/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#ifdef HAS_LINUX_EXT2_FS_H

#include <stdio.h>

#include "fs_ext2_quota_instance.h"

/*
    Doesn't have to take care of persistency as the QuotaManager
    guarantees that!
*/

NestReplyStatus
FsExt2QuotaInstance::add_entity(const char *name,
                                int is_user,
                                const char *dev,
                                double size,
                                int inode_count) {
    return NEST_SUCCESS;
}

NestReplyStatus
FsExt2QuotaInstance::delete_entity(const char *name,
                                   int is_user,
                                   const char *dev) {
    return NEST_SUCCESS;
}

NestReplyStatus
FsExt2QuotaInstance::update_entity(const char *name,
                                   int is_user,
                                   const char *dev,
                                   double new_size,
                                   int new_inode_count) {

    user_id_t   id;

    if (is_user) {
        if (NestGet_uid(&id, name, SystemPwdFile) < 0) {
            return NEST_USER_NOT_FOUND;
        }
    }
    else {
        group_ent   ge;
        if (Nestgetgrpnam(name, &ge, SystemGrpFile) < 0) {
            return NEST_GROUP_NOT_FOUND;
        }
        id = (user_id_t) ge.grp_gid;
    }

# ifdef DQF_MASK
    struct mem_dqblk    dqb;
# else
	struct dqblk		dqb;
# endif
    dqb.dqb_bhardlimit = dqb.dqb_bsoftlimit = NUM_BLKS(new_size);
    dqb.dqb_ihardlimit = dqb.dqb_isoftlimit = new_inode_count;
    dqb.dqb_btime = dqb.dqb_itime = 0;

    int cmd = QCMD(Q_SETQLIM, is_user ? USRQUOTA : GRPQUOTA);
    if (quotactl(cmd, dev, id, (caddr_t) &dqb) < 0) {
        return NEST_QUOTA_ENFORCEMENT_ERROR;
    }

    cmd = QCMD(Q_SYNC, is_user ? USRQUOTA : GRPQUOTA);
    if (quotactl(cmd, 0, id, NULL) < 0) {
        return NEST_QUOTA_ENFORCEMENT_ERROR;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
FsExt2QuotaInstance::view_entity_usage(const char *name,
                                       int is_user,
                                       const char *dev,
                                       double *size,
                                       int *inode_count) {

    if (!name || !dev || !size || !inode_count) {
        return NEST_QUOTA_ENFORCEMENT_ERROR;
    }

    user_id_t   id;

    if (is_user) {
        if (NestGet_uid(&id, name, SystemPwdFile) < 0) {
            return NEST_USER_NOT_FOUND;
        }
    }
    else {
        group_ent   ge;
        if (Nestgetgrpnam(name, &ge, SystemGrpFile) < 0) {
            return NEST_GROUP_NOT_FOUND;
        }
        id = (user_id_t) ge.grp_gid;
    }

# ifdef DQF_MASK
    struct mem_dqblk    dqb;
# else
	struct dqblk		dqb;
# endif

    int cmd = QCMD(Q_GETQUOTA, is_user ? USRQUOTA : GRPQUOTA);
    if (quotactl(cmd, dev, id, (caddr_t) &dqb) < 0) {
        return NEST_QUOTA_ENFORCEMENT_ERROR;
    }

    /*
    *size = SIZE_IN_MB(dqb.dqb_curblocks);
    */
# ifdef DQF_MASK
    *size = double(dqb.dqb_curspace) / 1024 / 1024 ;
# else
    *size = double(dqb.dqb_curblocks) / 1024 / 1024 ;
# endif
    *inode_count = dqb.dqb_curinodes;

    return NEST_SUCCESS;
}

NestReplyStatus
FsExt2QuotaInstance::initialize(const char *dev) {
# ifdef DQF_MASK
    struct mem_dqblk    dqb;
# else
	struct dqblk		dqb;
# endif
    user_id_t       id = 0;
    int             cmd;
    bool            first_time = true;

    while (1) {

        cmd = QCMD(Q_GETQUOTA, USRQUOTA);
        if (quotactl(cmd, dev, id, (caddr_t) &dqb) >= 0) {
            return NEST_SUCCESS;
        }

        double  size;
        int     inode_count;

        if (this->view_entity_usage("root",
                                    IS_USER,
                                    dev,
                                    &size,
                                    &inode_count) == NEST_SUCCESS) {
            return NEST_SUCCESS;
        }

        if (!first_time) {
            return NEST_NO_QUOTA_ENFORCEMENT;
        }

        cmd = QCMD(Q_QUOTAON, USRQUOTA);
        if (quotactl(cmd, dev, id, NULL) < 0) {
            if (errno != EBUSY) {
                return NEST_NO_QUOTA_ENFORCEMENT;
            }
        }

        first_time = false;
    }

    return NEST_NO_QUOTA_ENFORCEMENT;
}

#endif



