/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include <sys/utsname.h>
#include "nest_advertiser.h"
#include "nest_storage_manager.h"
#include "nest_collection_manager.h"
#include "nest_concurrency_manager.h"

class AdvertisingVisitor : public NestCollectionVisitor
{
public:
	AdvertisingVisitor( nestProtocol *dest ) { this->dest = dest; };
	~AdvertisingVisitor( void ) { };
	NestReplyStatus visit( const NestCollection &collection );
private:
	nestProtocol	*dest;
};

NestReplyStatus AdvertisingVisitor::visit( const NestCollection &collection )
{
    string buffer;
	const string *name = collection.getName( );
	const string *tmp;

	tmp = collection.getPath();
	if ( (!name->length()) || (!tmp->length()) ) {
		return NEST_BAD_COLLECTION;
	}

	buffer = *name + "Path = \"" + *tmp + "\"\n";

	time_t mtime = collection.getMtime( );
	if ( mtime > 0 ) {
		char	datebuf[32];
		NestSnprintf( datebuf, 32, "%d", mtime );
		buffer += *name + "Date = " + datebuf + "\n";
	}

	tmp = collection.getVersion( );
	if ( tmp->length() ) {
		buffer += *name + "Version = \"" + *tmp + "\"\n";
	}

	tmp = collection.getComment( );
	if ( tmp->length() ) {
		buffer += *name + "Comment = \"" + *tmp + "\"\n";
	}

	dest->Send( buffer );

	return NEST_SUCCESS;

};

nestAd::nestAd( nestProtocol *src, int len ) {
    myad  = (char *)NestMalloc( len + 1 );
    mylen = len;
    if ( len != src->Receiven( myad, len ) ) {
        nest_warn( __FILE__, __FUNCTION__, "Unable to read ad file\n" );
        NestFree( myad );
        myad = NULL;
        mylen = 0;
    }
    myad[len] = '\0';
}

nestAd::~nestAd( ) {
    if ( myad ) NestFree( myad );
}

char *
nestAd::toString( ) {
    return myad;
}

int
nestAd::length( ) {
    return mylen;
}

long uptime_start;

int
nestAdvertiser::init( ) {
    uptime_start = (long)time(NULL);
    return 0;
}

/*
NestReplyStatus
nestAdvertiser::readAd( nestProtocol *src ) {
        // TODO: read old collections out of file and pass to collection mngr
    return NEST_NOT_YET_IMPLEMENTED;
}
*/

nestAd *
nestAdvertiser::readAd( nestProtocol *src, int len ) {
    return ( len <= 0 ) 
        ? NULL
        : new nestAd( src, len );
}

NestReplyStatus
nestAdvertiser::writeAd( nestProtocol *dest ) {
    char buffer[NEST_MESSAGESIZE];
    struct utsname uname_buf;

    long now = (long)time(NULL);
    long uptime = now - uptime_start;


        // first write some meta data
    NestSnprintf( buffer, NEST_MESSAGESIZE, 
            "MyType = \"Storage\"\n" 
            "TargetType = \"Storage\"\n"
            "Subtype = \"%s\"\n",
            nestStorageManager::type()
            );
    dest->Sendn( buffer, NestStrlen( buffer ) );
    if ( uname( &uname_buf ) < 0 ) {
        nest_perror( "uname" );
    } else {
        NestSnprintf( buffer, NEST_MESSAGESIZE, "Name = \"%s\"\n", 
                uname_buf.nodename );
        dest->Sendn( buffer, NestStrlen( buffer ) );
    }

    
    NestSnprintf( buffer,
                  NEST_MESSAGESIZE,
                  "NestTime = %ld\n"
                  "Uptime   = %ld\n", 
                  now, uptime );
    dest->Sendn( buffer, NestStrlen( buffer ) );

        // now storage
    nestStorageManager::writeAd( dest );

        // now the supported security models
    NestSnprintf( buffer, NEST_MESSAGESIZE, "AuthenticationModel = \"GSI\"\n");
    dest->Sendn(buffer, NestStrlen(buffer) );

        // now the subnet
    NestSnprintf( buffer, NEST_MESSAGESIZE, "Subnet = \"%s\"\n", "my ip addr");
    dest->Sendn( buffer, NestStrlen( buffer ) );
        

        // now the concurrency
    nestConcurrencyManager::writeAd( dest );

        // now the debugging level
    NestSnprintf( buffer, NEST_MESSAGESIZE, "DebugLevel = %ld\n",
            get_nest_debug_level( ) );
    dest->Sendn( buffer, NestStrlen( buffer ) );

        // now the protocols
    for( int sd = nestProtocolManager::firstProtocolSocket();
            sd != nestProtocolManager::lastProtocolSocket();
            sd =  nestProtocolManager::nextProtocolSocket() )
    {
        if ( sd > 0 ) {
            NestSnprintf( buffer, NEST_MESSAGESIZE, "%sPort = %d\n",
                    nestProtocolManager::getProtocolNameFromSocket( sd ),
                    nestProtocolManager::getProtocolPortFromSocket( sd ) );
            dest->Sendn( buffer, NestStrlen( buffer ) );
        }
    }

        // now the subject name of the credential
    char *subject;
    long lifetime;
    NestReplyStatus status = nestProtocolManager::getCredentialInfo( 
            &subject, &lifetime );
    if ( status == NEST_SUCCESS ) {
        NestSnprintf( buffer, NEST_MESSAGESIZE, 
                "Credential_name     = \"%s\"\n"
                "Credential_lifetime = \"%ld\"\n",
                subject,
                lifetime );
        dest->Sendn( buffer, NestStrlen( buffer ) );
    }
        
        // now the local uid
    uid_t uid = NestGetuid();
    NestSnprintf( buffer, NEST_MESSAGESIZE, "uid = %d\n", uid );
    dest->Sendn( buffer, NestStrlen( buffer ) );


        // now the collections 
	NestCollectionManager	*colman = nestStorageManager::collection_manager;
	AdvertisingVisitor		*visitor = new AdvertisingVisitor( dest );
    if ( colman != NULL ) { // ghost storage doesn't have a collection_mgr
        colman->VisitAllCollections( *visitor );
    }


        // now the version
    NestSnprintf( buffer, NEST_MESSAGESIZE, "Version = %.2f\n", NEST_VERSION );
    dest->Sendn( buffer, NestStrlen( buffer ) );

    return NEST_SUCCESS;
}

