/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent,
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * For contact information, please visit:
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#ifndef NEST_CLIENT_H
#define NEST_CLIENT_H

#include "nest_common.h"
#include "nest_filetransfer.h"
#include "nest_request.h"
#include "nest_protocol.h"

#include "nest_proxy_fetch.h"

#include "nest_pthread_util.h"
#include "nest_advertiser.h"

#include "dynamic-lard.h"

#include <queue>

// why do I need forward declarations?????
class nestProtocol;
class nestFiletransfer;
class nestRequest;
class nestAd;

typedef NestReplyStatus ( taskFunction )( nestClient * ); 
typedef NestReplyStatus ( taskCleanupFunction ) (nestClient *, NestReplyStatus); 

class
nestClient {

public:

    nestClient( nestProtocol* connection, sockaddr_in *address );
	nestClient( const nestClient *);

    ~nestClient( );

    bool sendDirectoryList( NestFileInfo *list );

    bool sendAd( nestAd *ad );

    bool sendDiskspace( double availablebytes, double totalbytes );

    bool sendErrorMessage(  const char* filename, const char* func_name, int error );

    bool sendAck( );

    bool sendFinalAck( );

    bool sendFinalAck(  int transferType );

    bool sendInitialAck( );

    bool sendTransferInfo( );

    bool sendTimestamp( );

    bool sendPwd( const char *pwd );

    bool sendVersion( float version );

    bool sendBandwidthStatistics( );

    bool sendFilesize(  long filesize );

    bool sendFileinfo( NestFileInfo *info );   

    NestReplyStatus setFileattr( NestFileInfo *info, const char *fname, void *fattr );
    
    bool sendPhysName( const char *name );

    NestReplyStatus sendProxyIndex( DynamicLard *index );
    
    bool sendAcl( const char *user,
                  int is_user,
                  nest_access_rights_t access );

    bool sendAcl( const NestACLList *Acl );

    bool sendNestLot( NestLot *r);

    bool sendLotHandle( NestLotHandle *handle);

    NestReplyStatus sendFilehandle( NestFileInfo *handle );
    
    nestRequest *receiveRequest( fd_set *ready );

    void dump( );

    int storeRequest( nestRequest *req );

    void endRequest( NestReplyStatus status );

    NestReplyStatus assignStorageConnection( nestProtocol *storage );

    NestReplyStatus assignOtherStorageConnection( nestProtocol *storage);

    NestReplyStatus startTransfer( );
    NestReplyStatus startTransfer( nestProtocol *storage_con );
    NestReplyStatus startTransfer( int type, long length );

    NestReplyStatus endThirdPartyTransfer( );

    NestReplyStatus openThirdParty( char *protocol );
    
    void endTransfer( NestReplyStatus status );

    bool persistent();
    
    const char *toString();

    const char *getCwd();

    const char *ipAddress();
    void setIp( const char *ip );

    short getClientPort();

    sockaddr_in *getSocketAddress();

    NestReplyStatus setCwd( const char *dir );

    void setProxy( bool whether );
    void setProxy( const char *proxy_address );
    bool isProxy();
    const char *proxyAddress();
    
    nestRequest* getRequest( );
    nestRequest* getLastRequest();

    nestRequest* getProxyRequest();
    void assignProxyRequest( nestRequest *req );
    nestRequest* getLastProxyRequest();

    void assignProxyClient( nestClient *client );
    nestClient *getProxyClient( );

    void assignProxyFetch( nestProxyFetch * );
    nestProxyFetch *getProxyFetch();

    void assignTaskFunction( taskFunction *fp, taskCleanupFunction *fp );
    taskFunction * getTaskFunction();
    taskCleanupFunction * getTaskCleanupFunction();

    void setTaskStatus( NestReplyStatus );
    NestReplyStatus getTaskStatus( );

    nestFiletransfer* getTransfer( );
    nestFiletransfer* getLastTransfer( );

    char *getBuffer();

    nestProtocol* getConnection( );

    nestProtocol* getStorageConnection( );

    nestProtocol* getOtherStorageConnection( );

    void assignConnection( nestProtocol *connection );

    float getLastBandwidth( );

    int getId( );

    const char *getName();

    const char *getIdentity();

    void setName( const char * );

    void setIdentity( const char * );

    bool isActive();

    bool hasSocket();

	NestReplyStatus addDirtyWrite( nestRequest *req );
    nestRequest *getDirtyWrite( );
	
    void setFlushing( bool flushing );
    bool isFlushing( );

    void setActive();
    void setPassive();

    int getProtocol() { return getConnection()->getType(); } 

    void assignRequest( nestRequest *req );

    void   setLastAllocation( double allocation );
    double getLastAllocation( );

    void   setAllocation( double allocation );
    double getAllocation( );

    void        setUid( user_id_t uid );
    user_id_t   getUid();

    int operator==( const nestClient &other );

    void assignThirdPartyConnection( nestProtocol *third );
    nestProtocol *getThirdPartyConnection( );

private:

    void setState( int state );
    void my_endRequest( NestReplyStatus status );

    nestProtocol *connection;
    nestProtocol *storage;
    nestProtocol *otherStorage;
    nestProtocol *third_party;
    nestRequest *req, *last_req, *proxy_req, *last_proxy_req;
    nestFiletransfer *transfer, *last_transfer;

    queue< nestRequest *>  pending_reqs;
	queue< nestRequest *> *dirty_writes;

    bool flushing;  // flag that means it is flushing dirty writes

        // TODO: some configurations may not need thread safety
        // with globus threaded library they all do however
    pthread_mutex_t mutex;

    char buffer[FILEBUFFER];
    user_id_t uid;
    char *name;
    char *identity;
    char *cwd; 
    char *proxy_address;
    struct in_addr address;
    unsigned port;
    char *ip_address;

    nestClient *proxy_client;
    taskFunction *task_fp;
    taskCleanupFunction *cleanup_fp;
    NestReplyStatus taskStatus;
    long taskLength;
    nestProxyFetch *proxy_fetch;

    int state;
    bool waiting_to_close;
    bool is_proxy;
    int id;
    int receivedRequests;
    struct timespec connectTime;
    double last_allocation;
    double allocation;

};

#endif
