/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_config.h"
#include "nest_concurrency_manager.h"
#include "nest_concurrency_model.h"
#include "nest_concurrency_pot.h"
#include "nest_concurrency_nob.h"
#include "nest_concurrency_pop.h"
#include "network_limit.h"
#include "nest_scheduler.h"
#include "nest_filecache_model_mgr.h"

static fd_set *readfds, *writefds, *r_ready, *w_ready;

nestConcurrencyPot *pot_model;
nestConcurrencyPop *pop_model;
nestConcurrencyNob *nob_model;

// info about each concurrency, used for stats and for selection
struct model_list {
    nestConcurrencyModel *model;
    float bw;
    float total_bw;
    unsigned long kbsIn;
    unsigned long kbsOut;
    unsigned long transfers;
    unsigned long errors;
    unsigned long ave_concurrency;
    int current_tasks;
    int max_concurrency;
    int tickets;
} *models = NULL;
static int model_count = 0;

typedef struct {
    int current_tasks;
    unsigned long kbs_consumed;
    int stride;
} ProtocolStats;

ProtocolStats protocol_stats[PROTOCOL_COUNT];

int admission_control = INT_MAX;
nestClientPool *task_queue = NULL;
pthread_mutex_t *mutex = NULL;
nestFilecacheModelManager *filecache = NULL;

int max_concurrency = 0;
int current_tasks = 0;
unsigned long ave_concurrency = 0;
unsigned long total_sizes = 0;

unsigned long unscheduled_tasks    = 0;
unsigned long cachescheduled_tasks = 0;
unsigned long fifoscheduled_tasks  = 0;

enum {
    TAKE_TURNS,
    LOTTERY
};
int selection_criteria = LOTTERY;
int scheduling_policy = NEST_CACHE_AWARE;

// the callback to notify the main server
void (*notifyServer) (nestClient *, NestReplyStatus );


// TODO  make it so using favorites does not require admit control
// keep seperate count of current favorites
// only allow others if current favorites <= threshold (suggest 0)
// have seperate function called isFavorite( nestClient *)
// if no favorites, use cache-aware

int favoriteProtocol = NFS_PROTOCOL;

    // a single static network limit is shared between all threads
NetworkLimit *NestNetworkLimit = NULL;

int
nestConcurrencyManager::init( 
        fd_set *Readfds, fd_set *Writefds, fd_set *R_ready, fd_set *W_ready,
        void( *notifyServerCallback )(nestClient*,NestReplyStatus) )
{
    int worker_count = -1;
    bool use_clock = true;
    bool use_nob, use_pot, use_pop;
    use_pot = true;
    use_nob = false;
    use_pop = false;

    notifyServer = notifyServerCallback;

    const char *value = NULL;

    if (GetConfigValue(Config_ThreadPoolSize, &value) && *value != 0 ) {
        worker_count = atoi( value );
    } else {
        nest_die( "Must specify size of thread pool.\n" );
    }

    if (GetConfigValue(Config_AdmissionControl, &value) && *value != 0 ) {
        admission_control = atoi(value);
    }

    if (GetConfigValue(Config_Bandwidth, &value) && *value != 0 ) {
            nest_warn( __FILE__, __FUNCTION__, 
                    "Bandwidth limitation works only with threads.  "
                        "Disabling NOB and POP.\n" );
            use_nob = false;    
            use_pop = false;   
            NestNetworkLimit = new NetworkLimit( ); 
            NestNetworkLimit->setBandwidth((long)(atof(value) * 1024 * 1024));
    }

    if (GetConfigValue(Config_EnableThreadModel, &value) && *value ) {
        use_pot = (bool)atoi(value);
    }

    if (GetConfigValue(Config_EnableProcessModel, &value) && *value ) {
        use_pop = (bool)atoi(value);
    }

    if (GetConfigValue(Config_EnableEventModel, &value) && *value ) {
        use_nob = (bool)atoi(value);
    }

    if (GetConfigValue(Config_ConcurrencySelectionPolicy, &value) && *value!=0){
        if ( ! strcmp( value, "take-turns" ) ) {
            selection_criteria = TAKE_TURNS;
        } else if ( ! strcmp( value, "lottery" ) ) {
            selection_criteria = LOTTERY;
        } else {
            nest_die( "Bad arg (%s) to concurrency selection policy.\n", value);
        }
    }

    if (GetConfigValue(Config_CacheModelPolicy, &value) && *value!=0){
        if ( !strcmp(value, "clock" ) ) {
            use_clock = true;
        } else if ( !strcmp(value, "lru" )  ) {
            use_clock = false;
        } else {
            nest_die( "Bad arg %s for cache model replacement policy.\n",value);
        }
    }

    if (GetConfigValue(Config_SchedulingPolicy, &value) && *value!=0){
        if ( !strcmp(value, "fifo" ) ) {
            scheduling_policy = NEST_FIFO;
        } else if ( !strcmp(value, "sff" )  ) {
            scheduling_policy = NEST_SFF;
        } else if ( !strcmp(value, "cache" ) ) {
            scheduling_policy = NEST_CACHE_AWARE;
        } else {
            nest_die( "Bad arg (%s) for schedule command.\n", value );
        }
    }

    if (GetConfigValue(Config_PreferredProtocol, &value) && *value!=0){
        if ( !strcmp(value, "nfs" ) ) {
            favoriteProtocol = NFS_PROTOCOL;  
        } else if ( !strcmp(value, "http" )  ) {
            favoriteProtocol = HTTP_PROTOCOL; 
        } else if ( !strcmp(value, "chirp" ) ) {
            favoriteProtocol = NEST_PROTOCOL;
        } else if ( !strcmp(value, "gftp" ) ) {
            favoriteProtocol = GFTP_PROTOCOL;
        } else {
            nest_die( "Bad arg (%s) for protocol preference.\n", value );
        }
        nest_printf( "Preferring %s protocol.\n", value );
        scheduling_policy = NEST_PROTOCOL_PREFERENCE;
    }

    if ( ! use_nob && ! use_pot && ! use_pop ) {
        nest_die( "You must allow at least one concurrency model.\n" );
    }

        // seed random to be used for lottery concurrency selection
    srand( time(NULL) );

        // retain ptrs to the fd sets
    readfds  = Readfds;
    writefds = Writefds;
    r_ready  = R_ready;
    w_ready  = W_ready;

        // set up the classes we need for scheduling decisions
        // don't make task_queue thread safe, we'll do our own protection
    task_queue = new nestClientPool( false );
    mutex = (pthread_mutex_t *) NestMalloc( sizeof( pthread_mutex_t ) );
    nest_pthread_mutex_init( mutex, NULL );
    int cachemodel = ( (use_clock) ? NEST_CACHE_CLOCK : NEST_CACHE_LRU );
    if ( scheduling_policy == NEST_CACHE_AWARE ) {
        filecache  = new nestFilecacheModelManager( cachemodel );
    }

        // first start with equal strides
    for( int p = 0; p < PROTOCOL_COUNT; p++ ) {
        protocol_stats[p].kbs_consumed  = 0;
        protocol_stats[p].current_tasks = 0;
        protocol_stats[p].stride        = 1;
    }
    //protocol_stats[HTTP_PROTOCOL].stride = 4;
    
        // instantiate our concurrencies and set initial bandwidth values
    if ( worker_count <= 0 ) {
        use_pop = use_pot = false;
    }

    if ( use_nob ) {
        nob_model = new nestConcurrencyNob( readfds, writefds, taskDone, 
                model_count++ );
        nest_debug( NEST_DEBUG_CONCURRENCY, "Using nonblocking io.\n" );
    } else nob_model = NULL;

    if ( use_pop ) {
        pop_model = new nestConcurrencyPop( worker_count, readfds, taskDone,
                model_count++ );
        nest_debug( NEST_DEBUG_CONCURRENCY, "Created %d workers.\n", 
                worker_count );
    } else pop_model = NULL;

    if ( use_pot ) {
        pot_model = new nestConcurrencyPot( worker_count, readfds, taskDone,
                model_count++ );
        nest_debug( NEST_DEBUG_CONCURRENCY, "Created %d threads.\n", 
                worker_count );
    } else pot_model = NULL;

    setUpModelList();
    
    return 0;
}

// this function is used to compute the new weighted average 
// it tries to respect the new value without forgetting history too quickly
float
adjust_ave( float ave, float newval ) {
    
        // no bw should ever be no-pos
    if ( newval <= 0 ) {
        nest_warn( __FILE__, __FUNCTION__, "Ignoring weird new bw %.6f.\n", newval );
        return ave;
    }
    
        // if first time
    if ( ave == -1 ) {
            return newval;
    } else {
            // slowly age the average
        float new_ave = ( 0.95 * ave + 0.05 * newval );
        if ( new_ave <= 0 ) {
            nest_warn( __FILE__, __FUNCTION__, "New bw %.6f is <= 0....?", new_ave );
            return 0.00001;
        } else {
            return new_ave;
        }
    }
}


void
nestConcurrencyManager::setUpModelList( ) {

        // count models
    nest_debug( NEST_DEBUG_CONCURRENCY, "Init model list data\n" );

        // make an array for them and put them in
    models = (model_list *)
        NestCalloc( sizeof ( model_list ) * model_count );
    if ( nob_model != NULL ) {
        models[nob_model->getId()].model = nob_model;
    }
    if ( pop_model != NULL ) {
        models[pop_model->getId()].model = pop_model;
    }
    if ( pot_model != NULL ) {
        models[pot_model->getId()].model = pot_model;
    }

    for( int index = 0; index < model_count; index++ ) {
        models[index].bw = -1;
        models[index].kbsIn = models[index].kbsOut = 0;
        models[index].transfers = models[index].errors = 0;
    }
}

// this function is called whenever a task is completed
// return responsibility to server to listen again to the client
void
nestConcurrencyManager::taskDone( nestClient *client, NestReplyStatus status,
        int model_id ) 
{

        // look up the right model
    struct model_list *model = &models[model_id];
    taskFunction *chore = client->getTaskFunction();


        // end the transfer here, but don't tell the client that it's ended
        // once the transfer has been ended, the client might return
        // a closeConnectionRequest which will result in the client getting
        // deleting
    nestFiletransfer *transfer = client->getTransfer();
    const char *file = NULL;
    if ( transfer != NULL ) transfer->endTransfer( status );

    lockMutex();
    model->current_tasks--;
    current_tasks--;

        // adjust values only for successful transfers of positive size
        // ignore transfers of size 0, don't even count them
    long len = 0;
    if ( transfer != NULL ) {
        len = transfer->getBytesTransferred();
        if ( client->getRequest() != NULL ) {
            file = client->getRequest()->getNestFile();
        } else {
            assert( client->getProxyFetch() != NULL );
            file = client->getProxyFetch()->getProxyFile();
        }
    }

    if ( status == NEST_SUCCESS && len > 0 ) {
        float bw = transfer->getBandwidth();
        if ( bw < 0 ) {
            nest_warn( __FILE__, __FUNCTION__, "Received negative bw %.3f for a succesful transfer.\n",
                    bw );
        }

            // collect bandwidth info for future selections
            // may want to keep seperate bw info for short / long transfers
            // and maybe for different protocols as well
        model->bw = adjust_ave( model->bw, bw );
        model->total_bw += bw;
        model->transfers++;
        nest_debug( NEST_DEBUG_CONCURRENCY, 
                "%s_model transfer (%s) ended: %s.  "
                "%ld bytes transferred (%.3f MB/s)\n",
                model->model->toString(), 
                  nestProtocolManager::protocolName( client->getProtocol() ),
                NestErrorString( status ), len, bw );
        if ( transfer->isIncoming() ) {
            model->kbsIn += (len / 1024);
        } else {
            model->kbsOut += (len / 1024);
        }

        if ( scheduling_policy != NEST_PROTOCOL_PREFERENCE ) {
            protocol_stats[client->getProtocol()].kbs_consumed += (len / 1024); 
        }

        if ( filecache ) {
            filecache->taskDone( file, transfer, model_id, current_tasks );
        }
    } else if ( status != NEST_SUCCESS ) {
        nest_warn( __FILE__, __FUNCTION__, "Task failed: %s\n", NestErrorString( status ) );
        /*
        if ( status == NEST_NO_CONNECTION ) {
            // TODO: maybe we should set a flag in the client so it
            // will next return a close_connection_request
            // I think this should happen anyway though
            //client->setAlive( false );
        }
        */
        model->errors++;
    }

            // stride scheduling decisions here (if this was normal transfer)
    if ( chore == NULL ) {
        if ( scheduling_policy == NEST_PROTOCOL_PREFERENCE ) {
            strideDone( client, len );
            makeStrides(); 
        } else {
                // check for ability to schedule next task 
            makeSchedule();
        }
    }
    

        // return responsibility to server 
    notifyServer( client, status ); 

    unlockMutex();
    return;
    
}

int
nestConcurrencyManager::makeSchedule( ) {
    int started = 0;
    while ( task_queue->size() > 0 && current_tasks < admission_control ) {
        nestClient *nextclient = selectQueuedTask();
        nest_debug( NEST_DEBUG_CONCURRENCY, 
                "starting queued task for client %d. Queue len %d\n", 
                nextclient->getId(), task_queue->size() + 1 );
        startTask( nextclient );
        started++;
    }
    return started;
}

char *
nestConcurrencyManager::getConcurrencyModel( ) {
    static char model[1024];
    model[0] = '\0';
    if ( nob_model ) strcat( model, "nob " );
    if ( pop_model ) strcat( model, "pop " );
    if ( pot_model ) strcat( model, "pot " );
    return model;
}

void
nestConcurrencyManager::debugLevel( long level ) {
        // if we're using processes tell them all the new debug level also
    if ( pop_model != NULL ) pop_model->debugLevel( level );
}

nestConcurrencyModel *
nestConcurrencyManager::selectTakingTurns( ) {
    static int turn = 0;
    return models[turn++ % model_count].model;
}

ssize_t
nestConcurrencyManager::writeAd( nestProtocol *ad ) {
    char buffer[NEST_MESSAGESIZE];
    long total_in = 0, total_out = 0, total_trans = 0, total_err = 0;
    float total_bw = 0.0;

    NestSnprintf( buffer, NEST_MESSAGESIZE, "ConcurrencyModel = \"%s\"\n",
            getConcurrencyModel() );

    NestSnprintf( &buffer[NestStrlen(buffer)], 
            NEST_MESSAGESIZE - NestStrlen(buffer), 
            "Scheduling = \"%s\"\n",
            ( ( scheduling_policy == NEST_FIFO ) ? "fifo"
            : ( ( scheduling_policy == NEST_SFF ) ? "shortest file first" 
                : "cache-aware" ) ) );

    NestSnprintf( &buffer[NestStrlen(buffer)], 
            NEST_MESSAGESIZE - NestStrlen(buffer),
            "ConcurrencySelection = \"%s\"\n",
            ( ( selection_criteria == LOTTERY ) ? "lottery" : "take turns" ) );

    if ( admission_control != INT_MAX || task_queue->size() ) {
        float cache_sched_percent = (float)cachescheduled_tasks
            / ( cachescheduled_tasks + unscheduled_tasks + fifoscheduled_tasks)
            * 100;
        NestSnprintf( &buffer[NestStrlen(buffer)],
                      NEST_MESSAGESIZE - NestStrlen(buffer),
                      "Admission_Control = %d\n"
                      "Ave_Queue_Len  = %.4f\n"
                      "Max_Queue_Len  = %d\n"
                      "Cur_Queue_Len  = %d\n"
                      "Total_Q_Admits = %ld\n"
                      "Immediate_Scheduled = %ld\n"
                      "Cache_Scheduled     = %ld\n"
                      "Fifo_Scheduled      = %ld\n"
                      "Cache_Schedule_Perc = %.2f\n",
                      admission_control,
                      task_queue->getAveLen(),
                      task_queue->getMaxLen(),
                      task_queue->size(),
                      task_queue->getAdmissions(),
                      unscheduled_tasks,
                      cachescheduled_tasks, 
                      fifoscheduled_tasks,
                      cache_sched_percent );
    }

        // stride scheduling stuff
    for( int p = 0; p < PROTOCOL_COUNT; p++ ) {
        char protstring[25];
        NestSnprintf( protstring, sizeof(protstring), "%s_kbs",   
                      nestProtocolManager::protocolName( p ) );
        NestSnprintf( &buffer[NestStrlen(buffer)],
                NEST_MESSAGESIZE - NestStrlen(buffer),
                "%15s = %15ld\n",
                protstring,
                ( protocol_stats[p].kbs_consumed * protocol_stats[p].stride ) );
    }

    for ( int i = 0; i < model_count; i++ ) {
        NestSnprintf( &buffer[NestStrlen(buffer)],
                      NEST_MESSAGESIZE - NestStrlen(buffer),
                      "%s_bw = %.4f\n",
                      models[i].model->toString(),
                      models[i].bw );

        NestSnprintf( &buffer[NestStrlen(buffer)],
                      NEST_MESSAGESIZE - NestStrlen(buffer),
                      "%s_transfers = %ld\n",
                      models[i].model->toString(),
                      models[i].transfers );

        total_in    += models[i].kbsIn;
        total_out   += models[i].kbsOut;
        total_trans += models[i].transfers;
        total_err   += models[i].errors;
        total_bw    += models[i].total_bw;
    }
    int ret = NestSnprintf( &buffer[NestStrlen(buffer)],
                            NEST_MESSAGESIZE - NestStrlen(buffer),
                            "Total_kilobytes_in = %ld\n"
                            "Total_kilobytes_out = %ld\n"
                            "Total_transfers = %ld\n"
                            "Total_errors = %ld\n"
                            "Average_transfer_bw =%.4f\n"
                            "Current_concurrency = %d\n"
                            "Average_concurrency = %.4f\n"
                            "Max_concurrency = %d\n",
                            total_in, total_out, total_trans, total_err,
                            total_bw / total_trans,
                            current_tasks,
                            ( (float)ave_concurrency /
                              (total_trans + total_err + current_tasks) ),
                            max_concurrency );
    if ( ret < 0 ) {
        nest_warn( __FILE__, __FUNCTION__, "buffer truncated in %s:%s.\n", __FILE__, __FUNCTION__ );
    }

    ssize_t s = ad->Sendn( buffer, NestStrlen( buffer ) );
    if ( filecache ) s += filecache->writeAd( ad );
    return s;
}

// normally select highest average, but randomly allow others to get a chance
// the closer they are to the highest, the greater chance they should get 
nestConcurrencyModel *
nestConcurrencyManager::selectUsingAverages( long transfersize ) {

    bool hpdc_hack = false;
    if ( hpdc_hack ) {
        if ( transfersize <= 2048 ) {
            return nob_model;
        } else {
            return pot_model;
        }
    }
    
        // if anyone has not yet gone, let them go first
    for( int i = 0; i < model_count; i++ ) {
        if (models[i].transfers + models[i].errors == 0) return models[i].model;
    }

        // figure out max
    float max_bw = 0;
    for( int i = 0; i < model_count; i++ ) {
        max_bw = NEST_MAX( max_bw, models[i].bw );
    }

        // give everyone somewhere between 1 and 100 tickets and range them
    int total_tickets = 0;
    for( int i = 0; i < model_count; i++ ) {
        models[i].tickets = (int)(models[i].bw / max_bw * 100 );
        // inflation
    if ( models[i].bw == max_bw ) {
        models[i].tickets *= 3;
    } else {
        models[i].tickets /= 3;
    }
        if ( models[i].tickets <= 0 ) {
                // a model should not be allowed to have 0 tickets?
            nest_warn( __FILE__, __FUNCTION__,"%s had %d lottery tix\n",
                models[i].model->toString(),
                models[i].tickets );
            models[i].tickets = 1;
        }
    models[i].tickets += total_tickets;
        nest_debug(NEST_DEBUG_CONCURRENCY, "Lottery: %s got %d tix, %.3f bw.\n",
                models[i].model->toString(), 
                (models[i].tickets - total_tickets), 
                models[i].bw );
        total_tickets = models[i].tickets;
    }

        // have a lottery
    int winning_ticket = ( rand() % total_tickets );
    for( int i = 0; i < model_count; i++ ) {
        if ( winning_ticket < models[i].tickets ) {
            nest_debug( NEST_DEBUG_CONCURRENCY, "Lottery selected %s (%d)\n",
                    models[i].model->toString(), winning_ticket );
            return models[i].model;
        }
    }
        
        // if lottery math is correct, we should never get here
    assert( 0 );
    return NULL;
}

nestConcurrencyModel *
nestConcurrencyManager::selectConcurrencyModel( nestClient *client ) {

        // first of all, we may be constrained
    if ( model_count == 1 ) {
        return models[0].model;
    }

    if ( ! client->hasSocket() ) {
        nest_debug( NEST_DEBUG_CONCURRENCY, "Client has no sd, using pot.\n"); 
        return ( pot_model == NULL ) ? NULL : pot_model; 
    }

    if ( client->getTaskFunction() != NULL ) {
        // this means the client has specified their own specific function
        nest_debug(NEST_DEBUG_CONCURRENCY,"Special function, using pot.\n");
        return ( pot_model == NULL ) ? NULL : pot_model; 
    }

    if ( client->getRequest()->isThirdParty() ) {
        nest_debug(NEST_DEBUG_CONCURRENCY,"Third party transfer, using pot.\n");
        return ( pot_model == NULL ) ? NULL : pot_model; 
    }

    assert( client->getRequest() != NULL );
    long transferlength = 
            ( client->getRequest()->getEnd() == ENTIRE_FILE )
            ? ENTIRE_FILE : client->getRequest()->getLength();

    if ( transferlength == 0 ) {
        nest_debug( NEST_DEBUG_CONCURRENCY, "Trans of len zero, using pot.\n"); 
        return ( pot_model == NULL ) ? NULL : pot_model; 
    }
    
    if ( selection_criteria == LOTTERY ) {
        return selectUsingAverages( transferlength );
    } else {
        return selectTakingTurns();
    }
}

NestReplyStatus
nestConcurrencyManager::lockMutex( ) {
        // only necessary if multiple threads are calling taskDone
    if ( pot_model != NULL && pot_model->directCallback() ) {
        nest_pthread_mutex_lock( mutex );
    }
    return NEST_SUCCESS;
}

NestReplyStatus
nestConcurrencyManager::unlockMutex( ) {
    if ( pot_model != NULL && pot_model->directCallback() ) {
        nest_pthread_mutex_unlock( mutex );
    }
    return NEST_SUCCESS;
}

void
nestConcurrencyManager::strideDone( nestClient *client, long stride ) {
    int prot = client->getProtocol();
    long adjust = 0;
    
    if ( stride != client->getTransfer()->getLength() ) {
            // need to adjust - we initially set it either to the length
            // if it was known or we guessed
        long init = client->getTransfer()->getLength();
        if ( init < 0 ) init = client->getTransfer()->getLengthGuess();
        adjust = ( stride - init / 1024 ) / protocol_stats[prot].stride; 
    }

    if ( adjust ) protocol_stats[prot].kbs_consumed += adjust; 
    protocol_stats[prot].current_tasks--;

    nest_debug( NEST_DEBUG_CONCURRENCY | NEST_DEBUG_PROTOCOL_SCHED,
            "strideDone: client %d prot %d,  %ld total kbs, %d current\n",
            client->getId(),
            prot, protocol_stats[prot].kbs_consumed,
            protocol_stats[prot].current_tasks );

    assert( protocol_stats[prot].kbs_consumed  >= 0 );
    assert( protocol_stats[prot].current_tasks >= 0 );
}

NestReplyStatus
nestConcurrencyManager::makeStrides( ) {

    nestClient *client = NULL;
    NestReplyStatus status = NEST_SUCCESS;

    while ( task_queue->size() > 0 && ( client = selectLowestStride() ) ) {
            // here is a new client to schedule
        status = startTask( client );
        assert( status == NEST_SUCCESS );
        
            // adjust stride info
        long len = client->getTransfer()->getLength();
        if ( len < 0  ) {
                // we don't know the transfer length, just guess
            nest_debug( NEST_DEBUG_PROTOCOL_SCHED,
                    "Guessing at transfer length unknown.\n" );
            len = 8 * 1024; 
            client->getTransfer()->setLengthGuess( len );
        }
        int prot = client->getProtocol();
        long stride = ( len / 1024 ) / protocol_stats[prot].stride;
        protocol_stats[prot].kbs_consumed += stride; 
        protocol_stats[prot].current_tasks++;
        nest_debug( NEST_DEBUG_CONCURRENCY | NEST_DEBUG_PROTOCOL_SCHED,
                "Starting new stride: "
                "client %d: prot %d, %ld total kbs, %d current\n",
                client->getId(), prot, protocol_stats[prot].kbs_consumed,
                protocol_stats[prot].current_tasks );
    }
    return status;
}

unsigned long
nestConcurrencyManager::findLowestStride( ) {
        // first find lowest consumption of all active protocols
    unsigned long lowest = ULONG_MAX;
    for( int p = 0; p < PROTOCOL_COUNT; p++ ) {
        if (   protocol_stats[p].current_tasks > 0 
            && protocol_stats[p].kbs_consumed < lowest ) 
        {
            lowest = protocol_stats[p].kbs_consumed;
        }
    }
    return lowest;
}

// the counterpart to this function is taskDone 
// however this function is always called by admin thread
// whereas taskDone can be called by thread workers (actually not anymore)
NestReplyStatus
nestConcurrencyManager::addTask( nestClient *client ) { 

    NestReplyStatus status = NEST_SUCCESS;
    bool anticipatory = false;

        // take over responsibility for this client from the server
        // when task is done, return it
    if ( client->hasSocket() ) {
        NestFd_clear( client->getConnection()->getSocket(), readfds );
    }

    lockMutex();

        // new stride scheduling stuff
    if ( scheduling_policy == NEST_PROTOCOL_PREFERENCE ) {
        nest_debug( NEST_DEBUG_CONCURRENCY | NEST_DEBUG_PROTOCOL_SCHED,
                "Queuing client %d\n", client->getId() );
        task_queue->addClient( client );
        status = makeStrides( );
    } else if ( anticipatory ) {
            // just queue it now, schedule later
        nest_debug( NEST_DEBUG_CONCURRENCY | NEST_DEBUG_PROTOCOL_SCHED,
                "Queuing client %d\n", client->getId() );
        task_queue->addClient( client );
    } else {
            // check for admission_control
        if ( current_tasks >= admission_control ) {
            assert( current_tasks == admission_control );
            task_queue->addClient( client );
            nest_debug( NEST_DEBUG_TRANSFER | NEST_DEBUG_CONCURRENCY, 
                "Task queued (len %d) for client %d. Queue len %d\n", 
                client->getRequest()->getLength(),
                client->getId(), task_queue->size() );
            status = NEST_SUCCESS;
        } else {
            status = startTask( client );
            unscheduled_tasks++;
        }
    }

    unlockMutex( );
    return status; 
}

NestReplyStatus
nestConcurrencyManager::startTask( nestClient *client ) {
    NestReplyStatus status = NEST_SUCCESS;

    nestConcurrencyModel *model = selectConcurrencyModel( client );
    if ( model == NULL ) {
        nest_warn( __FILE__, __FUNCTION__, "Unable to find a concurrency model\n" );
        status = NEST_NOT_YET_IMPLEMENTED;
        return status;
    }
        
        // if we make it here, we're OK, go ahead and start the transfer
        // start the transfer here, end at taskDone
        // this will allow us to measure bw
        // note that pop worker will need to start it's own transfer

        // only do this stuff for normal transfer functions
        // the startTransfer might also fail especially with lot 
        // unavailability. 
    if ( client->getTaskFunction() == NULL ) {
        status = client->startTransfer();
        NEST_REPLY_CHECK(status);
        nestRequest *req = client->getRequest();
            // This was failing.  Move it up before addTask
            //assert( req != NULL );  

        if ( filecache ) {
            filecache->insert(req->getNestFile(), 
                    client->getTransfer(), req->getEnd());
        }
    }
    status = model->addTask( client );

        // adjust all our accounting info
    struct model_list *mod = &models[model->getId()];
    current_tasks++;
    mod->current_tasks++;
    max_concurrency = NEST_MAX( max_concurrency, current_tasks );
    mod->max_concurrency = NEST_MAX( mod->current_tasks, mod->max_concurrency );
    ave_concurrency += current_tasks;
    mod->ave_concurrency += mod->current_tasks;

    nest_debug( NEST_DEBUG_CONCURRENCY | NEST_DEBUG_TRANSFER, 
            "New task added to %s: %s.  Concurrency is %d\n", 
            model->toString(), 
            NestErrorString( status ),
            current_tasks );

    return status;
}

nestClient *
nestConcurrencyManager::selectFavoriteProtocol() {
    CLIENT_ITERATOR itr = task_queue->begin();
    nestClient *cursor = NULL;

        // if our favorite protocol is found, return it
        // else just the head
    while( itr != task_queue->end() ) {
        cursor = *itr++;
        if ( cursor->getProtocol() == favoriteProtocol ) {
            task_queue->removeClient( cursor );
            nest_debug( NEST_DEBUG_PROTOCOL_SCHED,
                    "Scheduling a favorite protocol.\n" );
            return cursor; 
        }
    }
        // no favorite protocols found, just return head
    return task_queue->getFront();
}

nestClient *
nestConcurrencyManager::selectQueuedTask( ) {

    nestClient *client = NULL;
    
    switch( scheduling_policy ) {
    case NEST_PROTOCOL_PREFERENCE:
        client = selectFavoriteProtocol();
        break;
    case NEST_FIFO:
        assert( ! task_queue->empty() );
        client = task_queue->getFront();
        break;
    case NEST_SFF:
        client = selectShortestFile();
        break;
    case NEST_CACHE_AWARE:
        client = selectCachedFile();
        break;
    default:
        nest_die( "Unknown scheduling policy\n" );
        return NULL;
    }
    return client;
}

nestClient *
nestConcurrencyManager::selectLowestStride( ) {
    unsigned long lowest = findLowestStride();

    CLIENT_ITERATOR itr = task_queue->begin();
    nestClient *cursor, *low = NULL;
    while( itr != task_queue->end() ) {
        cursor = *itr++;
        if ( protocol_stats[cursor->getProtocol()].kbs_consumed <= lowest ) {
            lowest = protocol_stats[cursor->getProtocol()].kbs_consumed;    
            low = cursor;
        }
    }

        // could be NULL if there are no queued clients for lowest strides
    if ( low != NULL ) task_queue->removeClient( low );
    return low;
}

nestClient *
nestConcurrencyManager::selectShortestFile( ) {
    CLIENT_ITERATOR itr = task_queue->begin();
    nestClient *cursor, *shortest = NULL;
    int shortestLength = INT_MAX; 

    while( itr != task_queue->end() ) {
        cursor = *itr++;
        if ( cursor->getRequest()->getLength() < shortestLength ) {
            shortest = cursor; 
            shortestLength = shortest->getRequest()->getLength();
        }
    }
    assert( shortest != NULL );
    task_queue->removeClient( shortest );
    return shortest;
}

nestClient *
nestConcurrencyManager::selectCachedFile( ) {
    CLIENT_ITERATOR itr = task_queue->begin();
    nestClient *cursor, *cached = NULL;

    assert( filecache );

    while( itr != task_queue->end() ) {
        cursor = *itr++;
        if ( filecache->contains(cursor->getRequest()->getNestFile())){
            cached = cursor;
            break;
        }
    }
    
    if ( cached != NULL ) {
        nest_debug( NEST_DEBUG_CACHEMODEL, 
                "Scheduling task for cached file %s for client %d.\n", 
                cached->getRequest()->getNestFile(),
                cached->getId() );
        task_queue->removeClient( cached );
        cachescheduled_tasks++;
        return cached;
    } else {
        cursor = task_queue->getFront();
        nest_debug( NEST_DEBUG_CACHEMODEL,
                "No queued reqs are cached.  Revert to FIFO for %s.\n",
                cursor->getRequest()->getNestFile());
        fifoscheduled_tasks++;
        return cursor;
    }
}

void
nestConcurrencyManager::makeProgress( ) {
        // check for ability to schedule more jobs now
        // turn off anticipatory scheduling for now
    //makeSchedule();
    
    if ( nob_model != NULL ) {
        nob_model->makeProgress( r_ready, w_ready);
    }
    if ( pop_model != NULL ) {
        pop_model->makeProgress( r_ready, w_ready );
    }
    if ( pot_model != NULL ) {
        pot_model->makeProgress( r_ready, w_ready );
    }
}

int
nestConcurrencyManager::getMaxfd( ) {
    int nob, pop, pot;
    nob = ( nob_model ) ? nob_model->getMaxfd() : 0;
    pop = ( pop_model ) ? pop_model->getMaxfd() : 0;
    pot = ( pot_model ) ? pot_model->getMaxfd() : 0;
    return NEST_MAX( nob, NEST_MAX( pot, pop ) );
}


// this function is called when worker children die.
// they shouldn't die but they might 
// I used to replace the dead workers within this function but
// this is a bad idea because malloc is not re-entrant
// so now just set a flag and clean up later 
// but it turns out that when workers die there are still problems
// workers shouldn't die however . . . 
void
nestConcurrencyManager::sigChild( int signal, siginfo_t *info, void *context ){
    if ( pop_model != NULL ) {
        pop_model->sigChild( signal, info, context, r_ready );
    }
}
