/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_concurrency_pop.h"

    // private helper functions
int workerLoop( int );
NestReplyStatus MessageHandler( char * msg, int whoami ); 

    // constructor
nestConcurrencyPop::nestConcurrencyPop( int pool_size, 
        fd_set *server_read_fds, taskDoneCallback cb, int id ) 
    : nestConcurrencyModel( id ) // call base class constructor
{
    this->taskDone = cb;
    this->readfds  = server_read_fds;
    this->deadKids = 0;
    pendingReqs = new nestClientPool( );
    workers     = new nestWorkerPool( pool_size, readfds, workerLoop );
}

NestReplyStatus
nestConcurrencyPop::addTask( nestClient *client ) {
    NestReplyStatus status = NEST_SUCCESS;
    nestWorker *worker;
    worker = workers->getIdleWorker();
    if ( ! worker ) {
        pendingReqs->addClient( client );
        nest_debug( NEST_DEBUG_POP, 
                "Queued req %d in pop due to lack of idle workers\n",
                pendingReqs->size() );
    } else {
        nest_debug( NEST_DEBUG_POP, 
                "assigning transfer to w%d\n", worker->getID() );
        status = worker->assignTask( client );
    }
    return status;
}

NestReplyStatus
nestConcurrencyPop::makeProgress( fd_set *r_ready, fd_set *w_ready ) {
    NestReplyStatus status = NEST_SUCCESS;
    HandleWorkerMessages( r_ready );
    if ( deadKids ) replaceDeadChildren( deadKids );
    return status;
}

char *
nestConcurrencyPop::toString() {
    static char name[] = "pop";
    return name;
}

int
nestConcurrencyPop::HandleWorkerMessages( fd_set *r_ready ) {

    int client_id, client_socket;
    NestReplyStatus status = 0;
    int worker_stream;
    int messages = 0;

    // nest_debug( NEST_DEBUG_POP, "Handling pop worker messages\n" );
    
    nestWorker *worker;
    int new_idle_workers = 0;
    struct IO_Measurement measurements;
    while ( ( worker = workers->canReceiveMessage( r_ready ) ) != NULL ) {
        worker_stream = worker->getStream();

            // actually receive the message here
        status = nestWorkerPool::ReceiveWorkerMessage( worker_stream, client_id,
                client_socket, &measurements );
        if ( status != NEST_SUCCESS ) {
            if ( status == NEST_NO_CONNECTION ) {
                nest_warn( __FILE__, __FUNCTION__, "POP: client %d lost after sending %ld bytes."
                        "  But not pruning.\n", client_id, 
                        measurements.total_bytes );
            } else {
                nest_warn( __FILE__, __FUNCTION__, "POP ReceiveWorkerMessage Error: %d: %s\n", 
                        status, NestErrorString( status ) );
            }
        }
        messages++;

            // end transfer and return responsibility for this client to server 
        nestClient *client = worker->getClient();
        client->getTransfer()->setMeasurements( &measurements );
        worker->endTask( status == NEST_SUCCESS, measurements.total_bytes );
        taskDone( client, status, getId() );

            // do some checking
            // TODO: This first assert failed once . . . ?
        assert( client->getId() == client_id );
        assert( client->hasSocket() );
        assert( client->getConnection()->getSocket() == client_socket );
        new_idle_workers++;
    }

        // assign any pending reqs to new workers 
    if( new_idle_workers) assignPendingReqs( new_idle_workers );

    if (messages) nest_debug( NEST_DEBUG_POP, 
            "Handled %d worker messages.\n", messages);
    return messages;
}

int
nestConcurrencyPop::getMaxfd() {
    return workers->getMaxfds();
}

void
nestConcurrencyPop::testWorkers( ) {
    nestWorker *worker = workers->getIdleWorker();
    char msg[NEST_MESSAGESIZE];
    
    nest_debug( NEST_DEBUG_POP, 
            "Sending 2 test messages to worker %d\n", worker->getPid());
    NestSnprintf( msg, sizeof(msg), "%d %s\n", TEST_MESSAGE, "hello, worker" );
    worker->sendMessage( msg );
    worker->sendMessage( msg );
}

NestReplyStatus
nestConcurrencyPop::debugLevel( long level ) {
    workers->debugLevel( level );
    return NEST_SUCCESS;
}

NestReplyStatus
nestConcurrencyPop::sigChild( int signal, siginfo_t *info, void *context,
        fd_set *readyfds ) 
{
        pid_t pid;
        int status;
        int saved_errno = errno;

        assert( signal == SIGCHLD || signal == SIGCLD );

        while( ( pid = waitpid( -1, &status, WNOHANG ) ) > 0 ) {
                // this just marks the worker as dead
            workers->removeWorker( pid, readfds, readyfds );
            deadKids++;
        }
        if ( pid != 0 ) {
            nest_perror( "waitpid" );
        }

        errno = saved_errno;

        return deadKids;
}

void
nestConcurrencyPop::replaceDeadChildren( int quantity ) {
    nest_warn( __FILE__, __FUNCTION__, "Replaced %d dead kids.\n", quantity );
    workers->createWorkers( quantity, readfds );
    assignPendingReqs( quantity );
    deadKids -= quantity; 
}

void
nestConcurrencyPop::assignPendingReqs( int new_idle_workers ) {

    nestClient *client;

    int i;
    for ( i = 0; i < new_idle_workers; i++ ) {

        if ( ( client = pendingReqs->getFront() ) != NULL ) {
            addTask( client );    
        } else {
            break;
        }    
    }
    if ( i ) nest_debug( NEST_DEBUG_POP, "assigned %d pending reqs.\n", i ); 
}

/** this is the function that each worker process lives in */
int
workerLoop( int id ) {

    char msg[NEST_MESSAGESIZE];

        // setup stdin and stdout so that we can read and write them just like
        // any other filedescriptors
    if ( setvbuf( stdin, NULL, _IOLBF, 0 ) != 0 
    ||   setvbuf( stdout, NULL, _IOLBF, 0 ) != 0 ) 
    {
        nest_perror( "setvbuf" );
        exit( -1 );
    }

        // ignore sigpipes, workers discover failed clients because blocking
        // transfer will fail with an EPIPE
    NestSignal( SIGPIPE, SIG_IGN );

    NestReplyStatus status;
    while( true ) {
        if (( status = ReceiveNestMessage(STDIN_FILENO, msg) ) != NEST_SUCCESS){
            if ( status == NEST_NO_CONNECTION ) {
                nest_debug( NEST_DEBUG_POP, "w%d Good-bye\n", id );
                exit( 0 );
            } else {
                nest_warn( __FILE__, __FUNCTION__, "w%d recv'd bad msg %s: %s\n",
                    id, msg, NestErrorString( status ) );
            }
        } else {
            nest_debug( NEST_DEBUG_POP, "w%d recv'd %s", id, msg );
            NestReplyStatus status = MessageHandler( msg, id );
            if ( status != NEST_SUCCESS ) {
                nest_debug(NEST_DEBUG_POP,"w%d message handler error: %s\n", id,
                        NestErrorString( status ) );
            }
        }
    }

    exit( 0 );

}

// this function is called by each worker to handle messages from the boss 
NestReplyStatus
MessageHandler( char * msg, int whoami ) {
    static char buffer[FILEBUFFER];

    int requestType = 0;
    int client_protocol = 0;
    int storage_protocol = 0;
    int other_storage_protocol = 0;
    long transferlen = 0;
    long offset = 0;
    long reqend = 0;

    char filename[NEST_MESSAGESIZE];
    char newfilename[NEST_MESSAGESIZE];
    char remotehost[NEST_MESSAGESIZE];
    char remotefile[NEST_MESSAGESIZE];

    NestReplyStatus retval = NEST_SUCCESS;
    int  client_id = 0;
    int  client_socket = 0;  // this is the server's socket for the client
                             // this is unneccessary info, TODO: remove.

    nestProtocol *client_con   = NULL;
    nestProtocol *storage_con  = NULL;
    nestProtocol *other_storage_con = NULL;
    nestRequest  *request      = NULL;
    nestFiletransfer *transfer = NULL;
    nestProtocol *src = NULL, *dst = NULL;

    if ( sscanf( msg, "%d", &requestType ) != 1 ) {
        nest_warn( __FILE__, __FUNCTION__, "W%d Bad formatted request (1): %s\n", whoami, msg );
        retval = NEST_BAD_PARAMETERS;
        goto DONE;
    }

    switch( requestType ) {
    case TEST_MESSAGE:
        nest_debug( NEST_DEBUG_POP, "w%d recv'd test msg:\n%s", whoami, msg );
        NestSnprintf( msg,
                      NEST_MESSAGESIZE,
                      "%d %d %d %s\n",
                      NEST_SUCCESS,
                      whoami,
                      TEST_MESSAGE,
                      "Hello, yourself boss." );

        SendNestMessage( STDOUT_FILENO, msg );
        nest_debug( NEST_DEBUG_POP, "w%d replied to boss\n", whoami );
        goto DONE;
    case PROXY_GET:
    case PROXY_PUT:
        nest_debug( NEST_DEBUG_PROXY, "w%d starting proxy op.\n", whoami );
            // fall through
    case CLIENT_RECV_FILE:
    case CLIENT_SEND_FILE:
            // do common setup for both file transfer requests
        if ( sscanf( msg,"%d %d %d %d %d %s %ld %ld", 
                    &requestType, 
                    &client_protocol, 
                    &storage_protocol,
                    &client_id, &client_socket,
                    filename, &offset, &transferlen ) != 8  )
        {
            nest_warn( __FILE__, __FUNCTION__, "bad formatted msg %s\n", msg );
            retval = NEST_BAD_PARAMETERS;
            goto DONE;
        } 
        storage_con = nestProtocolManager::ReceiveConnection( 
                STDIN_FILENO, storage_protocol );
        break;
    case COPY_REQUEST:
        if ( sscanf( msg,"%d %d %d %d %d %d %s %s %ld %ld",
                    &requestType,
                    &client_protocol,
                    &storage_protocol,
                    &other_storage_protocol,
                    &client_id, &client_socket,
                    filename, newfilename, &offset, &transferlen ) != 10  )
        {
            nest_warn( __FILE__, __FUNCTION__, "bad formatted msg %s\n", msg );
            retval = NEST_BAD_PARAMETERS;
            goto DONE;
        }
        storage_con = nestProtocolManager::ReceiveConnection(
                STDIN_FILENO, storage_protocol );
        other_storage_con = nestProtocolManager::ReceiveConnection(
                STDIN_FILENO, other_storage_protocol );
        break;
    case THIRD_PARTY_PUT:
    case THIRD_PARTY_GET:
            // do common setup for both third party transfers
        if ( sscanf( msg, "%d %d %d %d %d %s %s %s %ld",
                    &requestType, 
                    &client_protocol, 
                    &storage_protocol,
                    &client_id, &client_socket,
                    filename, remotefile, remotehost, &transferlen ) != 9 ) 
        {
            nest_warn( __FILE__, __FUNCTION__, "bad formatted msg %s\n", msg );
            retval = NEST_BAD_PARAMETERS;
            goto DONE;
        }
        storage_con = nestProtocolManager::ReceiveConnection(
                STDIN_FILENO, storage_protocol );
            // lseek was done when file was opened
            // worker doesn't know what offset was . . .
        offset = 0;
        break;
    case TEST_CLIENT_TO_SERVER:
    case TEST_SERVER_TO_CLIENT:
        if ( sscanf( msg, "%d %d %d %d %ld", &requestType, &client_protocol, 
                    &client_id, &client_socket, &transferlen ) != 5 ) 
        {
            nest_warn( __FILE__, __FUNCTION__, "bad formatted msg %s\n", msg );
            retval = NEST_BAD_PARAMETERS;
            goto DONE;
        }
        break;
    case DEBUG_REQUEST:
            // this is a special request, it is not a transfer request like
            // the others
        if ( sscanf( msg, "%d %ld", &requestType, &transferlen ) != 2 ) {
            nest_warn( __FILE__, __FUNCTION__, "bad formatted msg %s\n", msg );
            retval = NEST_BAD_PARAMETERS;
            goto DONE;
        }
        set_nest_debug_level( transferlen );
        nest_debug( NEST_DEBUG_DEBUG | NEST_DEBUG_POP, 
                "worker set debug to %ld\n", transferlen );
        goto DONE;
    default:
        nest_warn( __FILE__, __FUNCTION__, "w%d unknown msg type, msg %s\n", whoami, msg );
        retval = NEST_BAD_PARAMETERS;
        goto DONE;
    }

    client_con  = nestProtocolManager::ReceiveConnection( 
            STDIN_FILENO, client_protocol );
        
        // create necessary objects
        // TODO this is stupid hack
    reqend = ( transferlen == ENTIRE_FILE ) 
                    ? ENTIRE_FILE : ( transferlen + offset );
    if(requestType == COPY_REQUEST)
        request = new nestRequest( requestType, filename, newfilename, NULL,
                                   offset, reqend );
    else  
        request = new nestRequest( requestType, filename, remotehost, remotefile,
                offset, reqend );

    // TODO if 3pt, open 3pt connection here

        // now handle each type of file transfer specifically
    switch( requestType ) {
    case CLIENT_RECV_FILE: 
            // first send meta-info to client, then start sending 
        client_con->sendTransferInfo( request );
            // fall through to the proxys who don't send transfer info
    case PROXY_PUT:
        src = storage_con;
        dst = client_con;
        break;
    case CLIENT_SEND_FILE: 
            // send a go-ahead message to the client, then start receiving
        nest_debug( NEST_DEBUG_POP, "w%d sending initialack\n", whoami );
        client_con->sendInitialAck( );
        nest_debug( NEST_DEBUG_POP, "w%d sent initialack\n", whoami );
            // fall through to proxys, who don't send initial acks
    case PROXY_GET:
        src = client_con;
        dst = storage_con;
        break;
    case TEST_CLIENT_TO_SERVER:
        src = client_con;
        dst = NULL;
    case TEST_SERVER_TO_CLIENT: 
        src = NULL;
        dst = client_con;
        break;
    case THIRD_PARTY_PUT:
    case THIRD_PARTY_GET:
        nest_die( "w%d doing 3pt\n", whoami );
        break;
    case COPY_REQUEST:
        src = other_storage_con;
        dst = storage_con;
        break;
    default:
        nest_warn( __FILE__, __FUNCTION__, "Worker %d received weird msg: %s\n", whoami, msg );
        retval = NEST_UNKNOWN_REQUEST_TYPE; 
        goto DONE;
    }

        // create transfer object now
    transfer = new nestFiletransfer( 
            src, dst, transferlen, buffer, FILEBUFFER, requestType );
    transfer->startTransfer();

        // now do the actual filetransfer
    nest_debug( NEST_DEBUG_POP, "w%d entering blocking transfer\n", whoami );
    retval = transfer->BlockingTransfer( );
    nest_debug( NEST_DEBUG_POP, "w%d blocking transfer done: %s\n", whoami, 
            NestErrorString( retval ) );

        // do any final acking for a third party transfer
    if ( requestType == THIRD_PARTY_GET || requestType == THIRD_PARTY_PUT ) {
        nest_die( "Re-implement 3pt's now.\n" );
        //retval = client->endThirdPartyTransfer();
    }
    
        // send a final ack to client if transfer was OK 
    if ( retval == NEST_SUCCESS ) {
        nest_debug( NEST_DEBUG_POP, "w%d sending finalack\n", whoami );
        retval = ( client_con->sendFinalAck( requestType ) ? NEST_SUCCESS : 
                NEST_NO_CONNECTION );
        nest_debug( NEST_DEBUG_POP, "w%d sent finalack\n", whoami );
    } else {
        client_con->sendErrorMessage( __FILE__, __FUNCTION__,  retval );
    }

    DONE:

        // let the main server know that task is completed
    if ( ! nestWorkerPool::SendWorkerMessage( STDOUT_FILENO, whoami, client_id, 
                client_socket, retval, transfer ) ) 
    {
        nest_perror( "informServerTransferStatus" );
        assert( 0 );
    }

        // delete dynamically allocated classes 
        // later to optimize, should be persistent, just change internally
    if ( transfer )    delete transfer;
    if ( request  )    delete request;
    if ( client_con )  delete client_con;
    if ( storage_con ) delete storage_con;

    return retval;    
}
