/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_device_manager.h"

DeviceManager::DeviceManager() {
    devices = NULL;
    return;
}

DeviceManager::~DeviceManager() {
    return;
}

int
DeviceManager::add_entry(DiskEntry *d) {

    if (!d || !this->devices) {
        return -1;
    }

    string  str_key,
            str_class_ad = "";

    char    str_capacity[NEST_NAMELEN],
            str_user_usage[NEST_NAMELEN],
            str_group_usage[NEST_NAMELEN];

    NestSnprintf(str_capacity, NEST_NAMELEN, "%f", d->capacity);
    NestSnprintf(str_user_usage, NEST_NAMELEN, "%f", d->user_usage);
    NestSnprintf(str_group_usage, NEST_NAMELEN, "%f", d->group_usage);

    str_key = DEVICE_KEY_PREFIX;
    str_key = str_key + d->dev;

    str_class_ad =  str_class_ad +
                    "[  Type = \"" + DEVICE_KEY_PREFIX + "\"; " +
                    "   Device = \"" + d->dev + "\"; " +
                    "   Capacity = " + str_capacity + "; " +
                    "   UserUsage = " + str_user_usage + "; " +
                    "   GroupUsage = " + str_group_usage +
                    "]";

    ClassAd *device_class_ad = this->parser.ParseClassAd( str_class_ad, true);
    if (!device_class_ad) {
        return -1;
    }

    if (!this->devices->AddClassAd( str_key, device_class_ad )) {
        return -1;
    }

    memcpy((void *) &device_stat, (void *) d, sizeof(DiskEntry));
    return 0;
}

DiskEntry*
DeviceManager::get_entry(const char *dev) {

    if (!dev || !this->devices) {
        return NULL;
    }

    if (!strcmp(dev, device_stat.dev)) {
        return &device_stat;
    }

    return NULL;
}

double
DeviceManager::free_space(const char *dev, int is_user) {
    DiskEntry   *de;
    double usage;

    de = this->get_entry(dev);

    if (!de) {
        return 0.0;
    }

    if (is_user) {
        usage = de->user_usage;
    }
    else {
        usage = de->group_usage;
    }

    double free_space = de->capacity - usage;

    if (free_space < 0) {
        /*
            Should never happen!
        */
        nest_warn( __FILE__, __FUNCTION__, "Should never see this message! == capacity<total_usage!\n");
        return 0.0;
    }

    return free_space;
}

double
DeviceManager::free_space(int is_user) {
    return this->free_space(device_stat.dev, is_user);
}


NestReplyStatus
DeviceManager::get_best_device(char *dev,
                               double *avail,
                               int is_user) {

    double      usage;

    if (!dev || !avail) {
        return NEST_BAD_PARAMETERS;
    }

    usage = is_user ? device_stat.user_usage : device_stat.group_usage;

    strcpy(dev, device_stat.dev);
    *avail = device_stat.capacity - usage;

    return NEST_SUCCESS;
}

int
DeviceManager::incr_usage(const char *dev, double size, int is_user) {
    DiskEntry    *d = get_entry(dev);
    if (!d) {
        return -1;
    }
    if (is_user) {
        d->user_usage += size;
    }
    else {
        d->group_usage += size;
    }
    return 0;
}

int
DeviceManager::decr_usage(const char *dev, double size, int is_user) {
    DiskEntry    *d = get_entry(dev);
    if (!d) {
        return -1;
    }
    if (is_user) {
        d->user_usage -= size;
    }
    else {
        d->group_usage -= size;
    }
    return 0;
}


NestReplyStatus
DeviceManager::initialize(ClassAdCollection *resources,
                          char *dev,
                          double *storage_space,
                          double user_usage) {

    nest_debug( NEST_DEBUG_STORAGE,
                "DeviceManager::init(%s, %f, %f)\n",
                dev,
                *storage_space,
                user_usage );

    nest_fsstat     s_fs;
    DiskEntry       d;

    this->devices = resources;
    this->query.Bind(resources);
    this->devices_view = NEST_DEVICES_VIEW;

    ViewName    parent_view = "root";

    string view_constraint(""),
           view_rank(""),
           view_partition_expr("");

    view_constraint = view_constraint +
                      " other.Type == \"" + DEVICE_KEY_PREFIX + "\" ";

    if (!this->devices->CreateSubView( this->devices_view,
                                       parent_view,
                                       view_constraint,
                                       view_rank,
                                       view_partition_expr ) ) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "DeviceManager Couldn't create view for itself!\n");
        nest_debug(NEST_DEBUG_ALWAYS, "CondorErrMsg says: %s",
            CondorErrMsg.c_str());
        return NEST_CLASSAD_ERROR;
    }

    char path[NEST_MAXPATHLEN];

    if(!NestDeviceToPath(dev,path)) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "Device %s not recognized by getfsspec()-DeviceManager!\n",
                    dev );
        return ErrnoToNest(errno);
    }

    if (NestStatFS(path, &s_fs) < 0) {
        nest_perror("statfs failed:");
        nest_debug( NEST_DEBUG_ALWAYS,
                    "Cannot tolerate statfs fail!\n");
        return ErrnoToNest(errno);
    }

    strcpy(d.dev, dev);

    d.user_usage = d.group_usage = 0;
    d.capacity = ((s_fs.f_bavail * ((double)s_fs.f_bsize))/1024) / 1024;
    d.capacity += user_usage;

    if ((*storage_space >= 0) && (*storage_space < d.capacity)) {
        d.capacity = *storage_space;
    } else
        *storage_space = d.capacity;
    

    if (this->add_entry(&d) < 0) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Couldn't add device class ad to collection!\n" );
        return NEST_CLASSAD_ERROR;
    }

       nest_debug( NEST_DEBUG_STORAGE,
                "\n%s::UserUsage %lf| GroupUsage %lf::Capacity %lf\n",
                d.dev,
                d.user_usage,
                d.group_usage,
                d.capacity);


    return NEST_SUCCESS;
}

NestReplyStatus
DeviceManager::recover(ClassAdCollection *resources,
                       char *dev) {

    this->query.Bind(resources);
    this->devices = resources;
    this->devices_view = NEST_DEVICES_VIEW;

    this->query.Query(this->devices_view, NULL);
    this->query.ToFirst();

    string query_key;

    if (!this->query.Current(query_key)) {
        return NEST_UNKNOWN_ERROR;  //  No device found!
    }

    ClassAd *device_classad = this->devices->GetClassAd( query_key );
    if (!device_classad) {
        return NEST_CLASSAD_ERROR;
    }

    string  attr,
            str_dev;
    attr = "Device";
    device_classad->EvaluateAttrString(attr, str_dev);
    strcpy( device_stat.dev, str_dev.c_str() );
    attr = "Capacity";
    device_classad->EvaluateAttrNumber(attr, device_stat.capacity);
    attr = "UserUsage";
    device_classad->EvaluateAttrNumber(attr, device_stat.user_usage);
    attr = "GroupUsage";
    device_classad->EvaluateAttrNumber(attr, device_stat.group_usage);

    this->query.Query(NEST_LOT_CATALOG_VIEW, NULL);
    this->query.ToFirst();

    if (!this->query.Current(query_key)) {
        return NEST_SUCCESS;
    }

    double  size,
            tot_user_usage = 0,
            tot_group_usage = 0;

    bool    is_user,
            result;


    do {
        ClassAd *lot_class_ad = this->devices->GetClassAd(query_key);

        if (!lot_class_ad) {
            continue;
        }

        result = lot_class_ad->EvaluateAttrNumber( "Size", size );
        result &= lot_class_ad->EvaluateAttrBool( "IsUser", is_user );

        if (!result) {
            continue;
        }

        if (is_user) {
            tot_user_usage += size;
        }
        else {
            tot_group_usage += size;
        }

    } while (this->query.Next(query_key));

    device_stat.user_usage = tot_user_usage;
    device_stat.group_usage = tot_group_usage;

    if ((device_stat.capacity < device_stat.user_usage) ||
        (device_stat.capacity < device_stat.group_usage)) {
        nest_debug( NEST_DEBUG_STORAGE,
               "Weird! The sum of all lots granted is>device %s capacity!\n",
               device_stat.dev);
    }

    nest_fsstat   s_fs;
    char path[NEST_MAXPATHLEN];

    if(!NestDeviceToPath(device_stat.dev,path)) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "No information about the device %s !\n",
                    device_stat.dev );
        return NEST_UNKNOWN_ERROR;
    }

    if (NestStatFS(path, &s_fs) < 0) {
        nest_perror("statfs failed");
        nest_debug( NEST_DEBUG_STORAGE,
                    "cannot tolerate statfs fail!\n");
        return NEST_UNKNOWN_ERROR;
    }
    
    return NEST_SUCCESS;
}

NestReplyStatus
DeviceManager::synchronize() {
    return NEST_SUCCESS;
}

NestReplyStatus
DeviceManager::shutdown() {
    return NEST_SUCCESS;
}




