/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include "nest_error.h"
#include "nest_debug.h"

struct nest_error_s {
    char *desc;
    int  value;
};

struct nest_error_s nest_error_list[] = {
    { "Success", NEST_SUCCESS },
    { "Local file not found", NEST_LOCAL_FILE_NOT_FOUND },
    { "Remote file not found", NEST_REMOTE_FILE_NOT_FOUND },
    { "Insufficient server space", NEST_INSUFFICIENT_SPACE },
    { "No connection to server", NEST_NO_CONNECTION },
    { "Bad parameters", NEST_BAD_PARAMETERS },
    { "Not yet implemented", NEST_NOT_YET_IMPLEMENTED },
    { "Unknown request type", NEST_UNKNOWN_REQUEST_TYPE },
    { "Client lacks authorization", NEST_NONPRIVELEDGED_USER },
    { "User unknown", NEST_UNKNOWN_USER },
    { "Server is not run as root", NEST_NONPRIVELEDGED_SERVER },
    { "Entity already exists", NEST_ALREADY_EXISTS },
    { "Directory not empty", NEST_NOT_EMPTY },
    { "Received only partial message", NEST_PARTIAL_MESSAGE },
    { "Unknown NeST error", NEST_UNKNOWN_ERROR },
    { "Resource temporarily unavailable", NEST_TEMPORARILY_UNAVAILABLE },
    { "Unknown debug value", NEST_UNKNOWN_DEBUG_VALUE },
    { "Invalid authentication request", NEST_INVALID_AUTHENTICATION },
    { "Unable to authenticate", NEST_COULD_NOT_AUTHENTICATE },

    { "Server is out of resources - probably a transient error!",
        NEST_OUT_OF_RESOURCES },
    { "A config file or a system file is missing", NEST_SYSTEM_FILE_MISSING },

    { "File with the same name already exists", NEST_FILE_EXISTS },
    { "File not found", NEST_FILE_NOT_FOUND },
    { "Too many levels of symbolic links", NEST_TOO_MANY_LINKS },
    { "Filename too long", NEST_FILENAME_TOO_LONG },

    { "User already exists", NEST_USER_EXISTS },
    { "User not found", NEST_USER_NOT_FOUND },
    { "Maximum number of users reached", NEST_USER_LIMIT_REACHED },


    { "Group already exists", NEST_GROUP_EXISTS },
    { "Group not found", NEST_GROUP_NOT_FOUND },
    { "Maximum number of groups reached", NEST_GROUP_LIMIT_REACHED },
    { "Group consists of active users", NEST_GROUP_ACTIVE },

    { "No space left in the storage device", NEST_DISK_FULL },
    { "No quota enforcement", NEST_NO_QUOTA_ENFORCEMENT },
    { "Error returned by the quota enforcement module",
        NEST_QUOTA_ENFORCEMENT_ERROR },

    { "Lot has been scheduled", NEST_LOT_SCHEDULED },
    { "Respective Lot was not found", NEST_LOT_NOT_FOUND },
    { "Available Space in the Lots is not enough", NEST_LOT_NOT_ENOUGH },
    { "Lot is not empty", NEST_LOT_NOT_EMPTY },
    { "Maximum number of Lots reached", NEST_LOT_LIMIT_REACHED },
    { "Lot attributes specified are invalid", NEST_LOT_INVALID },
    { "Lots have been disabled", NEST_LOT_DISABLED },

    { "ClassAd operation failed", NEST_CLASSAD_ERROR },

    { "Filehandle has gone stale", NEST_STALE_FH },
    { "Interrupted system call", NEST_INTR },
    { "File is locked", NEST_FLOCK },
    
    { "A system file could not be updated", NEST_SYSTEM_FILE_UPDATE_FAILED },

    { "Failed to update the lot status", NEST_LOT_UPDATE_FAILED },
    
    { "Internal NeST server error - Inconsistent state", NEST_INCONSISTENT_STATE },

    { "Not yet initialized", NEST_NOT_YET_INITIALIZED },
    { "Not a file", NEST_NOT_A_FILE },
    { "Internal NeST server error - MAX_ERROR_VALUE", NEST_ERROR_MAX },
};

const char *
NestErrorString( NestReplyStatus error ) {
    int i;
    
    if( error < 0 ) error = NEST_UNKNOWN_ERROR;
    for ( i = 0; i < NEST_ERROR_MAX; i++ ) {
        if ( nest_error_list[i].value == error ) break;
    }
    return nest_error_list[i].desc;
}

int
NestToErrno( NestReplyStatus status ) {
    switch(status) {
    case NEST_SUCCESS:
        return 0;
    case NEST_INVALID_AUTHENTICATION:
    case NEST_COULD_NOT_AUTHENTICATE:
        return ECONNREFUSED; 
    case NEST_FILE_NOT_FOUND:
    case NEST_LOCAL_FILE_NOT_FOUND:
        return ENOENT;
    case NEST_REMOTE_FILE_NOT_FOUND:
        return ENOENT;
    case NEST_INSUFFICIENT_SPACE:
    case NEST_DISK_FULL:
        return ENOSPC;
    case NEST_LOT_NOT_ENOUGH:
        return EDQUOT;
    case NEST_NO_CONNECTION:
        return ENOTCONN;
    case NEST_BAD_PARAMETERS:
        return EINVAL;
    case NEST_NOT_EMPTY:
        return ENOTEMPTY;
    case NEST_NOT_YET_IMPLEMENTED:
        return ENOSYS;
    case NEST_UNKNOWN_REQUEST_TYPE:
        return EIO;
    case NEST_ALREADY_EXISTS:
        return EEXIST;
    case NEST_FILENAME_TOO_LONG:
        return ENAMETOOLONG;
    case NEST_FILE_EXISTS:
        return EISDIR;
    case NEST_FLOCK:
    case NEST_TEMPORARILY_UNAVAILABLE:
        return EAGAIN;
    case NEST_STALE_FH:
        return ESTALE;
    case NEST_INTR:
        return EINTR;
    case NEST_NOT_DIRECTORY:
        return ENOTDIR;
    default:
        nest_warn( __FILE__, __FUNCTION__, 
                "Add a value for nest error %d:%s to %s:%s.\n",
                status, NestErrorString( status ), __FILE__, __FUNCTION__ );
        return EINVAL;
    }
}

NestReplyStatus
ErrnoToNest( int error ) {
    switch( error ) {
        case( 0 ):
            return NEST_SUCCESS;
        case ENOTDIR:
            return NEST_NOT_DIRECTORY;
        case ENOENT:
            return NEST_REMOTE_FILE_NOT_FOUND;
        case EISDIR:
            return NEST_FILE_EXISTS;
        case EEXIST:
            return NEST_ALREADY_EXISTS;
        case ENAMETOOLONG:
            return NEST_FILENAME_TOO_LONG;
        case EPROTO:
            return NEST_BAD_PARAMETERS;
        case EPERM:
        case EACCES:
            return NEST_NONPRIVELEDGED_SERVER;
        case ENOTEMPTY:
            return NEST_NOT_EMPTY;
        case ECONNRESET:
        case ECONNREFUSED:
        case EHOSTUNREACH:
        case ENOTCONN:
        case EPIPE:
            return NEST_NO_CONNECTION;
        case ENOSYS:
            return NEST_NOT_YET_IMPLEMENTED;
        case EAGAIN:
            return NEST_TEMPORARILY_UNAVAILABLE;
        case ENOSPC:
            return NEST_DISK_FULL;
        case EDQUOT:
            return NEST_LOT_NOT_ENOUGH;
        case ENOMEM:
            return NEST_OUT_OF_RESOURCES;
        case ESTALE:
            return NEST_STALE_FH;
        case EINTR:
            return NEST_INTR;
        default:
            nest_warn( __FILE__, __FUNCTION__, "NestErrno called with unknown error (%d): %s\n",
                        error, strerror( error ) );
            return NEST_UNKNOWN_ERROR;
    }
}

