/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_nonblocking_transfers.h"
#include "nest_common.h"
#include "network_limit.h"
#include "nest_filetransfer.h"

extern NetworkLimit *NestNetworkLimit;

// TODO: cleanup non-blocking to work with connection objects not descriptors???

// static factory instantiate object that can return NULL on error
nestNonblockingTransfer*
nestNonblockingTransfer::newTransfer( nestClient *client, 
        NestReplyStatus &status ) 
{
        // only works with protocols that allow socket access
    if ( ! client->getConnection()->hasSocket() ) {
        status = NEST_NOT_YET_IMPLEMENTED;
        return NULL;
    }

    /*
        // don't do start transfer in here
        // do it in the concurrency manager
        // create a new object if and only if client->startTransfer is OK
    if ( ( status = client->startTransfer() ) != NEST_SUCCESS ) {
        return NULL;
    } else {
        return new nestNonblockingTransfer( client );
    }
    */
    return new nestNonblockingTransfer( client );
}

// constructor
nestNonblockingTransfer::nestNonblockingTransfer( nestClient *client ) {

    this->client        = client;
    this->acksent         = false;
    this->status          = WRITE_OK;
    this->buffer        = client->getBuffer();
    this->buffersize    = 0;
    this->bufferoffset  = 0;
    this->transferoffset= 0; 
    this->filetransfer     = client->getTransfer();
    this->measurements  = filetransfer->getMeasurements();
    this->type          = client->getRequest()->getType();
    this->src = this->dst = this->csd = NULL;

        // only works with sockets . . . 
    if ( getType() != TEST_SERVER_TO_CLIENT ) 
        assert( filetransfer->getSrc()->hasSocket() );
    if ( getType() != TEST_CLIENT_TO_SERVER ) 
        assert( filetransfer->getDst()->hasSocket() );

    switch( getType() ) {
    case CLIENT_RECV_FILE:
    case CLIENT_SEND_FILE:
    case THIRD_PARTY_PUT:
    case THIRD_PARTY_GET:
    case PROXY_GET:
    case PROXY_PUT:
        src = filetransfer->getSrc()->clone();
        dst = filetransfer->getDst()->clone(); 
        //src = filetransfer->getSrc();
        //dst = filetransfer->getDst();
        if ( getType() == CLIENT_RECV_FILE || getType() == PROXY_PUT ) {
            csd = dst;
        } else if ( getType() == CLIENT_SEND_FILE || getType() == PROXY_GET ) {
            csd = src;
        } else {
                // third party
            assert( client->getRequest()->isThirdParty() );
            //csd = client->getConnection()->clone();
            csd = client->getConnection();
        }
        assert( src != NULL && dst != NULL && csd != NULL );
        break;
    case TEST_SERVER_TO_CLIENT:
            // in this case, we just do a bunch of writes, no prior reads
            // so set the buffersize to full to simulate the initial read
        this->buffersize = FILEBUFFER; 
        NestBufferInit( this->buffer, FILEBUFFER );
        src = NULL;
        //dst = client->getConnection()->clone();
        dst = client->getConnection();
        break;
    case TEST_CLIENT_TO_SERVER:
        //src = client->getConnection()->clone(); 
        src = client->getConnection();
        dst = NULL;
        break;
    case COPY_REQUEST:
        src = filetransfer->getSrc()->clone();
        dst = filetransfer->getDst()->clone();
        csd = client->getConnection();
        // Initial Ack is not needed for Copy request
        this->acksent         = true;
        break;
    default:
        nest_die( "unknown transfer type\n" );
        break;
    }
}


nestNonblockingTransfer::~nestNonblockingTransfer( ) {
        // close our local dups
        // if we didn't clone then this wouldn't be necessary
        // however disabling cloning causes client puts to fail . . . 
    if ( src ) delete src;
    if ( dst ) delete dst;
    if ( getType() == THIRD_PARTY_GET || getType() == THIRD_PARTY_PUT ) {
        delete( csd );
    }
}

nestClient *
nestNonblockingTransfer::getClient( ) {
    return this->client;
}    

void
nestNonblockingTransfer::endTransfer( bool status ) {

        // do any final acking for a third party transfer
    int requestType = client->getRequest()->getType();
    if ( requestType == THIRD_PARTY_GET || requestType == THIRD_PARTY_PUT ) {
        client->endThirdPartyTransfer( );
    }

        // let the filetransfer object know how much was transferred
        // this could be done on every read / write if necessary
        // this shouldn't be necessary actually since we should be setting
        // it ourselves in the reads/writes
    //filetransfer->setBytesTransferred( transferoffset );
    
        // end the nonblocking transfer by telling the client to end it
        // this is now done in concurrencyMgr taskDone
    //client->endTransfer( status );
}

int
nestNonblockingTransfer::readSrc() {
    ssize_t targetsize, remaining; 
    struct timeval timer, end;
    long duration;

    assert( acksent );

    if ( filetransfer->getLength() != ENTIRE_FILE ) {
        remaining  = ( ssize_t )( filetransfer->getLength() - transferoffset );
        targetsize = ( remaining < FILEBUFFER ) ? remaining : FILEBUFFER; 
    } else {
        targetsize = FILEBUFFER;
    }
    
    
    NestGettimeofday( &timer );
    if ( ( buffersize = src->Receive( buffer, targetsize ) ) < 0 ) {
        nest_warn( __FILE__, __FUNCTION__, "read error: %s\n", strerror( errno ) );
        return CLIENT_LOST;
    }
    NestGettimeofday( &end );

    duration = NestTimediff( &timer, &end, __FUNCTION__ );
    measurements->total_in_time += duration; 
    if ( transferoffset == 0 ) {
        measurements->first_in_time = duration; 
        measurements->first_in_size = buffersize;
    }

        // if we return from a select and then read 0 bytes, client was lost
        // if length was unknown, then we're done with transfer
    if ( buffersize == 0 ) {
        return ( filetransfer->getLength() == ENTIRE_FILE ) 
            ? READ_DONE : CLIENT_LOST;
    }

        // reset bufferoffset for the soon to follow writeDst call
    bufferoffset = 0;

        // check to see if we're done if we know length of transfer 
    if ( filetransfer->getLength() != ENTIRE_FILE ) {
        if ( (long) buffersize + transferoffset >= filetransfer->getLength() ) {
            assert((long)buffersize+transferoffset==filetransfer->getLength());
            status = READ_DONE;
        } else { 
            status = READ_OK;
        }
    } else {
        status = READ_OK;
    }

    if ( getType() == TEST_CLIENT_TO_SERVER ) {
            // there will be no followup write call, simulate it here
        transferoffset += buffersize;
    }

    return status; 
}

// this function is the companion function to nestNonblockingTransfer::readSrc
// readSrc will have put some bytes from the src into a buffer
// writeDst will now write out of this buffer to the dst
// if writeDst finishes writing out of the buffer it will return
//    WRITE_DONE if the whole file has been transferred
//  WRITE_OK   if the buffer needs to be filled again by readSrc
// else it will return WRITE_CONTINUE to indicate that it did not write
// the complete buffer as filled by the readSrc function
// CLIENT_LOST is returned if the write returns an EPIPE 
// or an ECONNRESET signalling that the client side socket has been closed
int
nestNonblockingTransfer::writeDst() {
    ssize_t    nwrote = 0;
    ssize_t targetsize = 0;
    struct timeval timer, end;
    long duration;

        // TODO: this does only a straight byte copy from fd to fd
        // it is possible that the socket fd might have additional
        // metadata bytes that need to be written or read
        // for some protocols
    assert( acksent );

    targetsize = buffersize - bufferoffset;
    assert( targetsize >= 0 );
    if ( targetsize == 0 ) {
        assert( filetransfer->getLength() == ENTIRE_FILE );
        status = WRITE_DONE;
        return status;
    }

    NestGettimeofday( &timer );
    if ( (nwrote = dst->Send( &buffer[bufferoffset], targetsize ) ) < 0 ) {
        if ( errno == EPIPE || errno == ECONNRESET ) {
            status = CLIENT_LOST;
            return status;
        } else if ( errno == EAGAIN ) {
            nest_warn( __FILE__, __FUNCTION__,"Why would we get EAGAIN if we're using select?\n");
            nwrote = 0;
        } else {
            // TODO: don't die.  need to propogate to client
            nest_die( 
                "write error: %s\n"
                "dst is %d, bufoff %d, tarsize %d, bufsize %d, transoff %ld\n",
                strerror( errno ),
                dst, bufferoffset, targetsize, buffersize, transferoffset ); 
        }
    }
    NestGettimeofday( &end );

    duration = NestTimediff( &timer, &end, __FUNCTION__ );
    measurements->total_out_time += duration;
    if ( transferoffset == 0 ) {
        measurements->first_out_time = duration;
        measurements->first_out_size = nwrote; 
    }
    measurements->total_bytes += nwrote;

    if ( NestNetworkLimit != NULL ) NestNetworkLimit->consumeBandwidth(nwrote);

    bufferoffset += nwrote;

    if ( bufferoffset >= buffersize ) {
        assert( bufferoffset == buffersize );
        bufferoffset = 0;
        transferoffset += buffersize;
        if ( filetransfer->getLength() != ENTIRE_FILE ) {
            if ( transferoffset >= filetransfer->getLength() ) {
                assert( transferoffset == filetransfer->getLength() );
                status = WRITE_DONE;
            } else {
                status = WRITE_OK;
            }
        } else {
            status = WRITE_OK;
        }
    } else {
        status = WRITE_INCOMPLETE;
    }

    if ( getType() == TEST_SERVER_TO_CLIENT ) {
            // there will be no followup read call, simulate it here
        bufferoffset = 0;        
    }

    return status;
}

bool
nestNonblockingTransfer::ackSent() {
    return acksent;
}

// sendInitialAck is a misnomenclature
// it sends an ack if they client is going to send the file to let the client
// know that the request was accepted
// if the client is receiving the file however, then it sends transfer info
int
nestNonblockingTransfer::sendInitialAck() {

    bool retval; 

    switch( getType() ) {
    case CLIENT_RECV_FILE:
        retval = client->sendTransferInfo(  ); 
        break;
    case CLIENT_SEND_FILE: 
        retval = client->sendInitialAck( ); 
        break;
    case THIRD_PARTY_PUT:
    case THIRD_PARTY_GET:
    case TEST_CLIENT_TO_SERVER:
    case TEST_SERVER_TO_CLIENT:
    case PROXY_GET:
    case PROXY_PUT:
    case COPY_REQUEST:
        // no-op
        retval = true;
        break;
    default:
        assert( 0 );
    }

    acksent = true;
    
    if ( ! retval ) {
        this->status = CLIENT_LOST;
    }
    return this->status;
}

bool
nestNonblockingTransfer::sendFinalAck() {
    return client->sendFinalAck( );
}

bool
nestNonblockingTransfer::sendErrorMessage( const char* filename, const char* func_name, NestReplyStatus error ) {
    return client->sendErrorMessage( filename, func_name,  error );
}

// this function returns the duped client socket descriptor 
int
nestNonblockingTransfer::getCsd() {
    return ( csd ) ? csd->getSocket() : -1;
}

// this function returns the real client socket descriptor
int 
nestNonblockingTransfer::getRealClientSocket() {
    return client->getConnection()->getSocket();
}

int
nestNonblockingTransfer::getType() {
    return type;
}

// this function returns the dst descriptor 
int
nestNonblockingTransfer::getDst() {
    return ( dst ) ? dst->getSocket() : -1;
}

int
nestNonblockingTransfer::getStatus() {
    return status;
}

// this function returns the src descriptor 
int
nestNonblockingTransfer::getSrc() {
    return ( src ) ? src->getSocket() : -1;
}

int
nestNonblockingTransfer::getMaxfd() {
    return NEST_MAX( src->getSocket(), dst->getSocket() );
}
