/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_protocol_gridftp.h"
#include "nest_config.h"

// some global static stuff
globus_fifo_t nestProtocolGridFtp::fifo; 
pthread_mutex_t nestProtocolGridFtp::fifo_mutex;
int nestProtocolGridFtp::pipes[2];

enum FtpWaitWhich {
    WaitAuth,
    WaitAccept,
    WaitListen,
    WaitControlData,
    WaitDataConnect,
    WaitClose,
    WaitCommand,
};

class FtpWait {
  private:
    bool                        wait;
    globus_mutex_t              mutex;
    globus_cond_t               cond;
    const char                  *name;
    FtpWaitWhich                which;
    int                         id;
    nestProtocolGridFtp         *connection;

  public:
    FtpWait( const char *name, FtpWaitWhich which, int id,
             nestProtocolGridFtp *conn);
    ~FtpWait( void );

    // Initialize a wait; called by the constructor
    int StartWait( bool first = true );
    int WaitOnCallback( const char *func, int linenum );
    int CallbackDone( const char *func, int linenum, FtpWaitWhich which );
    int CancelWait( const char *func, int linenum );
    nestProtocolGridFtp *GetConnection( void ) { return connection; };
};

// Wait class methods: Constructor
FtpWait::FtpWait( 
    const char          *name,
    FtpWaitWhich        whichWait,
    int                 id,
    nestProtocolGridFtp *conn )
{
    this->name = NestStrdup( name );
    which = whichWait;
    this->id = id;
    this->connection = conn;

    wait = false;
    globus_mutex_init( &this->mutex, NULL );
    globus_cond_init( &this->cond, NULL );

    StartWait( true );
}

// Wait class methods: Descructor
FtpWait::~FtpWait( void )
{
    NestFree( (void *) name );
}

// Wait class methods: Start the wait process (call this *before* 
// registerring the callback function!!
int
FtpWait::StartWait( bool first)
{
    nest_debug( NEST_DEBUG_GRIDFTP, 
            "Gftp %d: Started waiting for %s\n", id, name );
    // If this isn the first call, we should not already be waiting...
    if ( first )
    {
        assert( wait == false );
    }
    wait = true;
    return 0;
}

// Wait class methods: Wait for the callback to be finish
// this function is called after registering callbacks.
// it turns the registered function into a blocking operation.
// StartWait() must be called before registering the callback.
int
FtpWait::WaitOnCallback( 
    const char  *funct,
    int         linenum )
{
    nest_debug( NEST_DEBUG_GRIDFTP, 
                "Gftp %d: Enter waitbusy for %s from %s:%d\n",
                id, name, funct, linenum );

    // Get exlusive access to 'wait' & cond...
    globus_mutex_lock( &this->mutex );
    while ( wait )
    {
        // !! This waits on 'cond' *and* unlocks 'mutex' !!
        globus_cond_wait( &this->cond, &this->mutex );
    }
        // No need to wait; just unlock the mutex
    globus_mutex_unlock( &this->mutex );

    nest_debug( NEST_DEBUG_GRIDFTP, 
                "Gftp %d: Leave waitbusy for %s from %s:%d\n", 
                id, name, funct, linenum );
    return 0;
}

// Wait class methods: Called to finish the callback;
int
FtpWait::CallbackDone( 
    const char      *funct,
    int             linenum,
    FtpWaitWhich    which )
{
    nest_debug( NEST_DEBUG_GRIDFTP,
                "Gftp %d: callbackDone for %s from %s:%d\n",
                id, name, funct, linenum );
    assert( wait == true );
    assert( which == this->which );
    globus_mutex_lock( &this->mutex );
    wait = false;
    globus_cond_signal( &this->cond );
    globus_mutex_unlock( &this->mutex );
    return 0;
}

// Wait class methods: Call when callback fails to register
int
FtpWait::CancelWait( 
    const char  *funct,
    int         linenum ) 
{
    nest_debug( NEST_DEBUG_GRIDFTP,
                "Gftp %d: CancelWait for %s from %s:%d\n",
                id, name, funct, linenum );
    nest_debug( NEST_DEBUG_GRIDFTP, "\tWait %d->0\n", wait );
    assert( wait == true );
    wait = false;
    return 0;
}

// constructor
nestProtocolGridFtp::nestProtocolGridFtp( globus_ftp_control_handle_t *handle,
        int id )
{
    nest_debug(NEST_DEBUG_GRIDFTP,"new gftp connection %d: handle 0x%p\n",
            id, handle );

    this->username = NULL;
    this->globusname = NULL;

    this->handle  = handle;
    this->id      = id;
    this->request = NULL;
    this->reqType = -1;
    this->isNew   = true;
    this->offset  = 0;
    this->eof     = false;
    this->transmissionLength = 0;
    this->bufsize = 0;
    this->offset  = 0;
}

// destructor
nestProtocolGridFtp::~nestProtocolGridFtp() {
    if ( this->reqType != CLOSE_CONNECTION_REQUEST ) {
        nest_warn( __FILE__, __FUNCTION__, "Gftp %d: killing a connection who's last req != close.\n",
                id );
    } else {
            // ack the close request
        sendAck();
    }

    if ( username != NULL )   NestFree( username );
    if ( globusname != NULL ) NestFree( globusname );

    if ( (authResult != GLOBUS_TRUE) || (acceptResult != GLOBUS_TRUE) ) {
        nest_warn( __FILE__, __FUNCTION__, "Gtp %d: Unable to remove a gftp handle, leaking it . . .\n",
                id );
        return;
    }

    globus_result_t result;
    // need to call globus_ftp_control_handle_destroy
    // the documentation says that first we must call either
    // globus_ftp_control_force_close or globus_ftp_control_quit
    // I think quit is for the client

    nest_debug( NEST_DEBUG_GRIDFTP, "Closing globus connection %d\n", id );
    FtpWait wait( "Close", WaitClose, id, this );
    // this dies with an assertion in globus_ftp_control_data.c:6825 if 
    // we haven't successfully authenticated
    result = globus_ftp_control_force_close(handle, closeCallback, &wait);
    //result = globus_ftp_control_quit( handle, closeCallback, &wait );
    if ( result == GLOBUS_SUCCESS ) {
        nest_debug(NEST_DEBUG_GRIDFTP, "waiting for closeCallback %d\n",id);
        wait.WaitOnCallback( __FUNCTION__, __LINE__ );
    } else {
        wait.CancelWait( __FUNCTION__, __LINE__ );
        nest_warn( __FILE__, __FUNCTION__, "Could not close handle: %s\n",
                globusResultToString( result ) );
    }

    nest_debug( NEST_DEBUG_GRIDFTP, "Destroying globus handle\n" );
    result = globus_ftp_control_handle_destroy( this->handle );
    if ( result != GLOBUS_SUCCESS ) {
            // TODO this is seen often when we run with simultaneous different
            // protocols.  If it runs for too long, it will then segfault
            // somewhere in the globus code
        nest_warn( __FILE__, __FUNCTION__, "Could not destroy handle: %s\n", 
                globusResultToString( result ) );
        nest_warn( __FILE__, __FUNCTION__, "Gtp %d: Unable to remove a gftp handle, leaking it . . .\n",
                id );
    } else {
        NestFree( this->handle );
    }
    
}

globus_ftp_control_handle_t *
nestProtocolGridFtp::getHandle() {
    return handle;
}

bool
nestProtocolGridFtp::isNewConnection( ) {
    bool ret = isNew;
        // a connection can only be new once
    isNew = false;
    return ret;
}

// static initialization routine
int
nestProtocolGridFtp::listen( int port ) {
    globus_result_t     result;
    globus_ftp_control_server_t      *server;
    
    globus_module_activate( GLOBUS_FTP_CONTROL_MODULE );

    server = (globus_ftp_control_server_t *)
        NestMalloc( sizeof( struct globus_ftp_control_server_s ) );
    
    result = globus_ftp_control_server_handle_init( server );
    if ( result != GLOBUS_SUCCESS ) {
        nest_warn( __FILE__, __FUNCTION__, 
                "globus server handle init error: %s\n",
                globusResultToString( result ) );
        NestFree( server );
        return -1;
    }
    
        // no need to wait for the listen callback
    short unsigned Port = (short unsigned) port;
    result = globus_ftp_control_server_listen( server, &Port, 
            listenCallback, NULL );
    if ( result != GLOBUS_SUCCESS ) {
        NestFree( server );
        return -1;
    }

        // do some initialization stuff, but do it only once
    static int count = 0;

    if ( count++ == 0 ) {

        if ( globus_fifo_init( &fifo ) < 0 ) {
            nest_perror( "globus_fifo_init" );
            return -1;
        }
        
        if ( NestSpipe( pipes ) < 0 ) {
            nest_perror( "s_pipe" );
            return -1;
        }

            // check for globus threads, if we are using non-threaded globus
            // then launch our own pthread here
        #ifndef HAVE_PTHREAD
        nest_debug( NEST_DEBUG_GRIDFTP, 
                "Launching thread for globus stuff.\n" );
        pthread_t thread;
        nest_pthread_create( &thread, NULL, GlobusCondWait, NULL );
        nest_pthread_mutex_init( &fifo_mutex, NULL );
        #endif
    }
        // use the pipe to signal nest
    return pipes[1];
}

nestProtocol*
nestProtocolGridFtp::accept( int socket ) {
    nest_debug( NEST_DEBUG_GRIDFTP, "Globus accept called\n" );
    clearNestSignal();
        // this should return NULL if empty.  Addendum: it does.
    return deQueueConnection(); 
}

NestReplyStatus
nestProtocolGridFtp::gssAccept( char **username , char **globus_username ) {
    *username = NestStrdup( this->username );
    *globus_username = NestStrdup( this->globusname );
    nest_debug( NEST_DEBUG_AUTH, "Using already acquired authentication "
            "for %s using gftp.\n", this->username );
    return NEST_SUCCESS;
}

static int signals = 0;

void
nestProtocolGridFtp::signalNest() {
    static char msg[256];
    signals++;
    nest_debug( NEST_DEBUG_GRIDFTP, 
            "Globus signaling Nest on socket %d, %d signals\n", 
            pipes[0], signals );
    NestSnprintf( msg, 256, "Globus needs attention\n" );
    NestWriten( pipes[0], msg, 256 );
}

void
nestProtocolGridFtp::clearNestSignal() {
    //clear message from pipes[1]
    static char msg[256];
    nest_debug( NEST_DEBUG_GRIDFTP, 
           "Clearing globus pipe %d, %d pending signals\n", pipes[1], signals );
    if ( signals > 0 ) {
        signals--;
        NestReadn( pipes[1], msg, 256 );
        assert( strcmp( msg, "Globus needs attention\n" ) == 0 );
        nest_debug( NEST_DEBUG_GRIDFTP, 
            "Cleared globus pipe, %d pending signals\n", signals );
    }
}

// only necessary to lock and unlock if we are not using globus threads
void
nestProtocolGridFtp::lockFifo() {
    #ifndef HAVE_PTHREAD
    nest_pthread_mutex_lock( &fifo_mutex );
    #endif
}

void
nestProtocolGridFtp::unlockFifo() {
    #ifndef HAVE_PTHREAD
    nest_pthread_mutex_unlock( &fifo_mutex );
    #endif
}


nestProtocol*
nestProtocolGridFtp::deQueueConnection( ) {
    nestProtocol *connection;
    lockFifo();
    connection = (nestProtocol *) globus_fifo_dequeue( &fifo );
    nest_debug( NEST_DEBUG_GRIDFTP, "fifo dequeue, size is %d.\n", 
            globus_fifo_size(&fifo) ); 
    unlockFifo();
    return connection;
}

int
nestProtocolGridFtp::enQueueConnection( nestProtocol *connection ) {
    lockFifo();
    int ret = globus_fifo_enqueue( &fifo, (void *) connection );
    nest_debug( NEST_DEBUG_GRIDFTP, "fifo enqueue, size is %d.\n", 
            globus_fifo_size(&fifo) ); 
    unlockFifo();
    return ret; 
}

void *
nestProtocolGridFtp::GlobusCondWait( void *args ) {
    globus_cond_t cond;
    globus_mutex_t mutex;
    globus_mutex_init( &mutex, 0 );
    globus_cond_init( &cond, 0);
    while ( true ) globus_cond_wait( &cond, &mutex );
}

void
nestProtocolGridFtp::assignRequest( nestRequest *request ) {
    // if this assertion ever fails, we will need to make a stack of requests
    // not just a single one
    assert( this->request == NULL );
    this->request = request;
    this->reqType = request->getType();
}

// once a request has been given away, it is made NULL
nestRequest *
nestProtocolGridFtp::getRequest( ) {
    nestRequest *req = this->request;
    this->request = NULL;
    return req;
}

// be careful, this is a blocking operation
ssize_t
nestProtocolGridFtp::Receive( char *recv_buffer, ssize_t n ) {

        // we may already have eof status from the previous receive
    if ( this->eof == GLOBUS_TRUE ) {
        nest_debug( NEST_DEBUG_GRIDFTP, "GFTP %d receive EOF previously\n", 
                id );
        return 0;
    }

    // Create our wait object
    FtpWait wait( "ControlData", WaitControlData, id, this );
    
    globus_result_t result = GLOBUS_SUCCESS;
    this->transmissionLength = 0;
    // we may have to try reading multiple times . . . 
    int attempts = 0;
    while ( result == GLOBUS_SUCCESS 
            && this->transmissionLength == 0 
            && this->eof != GLOBUS_TRUE ) 
    {
        attempts++;
        wait.StartWait( false );
        result = globus_ftp_control_data_read( 
                this->handle,
                (globus_byte_t *)recv_buffer, 
                n,
                controlDataCallback, &wait );
        if ( result == GLOBUS_SUCCESS ) {
            wait.WaitOnCallback( __FUNCTION__, __LINE__ );
            result = this->transferStatus;
        } else {
            wait.CancelWait( __FUNCTION__, __LINE__ );
        }

            // for some reason, must occasionally wait multiple times . . .
            // I think this is wrong actually . . . waiting multiple times
            // put some code in to break out and see if we ever get the below
            // message in normal situations.  The only time we maybe should get
            // it is when the client does an abort . . .
        if ( attempts >= 10 ) break;
    } 
    if ( attempts > 1 ) {
        nest_warn( __FILE__, __FUNCTION__, "Multiple waits (%d) for globus read\n", attempts );
    }

    // when we break, we either received bytes, error or eof
    if ( result != GLOBUS_SUCCESS ) {
            // we may fail to receive because we're at eof
        if ( this->eof == GLOBUS_TRUE ) {
            nest_debug( NEST_DEBUG_GRIDFTP, "GFTP %d receive EOF\n", id );
            return 0;
        } else {
            nest_warn( __FILE__, __FUNCTION__, "GfTP connection receive err: %s\n", 
                globusResultToString( result ) );
            return -1;
        }
    }

    nest_debug( NEST_DEBUG_GRIDFTP, 
            "Gftp conn %d Received %d bytes offset %d\n\tbuf: %.10s\n", 
            this->id, this->transmissionLength, this->offset,
            recv_buffer );
    return ( (ssize_t) this->transmissionLength );
}

// be careful, this is a blocking operation
ssize_t
nestProtocolGridFtp::Send( const char *buffer, ssize_t n ) {
    globus_result_t result;

    // Create the wait object
    FtpWait wait( "ControlData/Send", WaitControlData, id, this );

    // globus claims that the callback will only be called when ALL
    // of the data has been written - this means there is no need
    // to attempt to check whether all data has been written
    //nest_printf( "gftp con %d sending %d bytes\n", id, n );
    this->transmissionLength = 0;

    result = globus_ftp_control_data_write( this->handle, 
            (globus_byte_t *)buffer, n, this->offset, 
            GLOBUS_FALSE, controlDataCallback, &wait );
    if ( result == GLOBUS_SUCCESS ) {
        wait.WaitOnCallback( __FUNCTION__, __LINE__ );
        result = transferStatus;
        if ( result != GLOBUS_SUCCESS ) {
            nest_warn( __FILE__, __FUNCTION__, "GfTP connection sendn err: %s\n", 
                globusResultToString( result ) );
            return -1;
        }
    } else {
        wait.CancelWait( __FUNCTION__, __LINE__ );
        nest_warn( __FILE__, __FUNCTION__, "Unable to register a globus_data_write: %s\n",
                globusResultToString( result ) );
    }

    /*
    nest_printf( "Tried to send %d bytes in gftp, sent %d\n",
                n, this->transmissionLength );
    */
    if ( n != this->transmissionLength ) {
        nest_warn( __FILE__, __FUNCTION__, "Tried to send %d bytes in gftp, only sent %d\n",
                n, this->transmissionLength );
        //assert( n == this->transmissionLength );
    }
    this->offset += this->transmissionLength;
    return this->transmissionLength;
}

// callback for globus_ftp_control_data_write and
//              globus_ftp_control_data_read
void
nestProtocolGridFtp::controlDataCallback( 
                    void *callback_arg,
                    globus_ftp_control_handle_t * handle, 
                    globus_object_t * error, 
                    globus_byte_t * buffer, 
                    globus_size_t length, 
                    globus_off_t offset, 
                    globus_bool_t eof )
{
    FtpWait *wait = (FtpWait *) callback_arg;
    nestProtocolGridFtp *connection = wait->GetConnection( );
    assert( connection != NULL );

        // save the error status in our connection object
        // sometimes, the register for controlDataCallback succeeds but
        // the callback fails.  Since we wait for the callback done, put
        // any error code into the connection and check it after the callback
    connection->transferStatus = error;

    if( error != GLOBUS_SUCCESS ) {
        nest_warn( __FILE__, __FUNCTION__,">> controlDataCallback: error %s\n",
                globus_object_printable_to_string( error ) );
        // need to clean up all transfer state here
        // maybe all connection state as well
        // TODO - we get here due to client error - must remove client and
        // clean up somehow
        wait->CallbackDone( __FUNCTION__, __LINE__, WaitControlData );
        return;    
    } 

    nest_debug( NEST_DEBUG_GRIDFTP, 
        "controlDataCallback called for gftp %d.  Will set wait to false.\n",
        connection->id );

    nest_debug( NEST_DEBUG_GRIDFTP,
            "gftp old offset: %10lld, current %10lld, length %10ld, "
            "eof %3d\n\tbuffer %.70s\n", 
            connection->offset, offset, length, eof, buffer );
    if ( offset != connection->offset ) {
        nest_debug( NEST_DEBUG_GRIDFTP, "offset %lld != %lld\n",
                offset, connection->offset );
    }
    connection->transmissionLength = length;
    connection->offset += length; 
    connection->eof = eof;
    wait->CallbackDone( __FUNCTION__, __LINE__, WaitControlData );
}

bool 
nestProtocolGridFtp::sendDirectoryList( NestFileInfo *list ) {
    globus_result_t result;

    nest_debug( NEST_DEBUG_GRIDFTP, 
            "GridFtp %d sendDirectoryList: open data connection\n", id );

    // Create the wait object
    FtpWait wait( "DataConnect/SendDir", WaitDataConnect, id, this );

        // open a data connection
    result = globus_ftp_control_data_connect_write( 
        handle, dataConnectCallback, &wait );
    if( result == GLOBUS_SUCCESS ) {
        wait.WaitOnCallback( __FUNCTION__, __LINE__ );
    } else {
        nest_warn( __FILE__, __FUNCTION__,"gftp %d: sendDirList: data_connect error %s\n",
                id, globusResultToString( result ) );
        wait.CancelWait( __FUNCTION__, __LINE__ );
        sendErrorMessage( __FILE__, __FUNCTION__,  NEST_NO_CONNECTION );
        return false;
    }

        // send the ack, then the data
    if ( ! sendAck() ) return false;
    while ( list != NULL ) {
        NestSnprintf( buffer, FILEBUFFER, "%s\r\n", list->name );
        if ( Sendn( buffer, NestStrlen( buffer ) ) != 
             (int)NestStrlen( buffer ) ) {
            nest_warn( __FILE__, __FUNCTION__, "Could not send directory list in gftp %d.\n", id );
            sendErrorMessage( __FILE__, __FUNCTION__,  NEST_NO_CONNECTION );
            return false;
        }
        list = list->next;
    }

        // sendFinalAck also sends the eof
    return sendFinalAck( DIR_LIST_REQUEST );
}

bool 
nestProtocolGridFtp::sendErrorMessage( const char* filename, const char* func_name,  int error ){

        // check error type first, then use reqType
    if ( error == NEST_UNKNOWN_REQUEST_TYPE ) {
        NestSnprintf( buffer, FILEBUFFER, "500 : command not understood.\r\n" );
    } else if(error == NEST_INSUFFICIENT_SPACE ) {
        NestSnprintf( buffer, FILEBUFFER, "552 : Requested file action aborted. "
                                          "Exceeded storage allocation.\r\n" );
        //trying to forcefully close the data channel
        closeDataChannel();
    } else {
            // errors vary depending on the request type
        switch( reqType ) {
        case CONNECTION_REQUEST:
            nest_warn( filename, func_name, "Warning: sendErrorMessage called but sending ack\n" );
            NestSnprintf(   buffer,
                            FILEBUFFER, 
                            "230 User logged in, proceed.\r\n");
            break;
        case FILE_DELETE_REQUEST:
            NestSnprintf( buffer,
                          FILEBUFFER,
                          "550 DELE unsuccessful: %s.\r\n",
                          NestErrorString( error ) );
            break;
        case FILESIZE_REQUEST:
            NestSnprintf( buffer,
                          FILEBUFFER,
                          "550 File size not available: %s.\r\n",
                          NestErrorString( error ) );
            break;
        case CLIENT_RECV_FILE:
            NestSnprintf( buffer,
                          FILEBUFFER,
                          "550 Get file error: %s.\r\n",
                          NestErrorString( error ) );
            break;
        case CWD_REQUEST:
            NestSnprintf( buffer,
                          FILEBUFFER,
                          "550 CWD fail: %s.\r\n",
                          NestErrorString( error ) );
            break;
        case DIR_LIST_REQUEST:
            NestSnprintf( buffer,
                          FILEBUFFER,
                          "425 No connection.\r\n" );
            break;
        case CHANGE_USER_REQUEST:
            NestSnprintf( buffer,
                          FILEBUFFER,
                          "530 No local mapping.\r\n" );
            break;
        default:
            nest_warn( filename, func_name, "Warning: req %d using default err msg. Error: %s.\n",
                       reqType,
                       NestErrorString( error ) );

            NestSnprintf( buffer,
                          FILEBUFFER,
                          "550 command %d failure: %s.\r\n",
                          reqType,
                          NestErrorString( error ) );
            break;
        }
    }

    nest_debug( NEST_DEBUG_GRIDFTP, "Grid ftp %d sending error: %s\n", 
            id, buffer );
    if ( ! sendMessage( buffer ) ) {
        nest_warn( filename, func_name, "Globus %d send error failure: %s\n", id, buffer );
        return false;
    }
    
    return true;
}

bool 
nestProtocolGridFtp::sendTransferInfo(  nestRequest *req ){
        // no transfer info to send in gftp, this uses a seperate data 
        // channel so no need to send length info
    return sendAck();
}

bool 
nestProtocolGridFtp::sendFilesize( long size){
    NestSnprintf( buffer, FILEBUFFER, "213 %ld\r\n", size );
    return sendMessage( buffer );
}

bool
nestProtocolGridFtp::sendPwd( const char *pwd ) {
    assert( reqType == PWD_REQUEST );
    NestSnprintf( buffer,
                  FILEBUFFER,
                  "257 \"%s\" is current directory.\r\n",
                  pwd );
    return sendMessage( buffer );
}

bool
nestProtocolGridFtp::sendFinalAck( int transferType ){

    assert( transferType == this->reqType );

        // close the data connection
        // hopefully, this is done for us for CLIENT_SEND_FILE
    if ( transferType == CLIENT_RECV_FILE || transferType == DIR_LIST_REQUEST ){
        nest_debug(NEST_DEBUG_GRIDFTP,
                   "Gftp %d Sending eof on globus data channel.\n", id );
        // Create the wait object
        FtpWait wait( "ControlData/SendFinalAck", WaitControlData, id, this );
        globus_result_t result = globus_ftp_control_data_write( this->handle,
            (globus_byte_t *)buffer, 0, this->offset, GLOBUS_TRUE,
            controlDataCallback, &wait );
        if ( result == GLOBUS_SUCCESS ) {
            wait.WaitOnCallback( __FUNCTION__, __LINE__ );
        } else {
            nest_warn( __FILE__, __FUNCTION__, "GfTP connection %d data write err: %s\n",
                id, globusResultToString( result ) );
            wait.CancelWait( __FUNCTION__, __LINE__ );
            return false;
        }
    }

    NestSnprintf( buffer, FILEBUFFER, "226 Transfer complete.\r\n" );
    nest_debug(NEST_DEBUG_GRIDFTP,"Gftp %d Sending final ack: %s\n",id,buffer );
    if ( ! sendMessage( buffer ) ) {
        nest_warn( __FILE__, __FUNCTION__, "Globus send final ack failure: %s\n", buffer );
        return false;
    }
    return true;
}

// callback for globus_ftp_control_data_connect_write and
//              globus_ftp_control_data_connect_read
void 
nestProtocolGridFtp::dataConnectCallback(  void *callback_arg,
                        struct globus_ftp_control_handle_s *handle,
                        unsigned int stripe_ndx,
                        globus_bool_t reused,
                        globus_object_t *error)
{
    if( error != GLOBUS_SUCCESS ) {
        nest_warn( __FILE__, __FUNCTION__,">> dataConnectCallback: error %s\n",
                globus_object_printable_to_string( error ) );
        return;    
    } 

    FtpWait *wait = (FtpWait *) callback_arg;
    nestProtocolGridFtp *connection = wait->GetConnection( );
    nest_debug( NEST_DEBUG_GRIDFTP, 
            "gftp %d dataConnectCallback called - initialize transfer data\n",
            connection->id );
    connection->offset = 0;
    connection->eof = false;
    connection->bufsize = 0;
    connection->transmissionLength = 0;
    NestMemset( connection->buffer, 0, sizeof( connection->buffer ) ); 

    // Signal that we're done
    wait->CallbackDone( __FUNCTION__, __LINE__, WaitDataConnect );
}

void
nestProtocolGridFtp::closeCallback( void *arg,
        struct globus_ftp_control_handle_s *handle,
        globus_object_t *error,
        globus_ftp_control_response_t *response )
{
    if( error != GLOBUS_SUCCESS ) {
        nest_warn( __FILE__, __FUNCTION__,">> closecallback: error %s\n",
                globus_object_printable_to_string( error ) );
        return;    
    } 
    
    // Signal that we're done
    FtpWait *wait = (FtpWait *) arg;
    wait->CallbackDone( __FUNCTION__, __LINE__, WaitClose );
}
        
void 
nestProtocolGridFtp::commandCallback(   void * arg,
                    struct globus_ftp_control_handle_s * handle,
                    globus_object_t * error)
{
    nestProtocolGridFtp *connection = (nestProtocolGridFtp *)arg;
    assert( connection != NULL );

    if( error != GLOBUS_SUCCESS ) {
        nest_warn( __FILE__, __FUNCTION__,">> command_callback: error %s\n",
                globus_object_printable_to_string( error ) );
    } 

    // Signal that we're done
    FtpWait *wait = (FtpWait *) arg;
    wait->CallbackDone( __FUNCTION__, __LINE__, WaitCommand );
}


bool
nestProtocolGridFtp::sendAck(  ) {

        // acks vary depending on the request type
    switch( reqType ) {
    case CONNECTION_REQUEST:
        NestSnprintf( buffer,
                      FILEBUFFER,
                      "220 NeST FTP server ready.\r\n" );
        break;
    case CHANGE_USER_REQUEST:
        NestSnprintf( buffer,
                      FILEBUFFER,
                      "230 User logged on.\r\n" );
        break;
    case CLOSE_CONNECTION_REQUEST:
        NestSnprintf( buffer,
                      FILEBUFFER,
                      "221 Farewell.  Thanks for using NeST.\r\n");
        break;
    case FILE_DELETE_REQUEST:
        NestSnprintf( buffer,
                      FILEBUFFER,
                      "250 DELE command successful.\r\n" );
        break;
    case CLIENT_RECV_FILE:
    case CLIENT_SEND_FILE:
        NestSnprintf( buffer,
                      FILEBUFFER,
                      "150 NeST file status okay."
                      " Opening data connection.\r\n" );
        break;
    case DIR_LIST_REQUEST:
        NestSnprintf( buffer,
                      FILEBUFFER,
                      "150 NeST list status okay."
                      " Opening data connection.\r\n" );
        break;
    default:
        nest_debug( NEST_DEBUG_GRIDFTP,
                    "Warn: Grid ftp %d sending ack for unknown type: %d\n",
                    id,
                    reqType );
        NestSnprintf( buffer,
                      FILEBUFFER,
                      "220 NeST ack.  Command OK.\r\n" );
        break;
    }

    nest_debug( NEST_DEBUG_GRIDFTP,
                "Grid ftp %d sending ack: %s\n",
                id,
                buffer );

    if ( ! sendMessage( buffer ) ) {
        nest_warn( __FILE__, __FUNCTION__, "Globus send ack failure: %s\n", buffer );
        return false;
    }

    return true;
}

nestRequest*
nestProtocolGridFtp::receiveRequest(  fd_set *ready ){
        // ignore the fd_set
    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg,
                  NEST_MESSAGESIZE,
                  "Grid ftp %d receiveRequest called: ",
                  id );

    nestRequest *req = getRequest();
    if ( req != NULL ) {
        NestSnprintf(&(msg[NestStrlen(msg)]),NEST_MESSAGESIZE-NestStrlen(msg), 
            "Actually have a request to give.\n" );
    } else {
        NestSnprintf( &(msg[NestStrlen(msg)]), NEST_MESSAGESIZE-NestStrlen(msg),
                "No request to give\n" );
    }
    nest_debug( NEST_DEBUG_GRIDFTP, msg );
    return req; 
}

bool
nestProtocolGridFtp::sendMessage( const void *vptr ) {
    globus_result_t result;
    bool            rval;
    
    nest_debug( NEST_DEBUG_GRIDFTP, "Gftp %d sending ftp msg: %s\n", 
            id, (char *)vptr );

    // Create a new wait object (will probably not be top)
    FtpWait wait( "Command/SendMessage", WaitCommand, id, this );
    
    result = globus_ftp_control_send_response(
        handle, (char *)vptr, commandCallback, &wait );
    if ( result == GLOBUS_SUCCESS ) {
            // should we wait for the callback here?
            // we didn't use to, and when we sent messages, we would not
            // receive the callback for messages which were not handled
            // locally and the replies would get off by one . . . ?
        wait.WaitOnCallback( __FUNCTION__, __LINE__ );
        rval = true;
    } else {
        nest_warn( __FILE__, __FUNCTION__, "globus_ftp_control_send_response error: %s\n", 
                globusResultToString( result ) );
        NestFree( handle );
        wait.CancelWait( __FUNCTION__, __LINE__ );
        rval = false;
    }

    return rval;
}

char *
nestProtocolGridFtp::globusResultToString( globus_result_t result ) {
    return ( globus_object_printable_to_string( globus_error_get( result ) ) );
}

int
nestProtocolGridFtp::getAuthRequirements( ) {
    //return GLOBUS_FTP_CONTROL_AUTH_REQ_NONE;
    return  GLOBUS_FTP_CONTROL_AUTH_REQ_GSSAPI |
            GLOBUS_FTP_CONTROL_AUTH_REQ_USER |
            GLOBUS_FTP_CONTROL_AUTH_REQ_PASS;
}

// the listen callback is diffferent from all the other callbacks in that
// it is static and does not need to do the callbackDone thing when it is 
// done because no-one is waiting on it.
// when new clients show up, wait for the acceptCallback to return
// and if everything is OK, create a new client for the connection
void
nestProtocolGridFtp::listenCallback( void *args, 
        globus_ftp_control_server_t *server, globus_object_t *error )
{
    globus_result_t                         result;
    globus_ftp_control_handle_t             *handle;    
    static int                              id = 0;

    nest_debug( NEST_DEBUG_GRIDFTP, "listen callback called\n" );

    // check for error first
    if ( error != NULL ) {
        nest_warn( __FILE__, __FUNCTION__, "Globus listenCallback error: %s\n", 
                globus_object_printable_to_string( error ) );
        return;
    }

    handle = (globus_ftp_control_handle_t *) 
        NestMalloc( sizeof ( struct globus_ftp_control_handle_s ) );

    result = globus_ftp_control_handle_init( handle );
    if ( result != GLOBUS_SUCCESS ) {
        nest_warn( __FILE__, __FUNCTION__, "globus_ftp_control_handle_init error: %s\n", 
                globusResultToString( result ) );
        NestFree( handle );
        return;
    }
        
    // create the new connection object and try to accept it
    nestProtocolGridFtp *connection = new nestProtocolGridFtp( handle, ++id );

    // Create a new wait object & register an accept callback 
    nest_debug( NEST_DEBUG_GRIDFTP, "Registering accept\n" );
    FtpWait wait( "WaitAccept/Listen", WaitAccept, 
                  connection->id, connection );
    result = globus_ftp_control_server_accept( server, handle,
            acceptCallback, &wait );
    if ( result == GLOBUS_SUCCESS ) {
        // do we need to wait for the accept callback or not
        // I think not, but lets do it for the sake of being thorough
        // and if it works we can take it out later for the sake of performance
        // of course, if you do so, you must also remove its above creation
        wait.WaitOnCallback( __FUNCTION__, __LINE__ );
    } else {
        nest_warn( __FILE__, __FUNCTION__, "Error Could not complete a listen callback: %s\n", 
                globusResultToString( result ) );
        wait.CancelWait( __FUNCTION__, __LINE__ );
        connection->acceptResult = GLOBUS_FALSE;
    }

    if ( (connection->authResult == GLOBUS_TRUE)
            && (connection->acceptResult == GLOBUS_TRUE) ) 
    {
        nest_debug( NEST_DEBUG_GRIDFTP, "gftp %d: listen callback done\n",
                connection->id );

            // create a user request 
        nestRequest *req = new nestRequest( CHANGE_USER_REQUEST, 
                                            NULL, 
                                            connection->username, 
                                            NULL, 
                                            NEST_AUTH_GSS, 
                                            0 );
        connection->assignRequest( req );

            // only enqueue new connections, then signal nest
            // then register the readCommands callback for the handle
            // notice we don't wait on the readCommands because it will
            // not necessarily happen immediately and it will happen multiple X
            // TODO does the order of these operations matter ... ?
        if ( ! connection->isNewConnection() ) {
            nest_die( "Found an old connection in globus authCallback\n" );
        }
        enQueueConnection( connection );
        signalNest();    // do this in listenCallback after authCallback

            // this won't even work if the accept/auth fails
        nest_debug(NEST_DEBUG_GRIDFTP,"registering a read commands callback\n");
        result = globus_ftp_control_read_commands( 
                handle, readCommands, connection );
        if ( result != GLOBUS_SUCCESS ) {
            nest_warn( __FILE__, __FUNCTION__, 
                 "Could not register a globus read commands for gftp %d: %s\n", 
                 connection->id,
                 globusResultToString( result ) );
            delete( connection );
        }

    } else {
        nest_debug( NEST_DEBUG_GRIDFTP, 
                "gtfp %d: unable to authenticate, will delete object\n",
                connection->id );
        delete( connection );
    }

}

void
nestProtocolGridFtp::authCallback( 
        void *                                   args,
        globus_ftp_control_handle_t *            handle,
        globus_object_t *                        error,
        globus_ftp_control_auth_info_t *         auth_result)
{
    char *          username = NULL;
    
    nest_debug( NEST_DEBUG_GRIDFTP | NEST_DEBUG_AUTH, 
                "authCallback called\n" );
    FtpWait *wait = (FtpWait *) args;
    nestProtocolGridFtp *connection = wait->GetConnection( );
    connection->authResult = GLOBUS_TRUE;

        // check for error first
    if ( error != NULL ) {
        nest_warn( __FILE__, __FUNCTION__, "Globus authCallback Error: %s\n", 
                globus_object_printable_to_string( error ) );
        connection->authResult = GLOBUS_FALSE;
        NestSnprintf( connection->buffer,
                      FILEBUFFER,
                      "530 Globus auth error: %s\r\n",
                      globus_object_printable_to_string( error ) );
    } else {
            // mutual authentication OK, look for user in local mapping
            // TODO should not need strncmp, this is globus workaround
        if( ! strncmp(auth_result->user, ":globus-mapping:",16) ) {
            nest_debug( NEST_DEBUG_GRIDFTP | NEST_DEBUG_AUTH,
                    "user %s requesting globus mapping\n",
                    auth_result->auth_gssapi_subject );
            connection->globusname = NestStrdup( 
                        auth_result->auth_gssapi_subject );
            if ( ! globus_gss_assist_gridmap( auth_result->auth_gssapi_subject, 
                                              &username) ) 
            {
                connection->authResult = GLOBUS_TRUE;
            } else {
                connection->authResult = GLOBUS_FALSE;
                NestSnprintf( connection->buffer, FILEBUFFER, 
                    "530 No local mapping for Globus ID.\r\n" );
            }
        } else if(strcmp(auth_result->user, "ftp") == 0 ||
                  strcmp(auth_result->user, "anonymous") == 0)
        {
            const char *anonymous;
            connection->authResult = GLOBUS_TRUE;
            if ( GetConfigValue( Config_AnonymousUser, &anonymous ) ) {
                username = NestStrdup( anonymous );
            } else {
                username = NULL;
            }
        } else {
                // client asked for gsi authentication but did not pass
                // ":globus-mapping" or "ftp" or "anonymous" as name
            connection->authResult = GLOBUS_FALSE;
            NestSnprintf( connection->buffer, FILEBUFFER, 
                    "530 User unkown.\r\n" );
        }
    }

    if(connection->authResult == GLOBUS_TRUE) {
        connection->username = username;
        // TODO this is dangerous.  the auth_result pointer argument in this
        // function is created by globus.  don't know when it will disappear.
        // but globus has no copy method for it.
        // globus promises it shouldn't go away until connection is destroyed
        connection->delegated_credential_handle = 
            auth_result->delegated_credential_handle;
        
        /*
           // TODO if dcau is bad, should we do this?
            // don't do this at all anymore
           // don't do this here, just do it in response to the DCAU command
        globus_ftp_control_dcau_t dcau;
        dcau.mode = GLOBUS_FTP_CONTROL_DCAU_SELF;
        dcau.subject.mode = dcau.mode;
        dcau.subject.subject = NULL;
        globus_result_t result = globus_ftp_control_local_dcau( 
                handle, &dcau,
                connection->delegated_credential_handle );
        if ( result != GLOBUS_SUCCESS ) {
            nest_warn( __FILE__, __FUNCTION__,">> unable to proactively set dcau: error %s\n",
                        globusResultToString( result ) );
        }
        */
    } else {
        nest_debug( NEST_DEBUG_GRIDFTP | NEST_DEBUG_AUTH, 
                "Sending auth error reply: %s\n", 
                connection->buffer );
        connection->sendMessage( connection->buffer );
    }
    wait->CallbackDone( __FUNCTION__, __LINE__, WaitAuth );
}

void
nestProtocolGridFtp::acceptCallback( 
        void *args, 
        globus_ftp_control_handle_t *handle, 
        globus_object_t *error )
{
    globus_result_t                         result;
    globus_ftp_control_auth_requirements_t  authReqs;

    nest_debug( NEST_DEBUG_GRIDFTP, "acceptCallback called\n" );
    FtpWait *wait = (FtpWait *) args;
    nestProtocolGridFtp *connection = wait->GetConnection( );

    // check for error first
    if ( error != NULL ) {
        nest_warn( __FILE__, __FUNCTION__, "Error: %s\n",
                globus_object_printable_to_string( error ) );
        connection->acceptResult = GLOBUS_FALSE;
        wait->CallbackDone( __FUNCTION__, __LINE__, WaitAccept );
        return;
    } else {
        connection->acceptResult = GLOBUS_TRUE;
    }
    
    // TODO this probably should create a request object and enqueue
    // and signal here we just send a "220: NeST ready"
    // this would actually be difficult.
    // I think it is easier to only inform the server about the new
    // connection once all the auth code is done
    connection->reqType = CONNECTION_REQUEST;
    if ( ! connection->sendAck() ) {
        connection->acceptResult = GLOBUS_FALSE;
    }

    // set security requirements
    // TODO this is where it is not totally clean maybe
    // maybe only the server should know what the authReqs are and just
    // allow or disallow different types . . . 
    authReqs = getAuthRequirements( );

    if ( authReqs != GLOBUS_FTP_CONTROL_AUTH_REQ_NONE ) 
    {
        // try to authenticate and register an authentication callback
        nest_debug( NEST_DEBUG_GRIDFTP, "Calling globus_authenticate\n" );

        // Create our wait object
        FtpWait auth_wait( "Auth/Accept", WaitAuth, 
                           connection->id, connection );
        result = globus_ftp_control_server_authenticate( handle, 
                authReqs, authCallback, &auth_wait );
        if ( result == GLOBUS_SUCCESS ) {
            auth_wait.WaitOnCallback( __FUNCTION__, __LINE__ );
            nest_debug( NEST_DEBUG_GRIDFTP, "Globus authentication done\n" );
        } else {
            nest_warn( __FILE__, __FUNCTION__, "Auth gftp %d err: %s\n", 
                    connection->id, globusResultToString( result ) );
            auth_wait.CancelWait( __FUNCTION__, __LINE__ );
            connection->authResult = GLOBUS_FALSE;
        }
    } else {
        connection->authResult = GLOBUS_TRUE;
    }
    wait->CallbackDone( __FUNCTION__, __LINE__, WaitAccept );
}

/**
  this function reads a command and converts it into a generic nestRequest obj
**/
void
nestProtocolGridFtp::readCommands( 
        void *args, 
        globus_ftp_control_handle_t *handle,
        globus_object_t *error,
        globus_ftp_control_command_t *command ) 
{
    nestProtocolGridFtp             *connection             = NULL;
    nestRequest                     *request                = NULL;
    bool                            handleLocally           = false;
    char                            *message                = NULL;
    globus_result_t                 result;
    globus_ftp_control_host_port_t  connection_addr;
    int                             addr_part[6];
    int                             globus_request;
    //int                             count[1] = { 0 };
    FtpWait                         *wait;

    int type = -1;
    long start = 0, end = 0;
    char file[NEST_MESSAGESIZE];
    char user[NEST_MESSAGESIZE];
    memset( file, 0, NEST_MESSAGESIZE );
    memset( user, 0, NEST_MESSAGESIZE );

    connection = (nestProtocolGridFtp *) args;
    assert( connection != NULL );

        // check for error first
    if ( error != NULL ) {
        nest_warn( __FILE__, __FUNCTION__, "Error in readCommands, send error message: %s\n",
                globus_object_printable_to_string( error ) );
        connection->sendErrorMessage( __FILE__, __FUNCTION__,  NEST_UNKNOWN_REQUEST_TYPE );
            // do we need to re-register the read commands ?
        nest_debug( NEST_DEBUG_GRIDFTP, "re-initing a control handle\n" );
        result = globus_ftp_control_handle_init( connection->getHandle() );
        if ( result != GLOBUS_SUCCESS ) {
            nest_warn( __FILE__, __FUNCTION__, "Could not re-init a globus control handle: %s\n",
                    globusResultToString( result ) );
        }
        nest_debug( NEST_DEBUG_GRIDFTP,
                "re-registering a read commands callback\n" );
        result = globus_ftp_control_read_commands(
                connection->getHandle(), readCommands, args );
        if ( result != GLOBUS_SUCCESS ) {
            nest_warn( __FILE__, __FUNCTION__, "Could not register a globus read commands: %s\n",
                    globusResultToString( result ) );
        }
        return;
    } else {
        globus_request = command->base.code;
        nest_debug( NEST_DEBUG_GRIDFTP, "readCommands called: %d: %s\n",
            command->code, command->base.raw_command );
    }

    message = connection->buffer;

    switch( globus_request ) {
    case GLOBUS_FTP_CONTROL_COMMAND_AUTH:
        nest_warn( __FILE__, __FUNCTION__, "Auth request ignored . . . . already done(?)\n" );
        //return;
        // these should maybe be split up
    case GLOBUS_FTP_CONTROL_COMMAND_USER:
        if ( sscanf( command->base.raw_command, "%*s %s", user ) != 1 ) {
            nest_perror( "sscanf" );
            NestSnprintf( message,
                          NEST_MESSAGESIZE,
                          "550 User format error.\r\n" );
            handleLocally = true;
            break;
        }
        type = CONNECTION_REQUEST;
        break;
    case GLOBUS_FTP_CONTROL_COMMAND_PASS:
        NestSnprintf( message, NEST_MESSAGESIZE,
                "230 Promiscuous NeST requires no password.\r\n" );
        handleLocally = true;
        break;
    case GLOBUS_FTP_CONTROL_COMMAND_ADAT:
        NestSnprintf( message, NEST_MESSAGESIZE,
                "230 Proceed.  ADAT ignored. What the heck is ADAT?\r\n" );
        handleLocally = true;
        break;
    case GLOBUS_FTP_CONTROL_COMMAND_PWD:
        type = PWD_REQUEST;
        break;
    case GLOBUS_FTP_CONTROL_COMMAND_MKD:
    case GLOBUS_FTP_CONTROL_COMMAND_CWD:
    case GLOBUS_FTP_CONTROL_COMMAND_RMD:
        switch( globus_request ) {
        case GLOBUS_FTP_CONTROL_COMMAND_MKD:    type = MKDIR_REQUEST;  break;
        case GLOBUS_FTP_CONTROL_COMMAND_CWD:    type = CWD_REQUEST;    break;
        case GLOBUS_FTP_CONTROL_COMMAND_RMD:    type = RMDIR_REQUEST;  break;
        default:                                assert( 0 );
        }

        if ( sscanf( command->base.raw_command, "%*s %s", file ) != 1 ) {
            nest_perror( "sscanf" );
            NestSnprintf( message,
                          NEST_MESSAGESIZE,
                          "550 User format error.\r\n" );
            handleLocally = true;
            break;
        }
        break;
    case GLOBUS_FTP_CONTROL_COMMAND_CDUP:
        strcpy( file, ".." );
        type = CWD_REQUEST;
        break;
    case GLOBUS_FTP_CONTROL_COMMAND_LIST:
    case GLOBUS_FTP_CONTROL_COMMAND_NLST:
        type = DIR_LIST_REQUEST;
        break;
    case GLOBUS_FTP_CONTROL_COMMAND_HELP:
        NestSnprintf( message, NEST_MESSAGESIZE,
            "214-The following commands are recognized "
            "(* =>'s unimplemented).\r\n"
            "USER    PASV    MLFL*   ALLO    XCWD    NOOP    CDUP    PBSZ\r\n"
            "PASS    TYPE    MAIL*   REST    LIST    MKD     XCUP    CCC \r\n"
            "ACCT*   STRU    MSND*   RNFR    NLST    XMKD    STOU    MDTM\r\n"
            "SMNT*   MODE    MSOM*   RNTO    SITE    RMD     SIZE    ESTO\r\n"
            "REIN*   RETR    MSAM*   ABOR    SYST    XRMD    AUTH    ERET\r\n"
            "QUIT    STOR    MRSQ*   DELE    STAT    PWD     ADAT    FEAT\r\n"
            "PORT    APPE    MRCP*   CWD     HELP    XPWD    PROT    OPTS\r\n"
            "214 Direct comments to support@nestproject.org\r\n" );
        handleLocally = true;
        break;

    case GLOBUS_FTP_CONTROL_COMMAND_ABOR:
            // user is requesting that we abort a transfer
            // turn off wait to pop the last receive/send
            // out of waitCallback
        handleLocally = true;

        result = connection->closeDataChannel();
        if ( result == GLOBUS_SUCCESS ) {
            NestSnprintf( message,
                      NEST_MESSAGESIZE,
                      "225 ABOR command successful.\r\n" );
        } else {
            NestSnprintf( message, NEST_MESSAGESIZE, 
                    "450 Unable to ABOR\r\n" );
        }
        break;

    case GLOBUS_FTP_CONTROL_COMMAND_NOOP:
        NestSnprintf( message,
                      NEST_MESSAGESIZE,
                      "200 NOOP command successful\r\n" );
        handleLocally = true;
        break;
    case GLOBUS_FTP_CONTROL_COMMAND_TYPE:
        char t_type;
        globus_ftp_control_type_t gtype;
        if ( sscanf( command->base.raw_command, "%*s %c", &t_type ) != 1 ) {
            nest_perror( "sscanf" );
            NestSnprintf( message,
                          NEST_MESSAGESIZE,
                          "550 User format error.\r\n" );
            handleLocally = true;
            break;
        }

        gtype = (globus_ftp_control_type_t) t_type;
        result = globus_ftp_control_local_type(handle, gtype, 0 );
        if( result != GLOBUS_SUCCESS ) {
            nest_warn( __FILE__, __FUNCTION__,">> command_process (TYPE): error %s\n",
                    globusResultToString( result ) );
            NestSnprintf( message, NEST_MESSAGESIZE,
                    "450 Unable to change to type %c\r\n", t_type );
            handleLocally = true;
            break;
        }

        NestSnprintf( message, NEST_MESSAGESIZE,
                "200 Type Command okay: set to %c.\r\n", t_type );
        handleLocally = true;
        break;
    case GLOBUS_FTP_CONTROL_COMMAND_MODE:
        char m_type;
        if ( sscanf( command->base.raw_command, "%*s %c", &m_type ) != 1 ) {
            nest_perror( "sscanf" );
            NestSnprintf( message,
                          NEST_MESSAGESIZE,
                          "550 User format error.\r\n" );
            handleLocally = true;
            break;
        }

        result = globus_ftp_control_local_mode(handle, 
                    (globus_ftp_control_mode_t)m_type );
        if( result != GLOBUS_SUCCESS ) {
            nest_warn( __FILE__, __FUNCTION__,">> command_process (MODE): error %s\n",
                    globusResultToString( result ) );
            NestSnprintf( message, NEST_MESSAGESIZE,
                    "450 Unable to change to mode %c\r\n", m_type );
            handleLocally = true;
            break;
        }

        NestSnprintf( message, NEST_MESSAGESIZE,
                "200 Mode Command okay: set to %c.\r\n", m_type );
        handleLocally = true;
        break;
    case GLOBUS_FTP_CONTROL_COMMAND_SIZE:
        if ( sscanf( command->base.raw_command, "%*s %s", file ) != 1 ) {
            nest_perror( "sscanf" );
            NestSnprintf( message,
                          NEST_MESSAGESIZE,
                          "550 User format error.\r\n" );
            handleLocally = true;
            break;
        }
        assert( ! handleLocally );
        type = FILESIZE_REQUEST;
        break;
    case GLOBUS_FTP_CONTROL_COMMAND_DELE:
        if ( sscanf( command->base.raw_command, "%*s %s", file ) != 1 ) {
            nest_perror( "sscanf" );
            NestSnprintf( message,
                          NEST_MESSAGESIZE,
                          "550 User format error.\r\n" );
            handleLocally = true;
            break;
        }
        type = FILE_DELETE_REQUEST;
        break;
    case GLOBUS_FTP_CONTROL_COMMAND_SYST:
        NestSnprintf( message,
                      NEST_MESSAGESIZE,
                      "215 NeST Storage Appliance\r\n" );
        handleLocally = true;
        break;
    case GLOBUS_FTP_CONTROL_COMMAND_PORT:
        if ( 6 != sscanf(command->base.raw_command,"%*s %d,%d,%d,%d,%d,%d",
            &(addr_part[0]), &(addr_part[1]), &(addr_part[2]),
            &(addr_part[3]), &(addr_part[4]), &(addr_part[5]) ) )
        {
            nest_perror( "sscanf" );
            NestSnprintf( message,
                          NEST_MESSAGESIZE,
                          "450 Unable to do port\r\n" );
            handleLocally = true;
            break;
        }
        connection_addr.host[0]=addr_part[0];
        connection_addr.host[1]=addr_part[1];
        connection_addr.host[2]=addr_part[2];
        connection_addr.host[3]=addr_part[3];
        connection_addr.port=(addr_part[4]<<8)|(addr_part[5]);

        result = globus_ftp_control_local_port(handle, &connection_addr);
        if( result != GLOBUS_SUCCESS ) {
            nest_warn( __FILE__, __FUNCTION__,">> command_process (PORT): error %s\n",
                    globusResultToString( result ) );
            NestSnprintf( message,
                          NEST_MESSAGESIZE,
                          "450 Unable to do port\r\n" );
            handleLocally = true;
            break;
        }

        handleLocally = true;
        NestSnprintf( message,
                      NEST_MESSAGESIZE,
                      "200 Command okay.\r\n" );
        break;
    case GLOBUS_FTP_CONTROL_COMMAND_PASV:
            // clear any values from old address structure
        for( int i = 0; i < 4; i++ ) connection_addr.host[i] = 0;
        connection_addr.port = 0;

            // now attempt to set the passive address
        result = globus_ftp_control_local_pasv(handle, &connection_addr);
        if( result != GLOBUS_SUCCESS ) {
            /* TODO:  why does this sometimes fail?
             * >> command_process (PASV): error globus_ftp_control_local_pasv():
             * Handle not in the proper state 5.
             */
            nest_warn( __FILE__, __FUNCTION__,">> command_process (PASV): error %s\n",
                    globusResultToString( result ) );
            NestSnprintf( message,
                          NEST_MESSAGESIZE,
                          "450 Unable to go passive\r\n" );
            handleLocally = true;
            break;
        }
        NestSnprintf( message, NEST_MESSAGESIZE,
            "227 Entering Passive Mode (%d,%d,%d,%d,%d,%d).\r\n",
            connection_addr.host[0], connection_addr.host[1],
            connection_addr.host[2], connection_addr.host[3],
            ((unsigned char *)&connection_addr.port)[1],
            ((unsigned char *)&connection_addr.port)[0]);
        handleLocally = true;
        break;
    case GLOBUS_FTP_CONTROL_COMMAND_STOR:
            // read the file name in
        if ( sscanf( command->base.raw_command, "%*s %s", file ) != 1 ) {
            nest_perror( "sscanf" );
            NestSnprintf( message,
                          NEST_MESSAGESIZE,
                          "550 User format error.\r\n" );
            handleLocally = true;
            break;
        }

        start   = 0;
        end     = ENTIRE_FILE;
        type    = CLIENT_SEND_FILE;

            // Create the wait object
        wait = new FtpWait( "CommandStor", WaitDataConnect,
                            connection->id, connection );
        result = globus_ftp_control_data_connect_read(
            handle, dataConnectCallback, wait );
        if( result != GLOBUS_SUCCESS ) {
            nest_warn( __FILE__, __FUNCTION__,">> command_process: data_connect error %s\n",
                    globusResultToString( result ) );
            NestSnprintf( message,
                          NEST_MESSAGESIZE,
                          "550 Unable to create data connection.\r\n" );
            handleLocally = true;
            wait->CancelWait( __FUNCTION__, __LINE__ );
        } else {
            wait->WaitOnCallback( __FUNCTION__, __LINE__ );
        }
        delete( wait );
        break;
    case GLOBUS_FTP_CONTROL_COMMAND_RETR:
            // read the file name in
        if ( sscanf( command->base.raw_command, "%*s %s", file ) != 1 ) {
            nest_perror( "sscanf" );
            NestSnprintf( message,
                          NEST_MESSAGESIZE,
                          "550 User format error.\r\n" );
            handleLocally = true;
            break;
        }

        start   = 0;
        end     = ENTIRE_FILE;
        type    = CLIENT_RECV_FILE;

            // Create the wait object
        wait = new FtpWait( "CommandRetr", WaitDataConnect,
                            connection->id, connection );
        result = globus_ftp_control_data_connect_write(
            handle, dataConnectCallback, wait );
        if( result != GLOBUS_SUCCESS ) {
            nest_warn( __FILE__, __FUNCTION__,">> command_process: data_connect error %s\n",
                    globusResultToString( result ) );
            NestSnprintf( message,
                          NEST_MESSAGESIZE,
                          "425 Unable to create data connection.\r\n" );
            handleLocally = true;
            wait->CancelWait( __FUNCTION__, __LINE__ );
        } else {
            wait->WaitOnCallback( __FUNCTION__, __LINE__ );
        }
        delete( wait );
        break;

    case GLOBUS_FTP_CONTROL_COMMAND_QUIT:
        type = CLOSE_CONNECTION_REQUEST;
        break;
    // TODO: these need to be added to the globus code so we don't have to
    // parse them ourselves
    //case GLOBUS_FTP_CONTROL_COMMAND_PROT:
    //case GLOBUS_FTP_CONTROL_COMMAND_DCAU:
    //case GLOBUS_FTP_CONTROL_COMMAND_PBSZ:
    //case GLOBUS_FTP_CONTROL_COMMAND_CLNT:
    case GLOBUS_FTP_CONTROL_COMMAND_UNKNOWN:

			// PBSZ
        if ( strncasecmp( command->base.raw_command, "PBSZ", 4 ) == 0 ) {
            unsigned long buf_size;
            handleLocally = true;
            if (sscanf( command->base.raw_command, "%*s %lu", &buf_size ) != 1){
                nest_perror( "sscanf" );
                NestSnprintf( message,
                              NEST_MESSAGESIZE,
                              "550 User format error.\r\n" );
                break;
            }
            result = globus_ftp_control_local_pbsz( handle, buf_size );
            if( result != GLOBUS_SUCCESS ) {
                nest_warn( __FILE__, __FUNCTION__,">> command_process (PBSZ): error %s\n",
                        globusResultToString( result ) );
                NestSnprintf( message,
                              NEST_MESSAGESIZE,
                              "450 Unable to set buffer size %lu.\r\n",
                              buf_size );
            } else {
                NestSnprintf( message, NEST_MESSAGESIZE,
                    "200 PBSZ=%lu.\r\n", buf_size );
            }

            break;

			// PROT
        } else if ( strncasecmp( command->base.raw_command, "PROT", 4 ) == 0 ) {
            handleLocally = true;
            char prot_mode;
            globus_ftp_control_protection_t protection;
            if ( sscanf( command->base.raw_command, "%*s %c", &prot_mode )
                    != 1 ) {
                nest_perror( "sscanf" );
                NestSnprintf( message,
                              NEST_MESSAGESIZE,
                              "550 User format error.\r\n" );
                break;
            }
            protection = (globus_ftp_control_protection_t) prot_mode;
            // TODO: I think this is a globus bug.  local_prot fails
            // if dcau is set to NONE even if protection is set to CLEAR
            if ( protection != GLOBUS_FTP_CONTROL_PROTECTION_CLEAR ) {
                result = globus_ftp_control_local_prot( handle, protection );
            } else {
                result = GLOBUS_SUCCESS;
            }
            if( result != GLOBUS_SUCCESS ) {
                nest_warn( __FILE__, __FUNCTION__,">> command_process (PROT): error %s\n",
                        globusResultToString( result ) );
                NestSnprintf( message,
                              NEST_MESSAGESIZE,
                              "450 Unable to set prot mode %c.\r\n",
                              prot_mode );
            } else {
                NestSnprintf( message,
                              NEST_MESSAGESIZE,
                              "200 PROT=%c.\r\n",
                              prot_mode );
            }
            break;

			// DCAU
        } else if ( strncasecmp( command->base.raw_command, "DCAU", 4 ) == 0 ) {
            handleLocally = true;
            char dcau_mode_c;
            globus_ftp_control_dcau_t dcau;
            dcau.subject.subject = NULL;
            if ( sscanf( command->base.raw_command, "%*s %c", &dcau_mode_c )
                    != 1 ) {
                nest_perror( "sscanf" );
                NestSnprintf( message,
                              NEST_MESSAGESIZE,
                              "550 User format error.\r\n" );
                break;
            }
            dcau.mode = (globus_ftp_control_dcau_mode_t) dcau_mode_c;
            dcau.subject.mode = dcau.mode;
            if ( dcau.mode == GLOBUS_FTP_CONTROL_DCAU_SUBJECT ) {
                // get third token
                char *sub_name = strchr( command->base.raw_command, ' ' );
                sub_name = strchr( sub_name, ' ' ) + 1;
                dcau.subject.subject = NestStrdup( sub_name );
            }
               // I'm trying to fix the bad dcau mode, but it seems to make
               // globus-url-copy put work but get it makes send twice as
               // much data as it should
               // probably just better off sending a get set DCAU to 'T' message
            if (   dcau.subject.mode != GLOBUS_FTP_CONTROL_DCAU_SELF
                && dcau.subject.mode != GLOBUS_FTP_CONTROL_DCAU_SUBJECT 
                && dcau.subject.mode != GLOBUS_FTP_CONTROL_DCAU_NONE ) 
            {
                nest_warn( __FILE__, __FUNCTION__, "Passed a bad DCAU mode: %c.  Setting to \'S\'.\n",
                        dcau_mode_c );
                dcau.subject.mode = GLOBUS_FTP_CONTROL_DCAU_SELF;
            }
            result = globus_ftp_control_local_dcau( handle, &dcau,
                    connection->delegated_credential_handle );
            if( result != GLOBUS_SUCCESS ) {
                nest_warn( __FILE__, __FUNCTION__,">> command_process (DCAU): error %s\n",
                        globusResultToString( result ) );
                NestSnprintf( message,
                              NEST_MESSAGESIZE,
                              "432 Unable to set dcau mode %c.\r\n",
                              dcau.mode );
            } else {
                NestSnprintf( message,
                              NEST_MESSAGESIZE,
                              "200 DCAU=%c.\r\n",
                              dcau.mode );
            }
            break;

			// CLNT
        } else if ( strncasecmp( command->base.raw_command, "CLNT", 4 ) == 0 ) {
            handleLocally = true;
            char *client = &command->base.raw_command[5];
            nest_debug( NEST_DEBUG_CLIENT, "New ftp client: %s\n", client );
            NestSnprintf( message,
                          NEST_MESSAGESIZE,
                          "500 CLNT %s not understood.\r\n",
                          client );
            break;

        } 

        // if it is UNKNOWN and we don't handle it ourselves, fall into
        // default
    default:
        NestSnprintf( message,
                      NEST_MESSAGESIZE,
                      "502 Command not implemented.\r\n");
        handleLocally = true;
        nest_debug( NEST_DEBUG_GRIDFTP,
                    "Not handling a gftp command: %s\n",
                    command->base.raw_command );
        break;
    }

    if ( type > 0 ) {
        nest_debug( NEST_DEBUG_GRIDFTP,
                    "Globus read commands creating request of type %d\n",
                    type );
        request = new nestRequest( type, file, user, NULL, start, end );
        connection->assignRequest( request );
        if ( connection->isNewConnection() ) {
            nest_die( "Found a new connection in globus readCommands\n" );
            enQueueConnection( connection );
        }
        signalNest();
    } else {
        assert( handleLocally );
        
            // send message will do the debug message for us
        //nest_debug( NEST_DEBUG_GRIDFTP,
        //        "Globus read commands handling locally: %s\n", message);
        connection->sendMessage( message );
    }
}

globus_result_t
nestProtocolGridFtp::closeDataChannel( )
{

    FtpWait *wait;
    globus_result_t result;

    // Create our wait object
    wait = new FtpWait( "CommandAbr", WaitCommand,
                        this->id, this );

    // TODO - if a transfer was in progress, I need to send an eof
    result = globus_ftp_control_data_force_close(
            handle, commandCallback, wait );
    if( result != GLOBUS_SUCCESS ) {
        nest_warn( __FILE__, __FUNCTION__,">> command_process (ABOR): error %s\n",
                globusResultToString( result ) );
        wait->CancelWait( __FUNCTION__, __LINE__ );
    } else {
        wait->WaitOnCallback( __FUNCTION__, __LINE__ );
    }
    delete( wait );

    /*
    result = globus_ftp_control_data_send_eof( handle,
                count,
                sizeof( count ),
                GLOBUS_TRUE,
                commandCallback,
                NULL );
    if ( result != GLOBUS_SUCCESS ) {
        nest_warn( __FILE__, __FUNCTION__,">> command_process (ABOR): error %s\n",
                globusResultToString( result ) );
        NestSnprintf( message, NEST_MESSAGESIZE,
                "450 Unable to abort.\r\n" );
    }
    */

    return result;

}
