/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include <ctype.h>

#include "nest_protocol_http.h"
#include "nest_advertiser.h"

// static class variables
// why can't these be in the header file ???
static unsigned addr_sz;

NestReplyStatus
nestProtocolHttp::HttpStatusToNestStatus( int status ) {
    switch( status ) {
    case 200:
        return NEST_SUCCESS;
    case 404:
        return NEST_REMOTE_FILE_NOT_FOUND;
    default:
        return NEST_UNKNOWN_ERROR;
    }
}

    
int 
nestProtocolHttp::listen( int port ) {
    addr_sz = sizeof( struct sockaddr_in );
    return NestsetSocket( AF_INET, SOCK_STREAM, 0, port );
}

nestProtocol *
nestProtocolHttp::accept( int socket, sockaddr_in *addr ) {
    int csd =  NestAccept( socket, (sockaddr *)addr, &addr_sz ); 
    if ( csd <= 0 ) {
        return NULL;
    } else {
        nest_debug( NEST_DEBUG_HTTP, "New http connection: %d\n", csd );
        return new nestProtocolHttp( csd );
    }
}

NestReplyStatus
nestProtocolHttp::initiateThirdParty( nestClient *client,
        nestProtocol **connection ) 
{
    int sd = 0;
    NestReplyStatus status = NEST_SUCCESS;
    sockaddr_in sin;
    hostent *hostp = NULL;
    static char buffer[4096];

    nest_debug( NEST_DEBUG_HTTP, "HTTP init 3pt\n" );

        // make sure it is a get, should we do PUT?
        // and check other parameters
    if ( client->getRequest()->getType() != THIRD_PARTY_GET ) {
        return NEST_NOT_YET_IMPLEMENTED;
    }
    if ( client->getRequest()->getStart() != 0 
            || client->getRequest()->getEnd() != ENTIRE_FILE ) 
    {
        return NEST_BAD_PARAMETERS;
    }

        // open connection
    if (( hostp = NestGethostbyname(client->getRequest()->getRemoteHost()))
            == NULL)
    {
        return NEST_NO_CONNECTION;
    }
    memset( &sin, 0, sizeof( sockaddr_in ) );
    sin.sin_family = AF_INET;
    sin.sin_port = htons( 80 );
    memcpy( &sin.sin_addr, hostp->h_addr, hostp->h_length );
    if ( ( sd = NestSocket( AF_INET, SOCK_STREAM, 0 ) ) < 0 ) {
        status = NEST_NO_CONNECTION;
        goto DONE_THIRD_PARTY_INIT;
    }
    if ( NestConnect( sd, (sockaddr *)&sin, sizeof( sockaddr_in ) ) < 0 ) {
        status = NEST_NO_CONNECTION;
        goto DONE_THIRD_PARTY_INIT;
    }

        // create the instantiation of the protocol connection
    *connection = new nestProtocolHttp( sd );

        // attempt to move connection into transfer-ready state
        // send Connection: Keep-Alive for persistent connections
    NestSnprintf( buffer, 4096, "GET /%s HTTP/1.0\n\n", 
            client->getRequest()->getRemoteFile() );
    if ((**connection).Sendn( buffer, NestStrlen(buffer) ) 
            != (int)NestStrlen(buffer) )
    {
        status = NEST_NO_CONNECTION;
        goto DONE_THIRD_PARTY_INIT;
    }

        // now read until a blank line and parse the server reply
        // if we use fgets, it buffers and then that data is lost . . .
        // we should probably change it to use FILE * instead of descriptor
        // but then how will select work in NOB ???
        // what we really should do is do our own buffering in the client
        // buffer, then use anything leftover in the buffer for the transfer
    /*
    FILE *fp = fdopen( sd, "r" );
    if ( fp == NULL ) {
        nest_perror( "fopen" );
        return NEST_UNKNOWN_ERROR;
    }
    */
    do {
            // get next line
        // if ( fgets( buffer, sizeof( buffer ), fp ) == NULL ) {
        if ( NestFdGets( buffer, sizeof( buffer ), sd ) == NULL ) {
            return NEST_NO_CONNECTION;
        }
            // need to parse, check status and maybe length
        nest_debug( NEST_DEBUG_HTTP, "HTTP init 3pt read %s", buffer );
            // for time being, suffiencent to merely read first two tokens
        char tok1[1024], tok2[1024];
        if ( sscanf( buffer, "%s %s", tok1, tok2 ) == 2 ) {
            if ( strncmp( tok1, "HTTP", 4 ) == 0 ) {
                status = HttpStatusToNestStatus( atoi( tok2 ) ); 
                nest_debug( NEST_DEBUG_HTTP, "HTTP status is %s\n", 
                        NestErrorString(status) );
            } else if ( strcmp( tok1, "Content-Length:" ) == 0 ) {
                long length = atol( tok2 );
                nest_debug( NEST_DEBUG_HTTP, "HTTP transfer length is %ld\n", 
                        length );
                client->getRequest()->setEnd( length );
            }
        }
        
    } while( ( buffer[0] != '\r' && buffer[0] != '\n' ) );
    //fclose( fp ); // screws up if this is closed here

    DONE_THIRD_PARTY_INIT:
    if ( hostp != NULL ) NestFree( hostp );
    return status;
}


nestProtocolHttp::nestProtocolHttp( int socket ) {
    this->socket = socket;
        // TODO - parse HTTP better and set keep_alive correctly
    this->keep_alive  = false;
    this->still_alive = true;
    this->already_closed = false;
    memset( message, 0, HTTPMESSAGESIZE );
}

nestProtocolHttp::~nestProtocolHttp( ) {
    NestClose( socket );
}

nestProtocol *
nestProtocolHttp::clone() {
    return new nestProtocolHttp( NestDup( this->socket ) );
}

int
nestProtocolHttp::getSocket() {
    //return ( ( still_alive ) ? socket : -1 );
    return socket;
}

bool
nestProtocolHttp::hasSocket() {
    return true;
    //return still_alive;
}

bool
nestProtocolHttp::closeNonpersistent( ) {
    if ( ! keep_alive && still_alive ) {
        nest_debug( NEST_DEBUG_HTTP, "Closing non-persistent http client\n" );
        still_alive = false;
        NestShutdown( socket, 2 );
        return true;
    } else {
        return false;
    }
}

bool
nestProtocolHttp::persistent( ) {
    assert( ( still_alive ) || ( !keep_alive && !still_alive ) );
    return still_alive; 
}

bool
nestProtocolHttp::endThirdPartyTransfer( int type ) {
        // nothing to do for HTTP
        // what about PUTs?
    return true;
}

ssize_t
nestProtocolHttp::Send( const char *buffer, ssize_t n ) {
    return NestWrite( socket, buffer, n );
}

ssize_t
nestProtocolHttp::Receive( char *buffer, ssize_t n ) {
    return NestRead( socket, buffer, n );
}

bool
nestProtocolHttp::sendAd( nestAd *ad ) {

    int  hdrsize = 1024;     // also includes cushion for adding breaks
    int  adsize  = ad->length(); 
    int  msgsize = hdrsize + adsize; 
    char *msg = (char *)NestCalloc( msgsize ); 

    nest_debug( NEST_DEBUG_HTTP, "Sending BW stats to http client\n" );

        // write header into msg
    NestSnprintf( msg, msgsize, 
            "<html><title>NeST monitor page</title>\n"
            "<center><h1>NeST monitor page</h1></center>\n"
            "<hr><body>Hello! Welcome to "
            "<font color=\"#FF0000\"><h3>NeST</h3></font>\n" 
            "<hr><br>\n"
            );
    hdrsize = NestStrlen( msg );
        
        // read ad into msg
    char *line = strtok( ad->toString(), "\n" );
    while ( line ) {
        NestSnprintf( &msg[NestStrlen(msg)], msgsize - NestStrlen( msg ),
                "%s<br>\n", line );
        line = strtok( NULL, "\n" );
    }
    /*
    NestSnprintf( &msg[NestStrlen(msg)], msgsize - NestStrlen(msg),
            "%s\n", ad->toString() );
    */
    
        // then write the trailing html
    if ( NestSnprintf( &(msg[NestStrlen(msg)]), ( msgsize - NestStrlen(msg) ), 
            "<hr>"
            "<a href=\"http://www.ioappliance.com\">NeST homepage</a>, "
            "<a href=\"mailto:johnbent@cs.wisc.edu\">email</a> johnbent."
            "</body></html>" 
            "\n" ) < 0 ) 
    {
        nest_warn( __FILE__, __FUNCTION__, "Couldn't fit ad into allocated space in http.\n" );
    }
    sendMessage( msg, 200 );
    NestFree( msg );
        
    closeNonpersistent( );
    return true;
}

bool
nestProtocolHttp::sendFilesize( long size ) {
    assert( 0 );
    return false;
}

bool
nestProtocolHttp::sendErrorMessage( const char *filename, const char* func_name, int error ) {
    char msg[NEST_MESSAGESIZE];
    int errorcode = 0;

    nest_debug( NEST_DEBUG_HTTP, "http: sendError %s\n",NestErrorString(error));

    if ( error == NEST_REMOTE_FILE_NOT_FOUND ) {
        NestSnprintf( msg, NEST_MESSAGESIZE,
            "<HTML>"
            "<HEAD><TITLE>404 Not Found</TITLE></HEAD>"
            "<BODY>Sorry the url you requested was not found.</BODY>"
            "</HTML>\n" );
        errorcode = 404;
    } else {
        NestSnprintf( msg, NEST_MESSAGESIZE,
            "<HTML>"
            "<HEAD><TITLE>400 Unknown error</TITLE></HEAD>"
            "<BODY>NeST error: %s</BODY>"
            "</HTML>\n", NestErrorString( error ) );
        errorcode = 400;
    }
    sendMessage( msg, errorcode );

    closeNonpersistent( );
    return true;
}

bool
nestProtocolHttp::sendAck(  ) {
    return true;
}

bool
nestProtocolHttp::sendFinalAck( int transferType ) {
        // this is called when transfer is done
    closeNonpersistent();
    return true;
}

bool
nestProtocolHttp::sendTransferInfo( nestRequest *req ) {
    return sendHeader( req->getEnd(), 200 );
}

bool
nestProtocolHttp::fullMessage( char *msg ) {
    if ( NestStrlen( msg ) == HTTPMESSAGESIZE ) {
        nest_warn( __FILE__, __FUNCTION__, "HTTPMESSAGESIZE is of insuffiecient length\n" );
        return true;
    }
    if( NestStrlen( msg ) > 4 && ! 
            strcmp( &msg[NestStrlen(msg) - 4 ], "\r\n\r\n" ) )
        return true;
    /*
    if( NestStrlen( msg ) > 2 && ! strcmp( &msg[NestStrlen(msg) - 2 ], "\n\n"))
        return true;
    */
    return false;
}

nestRequest* 
nestProtocolHttp::receiveRequest(  fd_set *ready ){

    char *file = NULL;
    int nread = 0;
    
    if ( ! persistent() ) {
            // client was non-persistent and has already been serviced and
            // connection has already been closed
            // this only works in NOB and POT which are able to set stillAlive
            // in POP, we have to come back and read the 0 
            // there may be a race condition, where we return this close
            // connection request before the taskDone call completes
            // in which case the client object may get deleted and we'll 
            // reference it in taskDone
        return closeConnectionRequest();
    } else {
        if ( ! NestFd_isset( socket, ready ) ) return NULL;
    }

        // socket is set, read should return quickly
    /*
    if ( NestStrlen( message ) != 0 ) {
        nest_debug( NEST_DEBUG_HTTP, "Resuming partial http request read\n" );
    }
    */
    nread = NestRead( socket, &message[NestStrlen(message)], 
            HTTPMESSAGESIZE - NestStrlen(message) );

        // check whether connection was lost
    if ( nread <= 0 ) {
            // if 0, means other side closed
        if ( nread < 0 && ( errno != ECONNRESET && errno != EPIPE ) )
                nest_perror( "http read" );
        return closeConnectionRequest();
    }

        // if we've read a full request, parse it, otherwise wait till later
    if( fullMessage( message ) ) {
            // we have read a full http request, now we can parse it
        nest_debug( NEST_DEBUG_HTTP, "HTTP recv'd %d byte msg: %.53s\n", 
                NestStrlen(message), message );
        char *token = strtok( message, "\n\t " );
        while( token ) {
            if ( ! strcmp( token, "GET" ) ) {
                token = strtok( NULL, "\n\t " );
                file = NestStrdup( token );
                break;
            }
            token = strtok( NULL, "\n\t " );
        }
            // reset the message so we can use it again
        memset( message, 0, HTTPMESSAGESIZE );
    } else {
        nest_debug( NEST_DEBUG_HTTP, 
                "Received only a partial http req (len %d): %s.\n",
                NestStrlen( message ), message );
        return NULL;
    }
    
        // if we make it here, we've received a full message
    if ( file == NULL ) {
        nest_warn( __FILE__, __FUNCTION__, "Bad http req recvd.\n" );
        return new nestRequest( BAD_REQUEST, 0, 0, 0, NEST_BAD_PARAMETERS, 0 );
    } else {
            // skip past any http://server/ stuff
        char *filename = file;
        if ( strncmp( file, "http://", 7 ) == 0 ) {
                // skip http://
            filename = &(file[7]);
                // skip servername
            filename = strchr( filename, '/' );
        }
        filename++;     // skip the slash

        if ( NestStrlen( filename ) > 0 ) {
            char nest_file[NEST_MESSAGESIZE];

	    nestRequest *req = new nestRequest( 
                    CLIENT_RECV_FILE,
                    convertFilename( filename, nest_file, NEST_MESSAGESIZE ),
                    NULL, NULL, 
                    0, ENTIRE_FILE );
	    NestFree( file );
	    return req;

        } else {
	    if ( file != NULL ) NestFree( file );
            nest_debug( NEST_DEBUG_HTTP, "HTTP nest_ad request\n" );
            return new nestRequest( REQUEST_NEST_AD, 
                "http", NULL, NULL, 0, 0);
        }
    }

}

char *
nestProtocolHttp::convertFilename( char *req, char *filename, int len ) {

    char *ret;

    bool simple_storage = true;
    
    if ( simple_storage ) {
        ret = req;
            // skip slashes
        while ( ret[0] == '/' ) ret++;
        return ret; 
    } else {
            // client has already given a full path
        if ( req[0] == '~' ) {
            ret = req;
        } else {
                // else convert the path
            NestSnprintf( filename, len, "~%s%s/%s",
                NestSystemUser,
                nestStorageManager::getRootDir(), 
                req );
            ret = filename;
        }
    }

    nest_debug( NEST_DEBUG_HTTP, 
            "HTTP file request: %s -> %s\n", req, ret );

    return ret;
}

nestRequest*
nestProtocolHttp::closeConnectionRequest() {
    
        // only allow one close connection request to be returned
    if ( already_closed == true ) {
            // now we've moved the smarts for checking already_closed into
            // the client, so we should never get here . . .  
        assert( 0 );
        return NULL;
    } else {
        already_closed = true;
        return new nestRequest( CLOSE_CONNECTION_REQUEST, 0,0,0,0,0 );
    }
}

// this can be made analagous to send file info
// then the data becomes a concurrent transfer
// there is no final ack, since we close the http connection (for now)
bool
nestProtocolHttp::sendMessage( const void *vptr, int errorcode ) {
    char * msg = ( char * ) vptr;
    
    if ( ! sendHeader( NestStrlen( msg ), errorcode ) ) {
        return false;
    }
    return( (int)NestStrlen( msg ) == (int)Sendn( msg, NestStrlen( msg ) ) ); 
}

char *
nestProtocolHttp::errorstring( int errorcode ) {
    static char *notfound = "Not Found";
    static char *badreq   = "Bad Request";
    static char *ok       = "OK";
    static char *unknown  = "Unknown Error";

    switch( errorcode ) {
    case 400: return badreq;
    case 200: return ok;
    case 404: return notfound;
    default:  return unknown;
    }
}

bool
nestProtocolHttp::sendHeader( int len, int errorcode ) {
    char header[NEST_MESSAGESIZE];

    //return true;

    int status = NestSnprintf( header, NEST_MESSAGESIZE,
            "HTTP/1.1 %d %s\r\n"
            "Server: NeST/0.1\r\n"
            "MIME-version: 1.0\r\n"
            "Content-type: text/html\r\n"
            //"Last-Modified: %s\n"
            "Connection: close\r\n"
            "Content-Length: %d\r\n"
            "\r\n", 
            errorcode,
            errorstring( errorcode ),
            len );

    if ( status < 0 ) nest_perror( "NestSnprintf truncation" );

    nest_debug( NEST_DEBUG_HTTP, "Sending HTTP header: %s\n", header );
    return Sendn( header, NestStrlen( header ) );
}
