/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_protocol_manager.h"
#include "nest_config.h"
#include "nest_globus.h"

// array for mapping sockets to their protocols
static int           sockets    [PROTOCOL_COUNT];

static int maxsocket = -1;

static int  NestPort = 0;
static int  HttpPort = 0;
static int   FtpPort = 0;
static int  GftpPort = 0;
static int   NfsPort = 0;
static int ProxyPort = 0;

static int protocol_index = -1;
#define LAST_PROTOCOL_SOCKET -99;

int
nestProtocolManager::init( ) {

    const char *port_string;
    if (GetConfigValue(Config_NestPort, &port_string) && *port_string != 0 ) {
        NestPort = atoi(port_string);
    }
    if (GetConfigValue(Config_ProxyPort, &port_string) && *port_string != 0 ) {
        ProxyPort = atoi(port_string);
    }
    if (GetConfigValue(Config_HTTPPort, &port_string) && *port_string != 0 ) {
        HttpPort = atoi(port_string);
    }
    if (GetConfigValue(Config_FTPPort, &port_string) && *port_string != 0 ) {
        FtpPort = atoi(port_string);
    }
    if (GetConfigValue(Config_GFTPPort, &port_string) && *port_string != 0 ) {
        GftpPort = atoi(port_string);
    }
    if (GetConfigValue(Config_NFSPort, &port_string) && *port_string != 0 ) {
        NfsPort = atoi(port_string);
    }

        // now attempt to set up our gsi credentials if needed 
    if ( NestPort != 0 || GftpPort != 0 ) {
        getCredential();
    }

        // set chirp socket
    if ( NestPort != 0 ) {
        sockets[NEST_PROTOCOL] = nestProtocolNest::listen( NestPort );
    } else {
        sockets[NEST_PROTOCOL] = 0;
    }
    reportOpenPort( "NeST",     NestPort, sockets[NEST_PROTOCOL] );

        // set chirp socket
    if ( ProxyPort != 0 ) {
        sockets[PROXY_PROTOCOL] = nestProtocolNest::listen( ProxyPort );
    } else {
        sockets[PROXY_PROTOCOL] = 0;
    }
    reportOpenPort( "Proxy",     ProxyPort, sockets[PROXY_PROTOCOL] );

        // set http socket
    if ( HttpPort != 0 ) {
        sockets[HTTP_PROTOCOL] = nestProtocolHttp::listen( HttpPort );
    } else {
        sockets[HTTP_PROTOCOL] = 0;
    }
    reportOpenPort( "HTTP",     HttpPort, sockets[HTTP_PROTOCOL] );

        // set ftp socket
    if ( FtpPort != 0 ) {
        sockets[ FTP_PROTOCOL] = nestProtocolGridFtp::listen( FtpPort );
    } else {
        sockets[ FTP_PROTOCOL] = 0;
    }
    reportOpenPort( "FTP ",      FtpPort,  sockets[ FTP_PROTOCOL] );

        // set gftp socket
    if ( GftpPort != 0 ) {
        sockets[GFTP_PROTOCOL] = nestProtocolGridFtp::listen( GftpPort );
    } else {
        sockets[GFTP_PROTOCOL] = 0;
    }
    reportOpenPort( "GFTP", GftpPort, sockets[GFTP_PROTOCOL] );

        // set nfs socket
    if ( NfsPort != 0 ) {
		#ifdef BUILD_NFS
			sockets[NFS_PROTOCOL] = nestProtocolNFS::listen( NfsPort );
		#else
			sockets[NFS_PROTOCOL] = 0;
		#endif
    } else {
        sockets[NFS_PROTOCOL] = 0;
    }
    reportOpenPort( "NFS", NfsPort, sockets[NFS_PROTOCOL] );

    setMaxsocket();

    return 0;
}

NestReplyStatus credentialStatus = NEST_NOT_YET_INITIALIZED;
gss_cred_id_t credential_handle = GSS_C_NO_CREDENTIAL;
OM_uint32 lifetime;
gss_name_t *name;

NestReplyStatus
nestProtocolManager::getCredentialInfo( char **subject, long *expires ) {
    if ( credentialStatus != NEST_SUCCESS ) {
        return credentialStatus;
    }
    NestReplyStatus status = inquireGlobusCredential( 
            name,
            &lifetime,
            &credential_handle );
    *subject = (char*)name;
    *expires = (long)lifetime;
    return status;
}

void
nestProtocolManager::getCredential( ) {
    const char *value = NULL;

        // first check and set any specified environmental values used by 
        // x509 and globus code
    if ( GetConfigValue( Config_GlobusLocation, &value ) && *value ) {
        if ( ! NestIsDir( value ) ) {
            nest_die( "Globus location %s is not a valid directory!\n", value);
        }
        NestSetenv( "GLOBUS_LOCATION", value, 1 );
    }
    if ( GetConfigValue( Config_X509CertDir, &value ) && *value ) {
        if ( ! NestIsDir( value ) ) {
            nest_die( "x509_cert_dir %s is not a valid directory!\n", value);
        }
        NestSetenv( "X509_CERT_DIR", value, 1 );
    }
    if ( GetConfigValue( Config_X509HostCert, &value ) && *value ) {
        if ( ! NestIsFile( value ) ) {
            nest_die( "x509_host_cert %s is not a valid file!\n", value);
        }
        NestSetenv( "X509_USER_CERT", value, 1 );
    }
    if ( GetConfigValue( Config_X509HostKey, &value ) && *value ) {
        if ( ! NestIsFile( value ) ) {
            nest_die( "x509_host_key %s is not a valid file!\n", value);
        }
        NestSetenv( "X509_USER_KEY", value, 1 );
    }
    if ( GetConfigValue( Config_GridMapFile, &value ) && *value ) {
        if ( ! NestIsFile( value ) ) {
            nest_die( "grid_mapfile %s is not a valid file!\n", value);
        }
        NestSetenv( "GRIDMAP", value, 1 );
    }


    credentialStatus = acquireGlobusCredential( 
            &credential_handle, GSS_C_ACCEPT );
    nest_debug( NEST_DEBUG_CHIRP | NEST_DEBUG_AUTH, 
            "Acquired initial globus credential: %s\n",
                NestErrorString( credentialStatus ) );
    if ( credentialStatus != NEST_SUCCESS ) {
        nest_warn( __FILE__, __FUNCTION__, 
                "unable to acquire initial gsi certificate: %s\n",
                NestErrorString( credentialStatus ) );
    }
}

void
nestProtocolManager::reportOpenPort( char *protocol, int port, int status ) {
    if ( status < 0 ) {
        nest_warn( __FILE__, __FUNCTION__, 
                "Unable to listen to %s requests on port %d.\n",
                protocol, port );
    } else if ( status == 0 ) {
        nest_debug( NEST_DEBUG_PROTOCOL, 
                "%s protocol disabled at user request\n", protocol );
    } else {
        nest_debug( NEST_DEBUG_PROTOCOL, 
                "Listening to %s requests on port %d (sd %d).\n", 
                protocol, port, status );
    }
}

NestReplyStatus 
nestProtocolManager::initiateThirdPartyTransfer( nestClient *client, 
        nestProtocol **connection, char *protocol ) 
{
        // also, it might be nice to cache connections
        // also, how about authentication?
        // 

    NestReplyStatus stat;
    nest_debug( NEST_DEBUG_PROTOCOL, "Trying to initiateThirdParty with %s\n",
            protocol );

    if ( strcmp( protocol, "chirp" ) == 0 ) {
        stat = nestProtocolNest::initiateThirdParty
            ( client, connection );
    } else if ( strcmp( protocol, "http" ) == 0 ) {
        stat = nestProtocolHttp::initiateThirdParty( client, connection );
    } else {
        nest_warn( __FILE__, __FUNCTION__, 
                "Unable to negotiate 3 party transfer with %s.\n",
                protocol );
        return NEST_NOT_YET_IMPLEMENTED;
    }

    
    nest_debug( NEST_DEBUG_PROTOCOL,
            "initiateThirdParty with %s: %s\n", protocol,
            NestErrorString( stat ) );
    return stat;
}

// this is used by the nest_worker who is passed a protocol and a csd
nestProtocol* 
nestProtocolManager::newConnection( int csd, int protocol ) {
    switch( protocol ) {
    case NEST_PROTOCOL:
        return (nestProtocol *) new nestProtocolNest( csd );
    case PROXY_PROTOCOL:
        return (nestProtocol *) new nestProtocolProxy( csd );
    case HTTP_PROTOCOL:
        return (nestProtocol *) new nestProtocolHttp( csd ); 
    case STORAGE_PROTOCOL:
        return (nestProtocol *) new nestProtocolFilesystem( csd, (off_t)0 );
    case ADVERTISE_PROTOCOL:
        return (nestProtocol *) new nestProtocolAdvertise( ); 
    default:
            // we should never get here
        assert( 0 );
        return NULL;
    }
}

// this function receives a protocol connection from a socket
nestProtocol *
nestProtocolManager::ReceiveConnection( int sd, int protocol_id ) {
    
    nestProtocol *connection;
    int socket;
    
    assert( protocol_id != GFTP_PROTOCOL && protocol_id != FTP_PROTOCOL );
    
        // receive the fd and create the protocol connection 
    nest_debug( NEST_DEBUG_CONCURRENCY, "Recving fd on pipe %d\n", sd );
    socket = NestReceiveFd( sd );
    nest_debug( NEST_DEBUG_CONCURRENCY, "Recved fd %d on pipe %d\n", 
            socket, sd );
    connection = newConnection( socket, protocol_id );

        // receive any other state if necessary
    nest_debug( NEST_DEBUG_CONCURRENCY, "Recving state for protocol type %d "
            "on pipe %d\n", connection->getType(), sd );
    if ( connection->receive_state( sd ) != NEST_SUCCESS ) {
        assert( 0 );
        return NULL;
    } else {
        nest_debug( NEST_DEBUG_CONCURRENCY, "Recved state on pipe %d\n", sd );
        return connection;
    }

}

// this function sends a protocol connection over a socket
NestReplyStatus
nestProtocolManager::SendConnection( int sd, nestProtocol *connection ) {

    if ( ! connection->hasSocket() ) return NEST_NOT_YET_IMPLEMENTED;
    NestReplyStatus status;

        // pass the fd
    nest_debug( NEST_DEBUG_CONCURRENCY, "Sending fd %d on pipe %d\n", 
            connection->getSocket(), sd );
    if ((status = NestSendFd( sd, connection->getSocket() )) != NEST_SUCCESS ) {
        return status;
    }

    nest_debug( NEST_DEBUG_CONCURRENCY, "Sending state on pipe %d\n", sd );
    status = connection->send_state( sd );
    nest_debug( NEST_DEBUG_CONCURRENCY, "Sent state on pipe %d\n", sd );
    return status; 
}

nestClient* 
nestProtocolManager::accept( int protocol_socket, 
                             fd_set *readies, 
                             int accepts )
{

    nestProtocol *connection = NULL;
    nestClient   *client     = NULL;
    struct sockaddr_in addr;

        // all the protocols create a connection object except NFS
        // which goes ahead and creates the client already
        // some protocols only accept if the fd is ready, others don't care

    switch( getProtocolFromSocket( protocol_socket ) ) {
        case HTTP_PROTOCOL:
            if ( NestFd_isset( protocol_socket, readies ) ) {
                connection = nestProtocolHttp::accept( protocol_socket, &addr );
                NestFd_clear( protocol_socket, readies );
            }
            break;
        case NEST_PROTOCOL:
            if ( NestFd_isset( protocol_socket, readies ) ) {
                connection = nestProtocolNest::accept( protocol_socket, &addr );
                NestFd_clear( protocol_socket, readies );
            }
            break;
        case PROXY_PROTOCOL:
            if ( NestFd_isset( protocol_socket, readies ) ) {
                connection = nestProtocolProxy::accept( protocol_socket, &addr);
                NestFd_clear( protocol_socket, readies );
            }
            break;
        case FTP_PROTOCOL:
        case GFTP_PROTOCOL:
            if ( NestFd_isset( protocol_socket, readies ) ) {
                connection = nestProtocolGridFtp::accept( protocol_socket );
                memset( &addr, 0, sizeof( struct sockaddr_in ) );
            }
            break;
        case NFS_PROTOCOL:
			#ifdef BUILD_NFS
				client     = nestProtocolNFS::accept( protocol_socket, readies, 
                    accepts );
			#else
				client = NULL;
			#endif
            break;
        default:
            assert( 0 );
    }

    if ( connection != NULL ) {
        assert( client == NULL );
        client = new nestClient( connection, &addr );   
    }
    return client;
}

int
nestProtocolManager::getProtocolPortFromSocket( int protocol_socket ) {
    switch( getProtocolFromSocket( protocol_socket ) ) {
    case PROXY_PROTOCOL: return ProxyPort;
    case NEST_PROTOCOL:  return NestPort;
    case HTTP_PROTOCOL:  return HttpPort;
    case FTP_PROTOCOL:   return FtpPort;
    case GFTP_PROTOCOL:  return GftpPort;
    case NFS_PROTOCOL:   return NfsPort;
    default:
        return -1;
    }
}


char *
nestProtocolManager::getProtocolNameFromSocket( int protocol_socket ) {
    return protocolName( getProtocolFromSocket( protocol_socket ) );
}


char *
nestProtocolManager::protocolName( int protocol ) {
    static char gftp_string[] = "GridFTP";
    static char nest_string[] = "Chirp";
    static char ftp_string[]  = "FTP";
    static char http_string[] = "HTTP";
    static char nfs_string[]  = "NFS";
    static char stor_string[] = "Storage";
    static char prox_string[] = "Proxy";

    switch( protocol ) {
    case NEST_PROTOCOL:     return nest_string;
    case HTTP_PROTOCOL:     return http_string;
    case FTP_PROTOCOL:      return ftp_string;
    case GFTP_PROTOCOL:     return gftp_string;
    case NFS_PROTOCOL:      return nfs_string;
    case STORAGE_PROTOCOL:  return stor_string;
    case PROXY_PROTOCOL:    return prox_string;
    default:                return NULL;
    }
}

int
nestProtocolManager::getProtocolFromSocket( int protocol_socket ) {
    if ( protocol_socket == sockets[NEST_PROTOCOL] ) 
        return NEST_PROTOCOL;
    if ( protocol_socket == sockets[HTTP_PROTOCOL] )
        return HTTP_PROTOCOL; 
    if ( protocol_socket == sockets[FTP_PROTOCOL] )
        return FTP_PROTOCOL;
    if ( protocol_socket == sockets[GFTP_PROTOCOL] ) 
        return GFTP_PROTOCOL;
    if ( protocol_socket == sockets[NFS_PROTOCOL] ) 
        return NFS_PROTOCOL;
    if ( protocol_socket == sockets[PROXY_PROTOCOL] )
        return PROXY_PROTOCOL;

        // we better not make it here
    assert( 0 );
    return -1;
}

int
nestProtocolManager::firstProtocolSocket( ) {
    protocol_index = -1;
    return nextProtocolSocket();
}

int
nestProtocolManager::lastProtocolSocket( ) {
    return LAST_PROTOCOL_SOCKET;
}

int
nestProtocolManager::nextProtocolSocket( ) {
    protocol_index++;
    if ( protocol_index == PROTOCOL_COUNT ) {
        return LAST_PROTOCOL_SOCKET;
    } else {    
        return sockets[protocol_index];
    }
}

void
nestProtocolManager::setMaxsocket( ) {
    for( int i = 0; i < PROTOCOL_COUNT; i++ ) {
        if ( sockets[i] > maxsocket ) maxsocket = sockets[i];
    }
}    

int
nestProtocolManager::maxSocket( ) {
    if ( maxsocket <= 0 ) {
        nest_die( "Unable to listen on any ports.\n\tPerhaps a NeST is already "
                "running?\n" );
    }
    return maxsocket;
}    
