
/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent,
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * For contact information, please visit:
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_storage_fd_cache.h"



cache_entry *
entry_new( int fd, const char *key ) {
    cache_entry *entry = (cache_entry*)NestCalloc( sizeof( *entry ) );
    assert( key != NULL );
    entry->fd     = fd;
    entry->key    = NestStrdup( key );
    return entry;
}

void
entry_free( cache_entry *entry ) {
    NestFree( entry->key );
    NestFree( entry );
}

nestFdCache::nestFdCache( unsigned max_size ) {
    this->max_size = max_size;
}

nestFdCache::~nestFdCache( ) {
    this->max_size = 0;
    makeSpace();
}

int
nestFdCache::getSize() {
    return max_size;
}

void
nestFdCache::makeSpace() {
    assert( cache_list.size() == cache_map.size() );

    while( cache_list.size() > max_size ) {
        cache_entry *entry = cache_list.back();
        removeEntry( entry );
        NestClose( entry->fd );
        entry_free( entry );
    }
}

void
nestFdCache::removeEntry( cache_entry *old ) {
    nest_debug( NEST_DEBUG_FD_CACHE, 
            "Removing cached fd %s -> %d\n", old->key, old->fd );

        // find it in the map
    assert( cache_map.count( old->key ) > 0 );
    MAP_ITR mi;
    for(    mi = cache_map.lower_bound( old->key ); 
            mi != cache_map.upper_bound( old->key ); 
            mi++ )
    {
        LIST_ITR li = (*mi).second;
        cache_entry *cursor = *li;
        if ( cursor->fd == old->fd ) {
            assert( cursor == old );
            cache_map.erase( mi );
            cache_list.erase( li );
            break;
        }
    }
    if( mi == cache_map.upper_bound( old->key ) ) {
        nest_die( "Unable to find %s in the multimap\n", old->key );
    }
    assert( cache_list.size() == cache_map.size() );
}

bool 
nestFdCache::stow( int fd, const char *key ) {
    cache_entry *new_entry = entry_new( fd, key );
    cache_list.push_front( new_entry );
    //cache_map[key] = cache_list.begin();
    //cache_map.insert( key, cache_list.begin() );
    cache_map.insert(
            pair<const char* const, LIST_ITR>
                (new_entry->key, cache_list.begin()));

    nest_debug( NEST_DEBUG_FD_CACHE, "Adding cached fd %d for %s\n", fd, key );

    if ( cache_list.size() > max_size ) {
        assert( cache_list.size() - 1 == max_size );
        makeSpace();
    }
    assert( cache_map.size() == cache_list.size() );
    return true;
}

int
nestFdCache::invalidateAll( const char *subkey ) {
    nest_debug( NEST_DEBUG_FD_CACHE, "Asked to invalidate all that match %s\n",
            subkey );
    MAP_ITR cursor = cache_map.begin();
    while( cursor != cache_map.end() ) {
        MAP_ITR mi = cursor;
        if ( strstr( (*mi).first, subkey ) ) {
            nest_debug( NEST_DEBUG_FD_CACHE, 
                "Invalidating and removing %s\n", (*mi).first );
            LIST_ITR li = (*mi).second;
            cache_entry *entry = (*li);
            cache_list.erase( li );
            cache_map.erase( mi );
            NestClose( entry->fd );
            entry_free( entry );
        }
        cursor++;
    }
    return 0;
}
    
int  
nestFdCache::lookup( const char *key ) {
    int fd = -1;

    MAP_ITR mi = cache_map.find( key );
    if ( mi != cache_map.end() ) {
        LIST_ITR li = (*mi).second;
        cache_entry *found = (*li);
        fd = found->fd;
        cache_list.erase( li );
        cache_map.erase( mi );
        entry_free( found );
    }

    if ( fd < 0 && check_nest_debug_status (NEST_DEBUG_FD_CACHE) ) {
        nest_debug(NEST_DEBUG_FD_CACHE, "%s not found.\n", key );
        for( mi = cache_map.begin(); mi != cache_map.end(); mi++ ) {
            LIST_ITR li = (*mi).second;
            nest_debug(NEST_DEBUG_FD_CACHE, 
                    "%s -> %d\n", (*mi).first, (*li)->fd );
        }
    }

    nest_debug(NEST_DEBUG_FD_CACHE, "Sought fd %d from %s\n", fd, key );
    return fd;
}
