/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <assert.h>

#include "nest_speak.h"

// used to know how much total data we transferred
long transfer_len = 0;
long so_far = 0; 
void
quietProgress( int transferred, int total, void *args ) {
    so_far = transferred; 
    transfer_len = total;
}

void
verboseProgress( int transferred, int total, void *args ) {
    struct transferInfo *transfer = ( struct transferInfo *) args;
    struct timeval now;
    NestGettimeofday( &now );
    assert( transfer );
    so_far = transferred;
    transfer_len = total;

    long micros = NestTimediff( &transfer->starttime, &now, __FUNCTION__ );
    float bw = NestBandwidth( transferred, micros );
    int percent = (int)( ( transferred * 100.0 ) / total );

        // print the progress info
    fprintf( stdout, "\r%3d%% %10d / %10d (%6.2f MB/s) ", 
            percent, transferred, total, bw );
    fflush(  stdout );
}


bool
Match( char *one, char *two ) {
    return ( !strcasecmp( one, two ) );
}

void
Usage( char *prog, int line ) {
    fprintf( stderr, 
           "Usage (%d): %s: [-a] [-v] [-p #] [-q] server file1 [file2..]\n"
           "\t-a anonymous connection\n"
           "\t-v verbose; shows progress\n"
           "\t-p specify a port on which to connect to server\n"
           "\t-q silent\n",
           line, prog);
    exit( 1 );
}

void
PrintProxy( char *proxy, void *args ) {
    int *found = (int *)args;
    (*found)++;
    printf( "%s\n", proxy );
}

int
ProxyQuery( NestConnection con) {
    int found = 0;
    NestReplyStatus stat = NestProxyQuery( con, PrintProxy, &found );
    if ( stat != NEST_SUCCESS ) return -1;
    else return 0;
}

int 
main( int argc, char **argv ) {
    char *server;
    int arg = 1; 
    NestProgressFunction *progress = quietProgress;
    struct transferInfo *transferInfo =  NULL; 
    int port = NEST_SERVER_PORT;
    int quiet = 0;
    int anon  = 0;
    int c;

    /*
       // can use this for a function pointer if might either read or write
    NestReplyStatus ( *transfer ) ( char *, char *, NestConnection, 
        NestProgressFunction *, void * ) = NULL;
    */

    char *prog = argv[0];

    while( ( c = getopt( argc, argv, "avqp:" )) != EOF ) {
        switch( c ) {
        case 'a':
            anon = 1;
            break;
        case 'v':
            progress = verboseProgress;
            transferInfo = NestmakeTransferInfo();
            break;
        case 'q':
            quiet = 1;
            break;
        case 'p':    
            port = atoi( optarg );
            break;
        default:
            Usage( prog, __LINE__ );
        }
    }
    arg = optind;

        // next comes server name
    server = argv[arg++];

        // try to connect to the server first
    NestConnection connection;
    NestReplyStatus status;
    if ( anon ) status = NestOpenAnonymously( connection, server, port );
    else status = NestOpenConnection( connection, server, port ); 
    if ( status != NEST_SUCCESS ) {
        fprintf( stderr, "Could not connect to NeST server %s: %s\n",
                server, NestErrorString( status ) );
        exit( 1 );
        exit( -1 );
    }

        // init random
    srand( time( NULL ) );

        // enter a loop and do random reads from specified workload
    struct timeval start, end;
    while( 1 ) {
        int which = arg + (rand() % (argc - arg));
        char *file = argv[which];
        //printf( "Will read %s.\n", file );
        NestGettimeofday( &transferInfo->starttime );
        NestGettimeofday( &start );
        NestReplyStatus status = 
            NestReceiveFile( file, "/dev/null", connection, 
                    progress, transferInfo );
        NestGettimeofday( &end );
        if ( status != NEST_SUCCESS ) {
            fprintf( stderr, "Could not receive %s: %s\n", 
                file, NestErrorString( status ) );
            break;
        } else if ( ! quiet ) {
            fprintf( stderr, "%s: %ld bytes, %.5f MB/s\n", 
                file, so_far, 
                NestBandwidth(transfer_len, &start, &end, __FUNCTION__));
        }
    }
    
    return 0;
}
