#include "nest_common.h"
#include "nest_speak.h"

#include <sys/mman.h>

#define ERROR error( status, __LINE__ )

char filepath[] = "/usr/dict/words";

void
error( int err, int line ) {
    printf( "Error on line %d: %s\n", line, NestErrorString( err ) );
    exit( -1 ); 
}

int
main( int argc, char* argv[] ) {

    NestConnection con;
    NestReplyStatus status;
    long filelength;
    long timestamp;
    char *host;
    char *file;
    int  fp;
    int readsize = 100;
    int offset   = 100;
    char buff1[readsize];
    char buff2[readsize];
    int  nread;
    float nestspace;
    float oldspace;
    
	if (argc < 2) {
		fprintf( stderr, "usage: %s server_hostname\n", argv[0] );
		exit( -1 );
	}
    host = argv[1];
    file = ExtractFilename( filepath );

        // clean out our buffers
    memset( buff1, 0, readsize );
    memset( buff2, 0, readsize );

        // first open our nest connection
    printf( "Opening connection to %s\n", host );
    if ( ( status = OpenNestConnection( con, host ) ) != NEST_SUCCESS ) {
        ERROR;
    }

        // check timestamp
    if ( ( status = GetNestTimestamp( timestamp, con ) ) != NEST_SUCCESS ) {
        ERROR;
    }
    printf( "NeST time is %s", ctime( (time_t *)&timestamp ) );
    
        // first make sure that our test file is not already a nest object
    printf( "Removing any ghost files\n" );
    status = RemoveNestFile( file, con );
    if ( status != NEST_SUCCESS && status != NEST_REMOTE_FILE_NOT_FOUND ) {
        ERROR;
    }
        // file better not exist now
    if (  ( status = NestFilesize( filelength, ExtractFilename( file ), con ) )
            != NEST_REMOTE_FILE_NOT_FOUND ) 
    { ERROR; }

        // check available space at nest
    printf( "Checking available space\n" );
    if ( ( status = AvailableNestSpace( nestspace, con ) ) != NEST_SUCCESS ) {
        ERROR;
    }
    
        // send the file
    printf( "Check sending %s\n", file );
    if ( ( status = SendNestFile( filepath, filepath, con ) ) != NEST_SUCCESS ) {
        ERROR;
    }
    
        // make sure it exists and is the right length
    printf( "Checking length of %s\n", file );
    if (  ( status = NestFilesize( filelength, ExtractFilename( file ), con ) )
            != NEST_SUCCESS ) 
    { ERROR; }
    if ( filelength != Filelength( filepath ) ) {
        printf( "%ld != %ld\n", filelength, Filelength( file ) ); 
    }

        // recheck available space
    printf( "Rechecking available space\n" );
    oldspace = nestspace;
    if ( ( status = AvailableNestSpace( nestspace, con ) ) != NEST_SUCCESS ) {
        ERROR;
    }
    if( nestspace != oldspace - filelength ) {
        printf( "%d WARN: %.0f - %ld != %.0f\n", __LINE__, oldspace, 
                filelength, nestspace);
            // switch to KB's and see if the numbers match up . . . 
        long l_old, l_cur;
        l_old = (long)( oldspace  / 1024 );
        l_cur = (long)( nestspace / 1024 );
        if ( l_old - ( filelength / 1024 ) != l_cur ) {
            printf( "ERROR: %ld - %ld != %ld\n", l_old, filelength/1024, l_cur);
        }
    }

        // delete any local copies and try to read the file
    printf( "Checking receive file %s\n", file );
    if ( Filelength( file ) > 0 ) {
        if ( unlink( file ) < 0 ) {
            perror( "unlink" );
            exit( - 1 );
        }
    }
    fp = Open( file, O_CREAT | O_TRUNC | O_RDWR );
    if ( ( status = ReceiveNestFile( file, file, con ) ) != NEST_SUCCESS ) {
        ERROR;
    }
    Close( fp );
    assert( Filelength( file ) == Filelength( filepath ) );
    printf( "Checking file contents\n" );
    fp = Open( file, O_RDONLY );
    char *copy = (char *) mmap( 0, Filelength( file ), PROT_READ, 
            MAP_PRIVATE, fp, 0 );
    Close( fp );
    fp = Open( filepath, O_RDONLY );
    char *orig = (char *) mmap( 0, Filelength( filepath ), PROT_READ, 
            MAP_PRIVATE, fp, 0 );
    if ( strncmp( copy, orig, Filelength( filepath ) ) != 0 ) {
        printf( "ERROR: Files are different\n" );
        exit( -1 );
    }
    Close( fp );
    if ( munmap( copy, filelength ) < 0 || munmap( orig, filelength ) ) {
        perror( "munmap" );
    }

        // try a partial read
    printf( "Testing a byte read\n" );
    fp = Open( filepath, O_RDONLY );
    if ( lseek( fp, offset, SEEK_SET ) < 0 ) {
        perror( "lseek" );
        exit( -1 );
    }
    if ( Readn( fp, buff1, readsize ) < 0 ) {
        perror( "readn" );
        exit( -1 );
    }
    if ( ( status = ReadNestBytes( con, file, offset, buff2, readsize, nread ) )
            != NEST_SUCCESS ) 
    {
        ERROR;
    }
    if ( strncmp( buff1, buff2, readsize ) != 0 ) {
        printf( "%d: %s != %s\n", __LINE__, buff1, buff2 );
        exit( -1 );
    }

        // try a byte read beyond eof
    printf( "Testing byte read beyond eof\n" );
    offset = filelength - 50;
    if ( lseek( fp, offset, SEEK_SET ) < 0 ) {
        perror( "lseek" );
        exit( -1 );
    }
    if ( ( nread = Read( fp, buff1, readsize ) ) < 0 ) {
        perror( "Read" );
        exit( -1 );
    }
    if ( nread != 50 ) {
        printf( "WARN: nread %d != %d\n", nread, readsize - 50 );
    }
    if ( ( status = ReadNestBytes( con, file, offset, buff2, nread, nread ) ) 
            != NEST_SUCCESS )
    {
        ERROR;
    }
    if ( strncmp( buff1, buff2, nread ) != 0 ) {
        printf( "%d: %s != %s\n", __LINE__, buff1, buff2 );
        exit( -1 );
    }

        // what does read do?
    printf( "Testing behavior of read when offset beyond eof\n" );
    offset = filelength + 1;
    if ( lseek( fp, offset, SEEK_SET ) < 0 ) {
        perror( "lseek" );
        exit( -1 );
    }
    nread = read( fp, buff1, 50 );
    printf( "read returned %d\n", nread );
    perror( "read" );
            
            // try a byte read really beyond eof
    printf( "Testing byte read when offset is beyond eof\n" );
    offset = filelength + 1;
    nread = 50;
    if ( ( status = ReadNestBytes( con, file, offset, buff2, nread, nread ) )
            != NEST_SUCCESS )
    {
        ERROR;
    }
    assert( nread == 0 );

        // try a reservation
    printf( "Checking reservation - assume reserve creates empty file\n" );
    oldspace = nestspace;
/*
    if (( status = ReserveNestSpace(Filelength( filepath ), "my_tmp", con))
            != NEST_SUCCESS )
    {
        ERROR;
    }
*/
    //  Request a 20MB user-space for 1 hour!
	NestLotHandle	handle;
    if (( status = ReserveUserSpace(&handle, con, "20", "0:1:0"))!=NEST_SUCCESS) {
        ERROR;
    }

    if (( status = NestFilesize( filelength, "my_tmp", con)) != NEST_SUCCESS) {
        ERROR;
    }
    if ( filelength != Filelength( filepath ) ) {
        printf( "%d: WARN %.0f != %ld\n", __LINE__, nestspace, filelength );
    }
    if (( status = AvailableNestSpace( nestspace, con ) ) != NEST_SUCCESS ) {
        ERROR;
    }
    if ( nestspace != oldspace - Filelength( filepath ) ) {
        printf( "%d: WARN %.0f != %.0f - %ld\n", __LINE__, nestspace, oldspace,
                Filelength( filepath ) );
    }
    if (( status = RemoveNestFile( "my_tmp", con ) ) != NEST_SUCCESS ) {
        ERROR;
    }
    if (( status = AvailableNestSpace( nestspace, con ) ) != NEST_SUCCESS ) {
        ERROR;
    }
    if ( oldspace != nestspace ) {
        printf( "%d: ERROR: %.0f != %.0f\n", __LINE__, oldspace, nestspace );
        exit( -1 );
    }

        // try a nest byte write
    printf( "Checking byte write\n" );
    char message[] = "This is my secret message";
    status = WriteNestBytes( con, file, offset, message, strlen(message), true);
    if ( status != NEST_SUCCESS ) { ERROR; }
    status = ReadNestBytes( con, file, offset, buff1, strlen(message), nread );
    if ( status != NEST_SUCCESS ) { ERROR; }
    if ( nread != (int)strlen( message ) ) {
        printf( "%d ERROR: %d != %d\n", __LINE__, strlen(message), nread );
        exit( -1 );
    }
    if ( strncmp( buff1, message, nread ) != 0 ) {
        printf( "%d ERROR: %s != %s\n", __LINE__, buff1, message ); 
        exit( -1 );
    }

        // check incremental sends and receives
    printf( "Checking write bytes by sending file incrementally\n" );
    status = RemoveNestFile( file, con );
    filelength = Filelength( filepath );
    fp = Open( filepath, O_RDONLY );
    orig = (char *)mmap( 0, filelength, PROT_READ, MAP_PRIVATE, fp, 0 );
    offset = 0; 
    struct timeval start, cur;
    int writesize = 0;
    Gettimeofday( &start );
    while( offset < filelength ) {
        long remaining = filelength - offset;
        writesize = ( remaining < FILEBUFFER ) ? remaining : FILEBUFFER;
        status = WriteNestBytes( con, file, offset, &orig[offset], writesize, 1);
        if ( status != NEST_SUCCESS ) { ERROR; }
        Gettimeofday( &cur );
        offset += writesize;
        printf( "%.2f, %d\n", Bandwidth( offset, &start, &cur ), writesize );
    }
    status = NestFilesize( filelength, file, con );
    if ( status != NEST_SUCCESS ) { ERROR; }
    assert( offset == filelength );
    assert( filelength == Filelength( filepath ) );

    printf( "Checking read bytes by receiving file incrementally\n" );
    int fp2 = Open( file, O_RDWR | O_CREAT | O_TRUNC );
    char buf[FILEBUFFER];
    nread = 0;
    offset     = 0;
    Gettimeofday( &start );
    do {
        status = ReadNestBytes( con, file, offset, buf, FILEBUFFER, nread );
        Writen( fp2, buf, nread ); 
        if ( status != NEST_SUCCESS ) { ERROR; }
        Gettimeofday( &cur );
        offset += nread;
        printf( "%.2f %d\n", Bandwidth( offset, &start, &cur ), nread );
    } while( nread > 0 );
    
    printf( "Comparing contents\n" );
    copy = (char *)mmap( 0, filelength, PROT_WRITE | PROT_READ, MAP_PRIVATE, 
            fp2, 0 );
    if ( strncmp( copy, orig, Filelength( filepath ) ) != 0 ) {
        printf( "ERROR: Files are different\n" );
        exit( -1 );
    }
    
    printf( "WARN: This test program does NOT test the format function\n" );
	return 0;
}
